/* 
 * This code implements something like the queue of Sterling and
 * Shapiro (1994, p. 299), in which it is possible to dequeue an
 * element before it is enqueued. It is intended to be used as
 * described by van Noord & Gerdemann (2000) in the section on
 * minimization of transducers. The code here combines the output
 * buffer and the queue of identities. When an identity is required
 * (member function pop_identity) then either an existant identy is
 * popped from the queue of identities or buffering is set to true and 
 * pointers to subsequent outputs will be enqueued. Each time an
 * identity is required a null pointer (0) will be enqueued on the
 * output buffer. Whenever an input identity is encountered the buffer 
 * is flushed to an output stream that is specified in the
 * constructor.
 */


#include <iostream>
#include <queue>
#include <cassert>   // Should change to use exceptions


template<class T, class Out>
class output_delay
{
public:
  output_delay(Out& res) : 
    identities(), buffer(), buffering(false),
    res_(res) { }

  // When you're done with an output delay, the buffers should be
  // empty. It would be tempting to put this in the destructor so 
  // that an error would be thrown in this case. But the C++ rule is
  // that destructors should never throw. Otherwise you can't heve
  // exception safe and exception neutral containers (Sutter,
  // "Exceptional C++).

  bool final_check() {
    return (buffer.empty() && identities.empty());
  }

  void push_output(const T& x) { 
    if (buffering) 
      buffer.push(&x);
    else 
      *res_++ = x;	
  }



  void push_identity(const T& x) {
    if (buffering) {
      *res_++ = x;
      // In this case the the top of the buffer should be a null pointer
      assert(!buffer.front());
      buffer.pop();

      // Flush the buffer as much as you can, i.e., up to the next
      // null pointer indicating an identity yet to be encoutered
      // on the input side
      while (!buffer.empty() && buffer.front()) {
	*res_++ = *buffer.front();
	buffer.pop();
      }
      if (buffer.empty()) buffering = false;
    } else {
      identities.push(x);
    }
  }


  void pop_identity() {
    if (!identities.empty()) {
      *res_++ = identities.front();
      identities.pop();
    } else {
      buffer.push(0);
      buffering = true;
    }
  }

  void pop_phantom_identity() {
    assert(!identities.empty());
    identities.pop();
  }


private:
  queue<T> identities;
  queue<const T*> buffer;
  bool buffering;
  Out& res_;  // This makes class uncopyable
  output_delay(const output_delay&);  // copy constructor
  output_delay& operator=(const output_delay&); // assignment
};


int main()
{

  // The following sequence prints: 354687910
  // The sequences of pushes and pops represents the following sequence
  // of transitions in the form input/output and ignoring states:
  //
  // 1. pred/[3]
  // 2. id(pred)/[5]     where pred matches 4
  // 3. pred/[id(pred)]  where pred corresponds to pred in line 2
  // 4. pred/[6,id(pred),7]   where pred corresponds to pred in line 5
  // 5. id(pred)/[9]          where pred matches 8
  // 6. id(pred)/[1]          where pred matches 2
  // 7. pred/[0,dequeue]      where dequeue throws away what was matched in 6

  ostream_iterator<int> os(cout);
  output_delay<int,ostream_iterator<int> > q(os);

  q.push_output(3);
  q.push_identity(4);
  q.push_output(5);
  q.pop_identity();
  q.push_output(6);
  q.pop_identity();
  q.push_output(7);
  q.push_identity(8);
  q.push_output(9);
  q.push_identity(2);
  q.push_output(1);
  q.push_output(0);
  q.pop_phantom_identity();

  // This line checks that the buffers are empty at the end
  cout << '\n' << q.final_check() << endl;

}


