:- multifile macro/2.

%% Finite-state grapheme to phoneme conversion for Dutch, as described in:
%% Gosse Bouma, A finite-state and data-oriented method for grapheme to
%% phoneme conversion, Proceedings of NAACL 2000,
%% X-archiv: http://xxx.lanl.gov/abs/cs.CL/0003074


%% define replace operator (Gerdemann van Noord 99)
:- ensure_loaded('../GerdemannVannoord99/replace_nm').

%% spy points for regular expression compiler
%% useful to see what's going on (use debug level 1)
:- bb_put(fsa_rx_spy:string2phon,on).
:- bb_put(fsa_rx_spy:segment,on).
:- bb_put(fsa_rx_spy:graph2phon,on).
:- bb_put(fsa_rx_spy:graph2freq,on).

:- bb_put(fsa_rx_spy:mark_begin_end,on).
:- bb_put(fsa_rx_spy:remove_markers,on).

:- bb_put(fsa_rx_spy:conversion,on).
:- bb_put(fsa_rx_spy:conversion_to_most_freq,on).
:- bb_put(fsa_rx_spy:special_vowel_rules,on).
:- bb_put(fsa_rx_spy:short_vowel_rules,on).
:- bb_put(fsa_rx_spy:special_consonant_rules,on). 
:- bb_put(fsa_rx_spy:no_context_rules,on). 

%%%%%%%%%%%% Top-level expressions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% 1. Do everything from character string to phoneme string
macro(g2p,[[cache(string2phon),' '+]*,cache(string2phon), ' '*]).

macro(string2phon,
		segment 
	o 	graph2phon
	o	remove_markers
	).

%% 2. Segmentation only
macro(segment, 
	replace([ idsegment, [] : '-' ],[],[]) 
     ).

%% 3. Conversion of segmented input 
macro(graph2phon,
      	sigma({a..z,'-'},{vowel,cons,'-','+','#'},
	     mark_begin_end
	   o ( ~ $[-,-] )
	   o conversion
	   )
     ).

%% 4. Conversion of segmented input to most frequent phoneme
%%    (Used to compare hand-crafted rules with a base-line using frequency.)
macro(graph2freq, 
	sigma({a..z,'-'},{vowel,cons,'-','+','#'},
	       mark_begin_end
	     o ( ~ $['-','-'] )
	     o conversion_to_most_freq
	     )	  
     ).

%%%%%%%%%%%%%%%%%%%%%%%%%% Utils %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% aa-n-b-i-dd-en -> #-aa-n-b-i-dd-en#
macro(mark_begin_end, 
	  [[] x ['#','-'], ? * ,[]:'#'] 
	% make contexts stricly local, otherwise we can't t_determinize...!
	o ( ~ $[-, -] )
     ).

macro(remove_markers, replace({'#','-','+'} x [], [],[])  ).

macro(vowel,{a,'A',')',e,'E','@',i,'I',o,'O',u,y,'K','M','L','|','}'}).

macro(cons,{b,c,d,f,g,'G',h,j,k,l,m,n,'N',p,q,r,s,'S',t,v,w,x,z}).

macro(non_vowel,{cons, '#'}).

%% g2p: zorg ervoor dat op ieder segment slechts 1 regel van toepassing is
%% (door rechts het scheidingsteken '-' te vervangen door '+')
%% g2p-macro: Ensure that at most a single rule applies to each segment
%% and that each rule applies to a full segment.
macro(g2p(Target,LeftContext,RightContext),
      replace([Target,'-':'+'],
	      [ignore(LeftContext,{+,-}),{-,+}],
	      ignore(RightContext,{+,-})
	     )
     ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%% Segments %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

macro(idsegment, identity({a, [a,a], [a,u], [a,i], [a,a,i],
			   e, [e,e], [e,i], [e,u], [e,a,u], [e,e,u],
			   i, [i,e], [i,e,e], [i,e,u], [i,j], 
			   o, [o,e], [o,e,i], [o,o], [o,o,i], [o,u],
			   u, [u,u], [u,i],
			   y,
			   b, [b,b],
			   c, [c,c], [c,h],[c,k],
			   d, [d,d], [d,t],
			   f, [f,f], [f,v],
			   g, [g,g],
			   h, 
		           j,
			   k, [k,k],
			   l, [l,l],
			   m, [m,m],
			   n, [n,n], [n,g],
			   p, [p,b], [p,p], [p,h],
			   q, [q,u],
			   r, [r,r],
			   s, [s,c,h], [s,s],[s,j],[s,h],[s,z],
			   t, [t,t],[t,h],[t,i,e],[t,i,e,e],[t,i,o],
			   v,
			   w,
			   x,
			   z			   
			  } )
     ).

%%%%%%%%%%%%%%%%%%%%%%%%%% Conversion Rules   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%

macro(conversion,
        special_vowel_rules   
      o short_vowel_rules
      o g2p(e: @, [], [])
      o special_consonant_rules
      o	no_context_rules
     ).

macro(special_vowel_rules,      
        g2p([a,i] x ')',       [],  r)         %% revolutionair
      o g2p([e,u] x [e,j,'}'], [],  m)         %% petroleum
      o g2p([] x j,   {i,e}, {[e,e],a,o,u})    %% pias pion video
      o g2p(y x 'I',       [], {n,f,p,b})
      o g2p(i x '@',     [],     g)            %% moedig(st)
      o g2p([i,j] x '@', l,      k)            %% mogelijkheid    
     ).

macro(short_vowel_rules,
	g2p({ a x 'A'                          %% bak, bakken
	    , i x 'I'                          %% wit
	    , o x 'O'                          %% bok
	    , u x '}'                          %% dun
	    },[],[cons,non_vowel]
	   )
	o g2p(e: 'E', [], {[t,t],[k,k],x,[n,{n,g,[t,i,e]}],[g,{g,s,d}],
			   [d,d],[m,m],[f,f],[b,b],[c,{c,h,k,t}],
			   [s,s,i],[k,{s,t}], [{k,t},#], [p,t]})

	).

macro(special_consonant_rules,
        g2p(b x p,         [], {s, t, '#'})
      o g2p([c:k^,c:s], [], {e,'E',i,'I',y})	
      o g2p([d,t^] x t,         [], {s,g,k,j,v,h,z,'#'})	
      o g2p({ f:v, s:z},      [], {b,d})	
      o g2p(g x 'G',       vowel, vowel)
      o g2p(g x 'G',       [r], [])              
      o g2p(n x 'N',      [], {k,q})
      o g2p(n x [],        [@],['#'])
      o g2p(n x m,         [], {b,p})
      o g2p([n,g] x ['N','G'], {a,['#',{'I','O'}]}, [])
                       %% aangaven, ingeval, ongestraft
      o g2p([q,u] x [k,w], [], [a])
      o g2p(s x z,         [], [d])
      o g2p([s,c,h] x s,   [i],[])
      o g2p([s,j] x 'S',   [#],[])           %% sjoelen
      o g2p({ v x f,  z x s},      {d,f,g,p,s,t}, [])
       ).

macro(no_context_rules, 
	g2p({ %% 1. long vowels and diphtongs
	      [a,a] x a          %% slaap
	    , [a,a,i] x [a,j]    %% waai-en 
            , [a,i] x e          %% container
	    , [a,u] x 'M'        %% blauw
	    , [e,a,u] x o        %% cadeau
	    , [e,e] x e          %% beek
	    , [e,e,u] x e        %% leeuw
	    , [e,i] x 'K'        %% eik
	    , [e,u] x '|'        %% deur
	    , [i,j] x 'K'        %% wijn
	    , [i,e] x i          %% bier
	    , [i,e,e] x [i,j,e]  %% prieel
	    , [i,e,u] x i        %% nieuw
	    , [o,e] x u          %% schoen
	    , [o,e,i] x [u,j]    %% moeilijk
	    , [o,o] x o          %% boot
	    , [o,o,i] x [o,j]    %% mooi
	    , [o,u] x 'M'        %% koud
	    , [u,u^] x  y        %% duur
	    , [u,i] x 'L'        %% duif
	    , y x i              %% typisch   
	    %% 2. double consonants
	    , [b,b] x b          %% ribben
	    , [c,{c,k}^] x k     %% accu, 
	    , [d,d] x d          %% etc.
	    , [f,{f,v}] x f
	    , g x 'x'
	    , [g,g] x 'G'
	    , [k,k] x k
	    , [l,l] x l
	    , [m,m] x m
	    , [n,n] x n
	    , [p,p] x p
	    , [q,u] x k 
	    , [r,r] x r
	    , [s,{s,z}] x s
	    , [t,t] x t
	    %% 3. misc consonants
	    , [c,h] x 'x'        %% lachen
	    , [s,c,h] x [s,x]    %% school
	    , [s,h] x 'S'        %% show sjoelen
	    , [n,g] x 'N'        %% lang
	    , [t,i,e] x [t,s,i]  %% organisatie
	    , [t,i,e,e] x ['S',e]%% differentieel
	    , [t,i,o] x ['S',o]  %% differentieel
	    , 'x' x [k,s]        %% sex
	    },
	    [], []   
	   ) 
     ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%% Frequency data %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
     
macro(conversion_to_most_freq,
      g2p({ '#' x '#'
	, a x 'A'
	, [a,a] x a
	, [a,a,i] x [a,j]
	, [a,i] x ')'
	, [a,u] x 'M'
	, b x b
	, [b,b] x b
	, c x k
	, [c,c] x k
	, [c,h] x x
	, [c,k] x k
	, d x d
	, [d,d] x d
	, [d,t] x [d,t]
	, e x '@'
	, [e,a,u] x o
	, [e,e] x e
	, [e,e,u] x e
	, [e,i] x 'K'
	, [e,u] x [e,j,'}']
	, f x f
	, [f,f] x f
	, [f,v] x f
	, g x 'G'
	, [g,g] x 'G'
	, h x h
	, i x '@'
	, [i,e] x i
	, [i,e,e] x [i,j,e]
	, [i,e,u] x i
	, [i,j] x '@'
	, j x j
	, k x k
	, [k,k] x k
	, l x l
	, [l,l] x l
	, m x m
	, [m,m] x m
	, n x []
	, [n,g] x 'N'
	, [n,n] x n
	, o x 'O'
	, [o,e] x u
	, [o,e,i] x [u,j]
	, [o,o] x o
	, [o,o,i] x [o,j]
	, [o,u] x 'M'
	, p x p
	, [p,b] x [p,b]
	, [p,h] x [p,h]
	, [p,p] x p
	, [q,u] x [q,u]
	, r x r
	, [r,r] x r
	, s x s
	, [s,c,h] x [s,x]
	, [s,h] x 'S'
	, [s,j] x 'S'
	, [s,s] x s
	, [s,z] x s
	, t x t
	, [t,h] x [t,h]
	, [t,i,e] x [t,s,i]
	, [t,i,e,e] x ['S',e]
	, [t,i,o] x ['S',o]
	, [t,t] x t
	, u x y
	, [u,i] x 'L'
	, [u,u] x y
	, v x f
	, w x w
	, x x [k,s]
	, y x 'I'
	, z x s
	  }, [], []
	 )
     ). 
      
