%%% (c) Edmund Grimley-Evans.
%%% adapted for fsa6 by Gertjan van Noord
%%%
%%% cf. http://www.bell-labs.com/projects/tts/gkiraz-wia97.ps
%%% George Anton Kiraz and Edmund Grimley-Evans, Multi-Tape Automata
%%% for Speech and Language Systems: A Prolog Implementation.  
%%% In D. Wood and S. Yu (eds.), Automata Implementation,
%%% Lecture Notes in Computer Science 1436, Springer, 1998.

:- module(ge_minimise, [ ge_minimise/2 ]).

ge_minimise(Fsa0,Fsa):-
    fsa_data:fsa_symbols_decl(Fsa0,Syms),
    (	Syms \== r(fsa_frozen)
    ->	fsa_util:debug_message(0,
         "Warning: this minimizer ignores symbols on predicates~n",
		      [])
    ;   true
    ),
    retractall(final_(_)),
    retractall(transition_(_,_,_)),
    fsa_data:fsa_components(Symbols,_,Starts0,Finals0,Trans0,[],Fsa0),
    fsa_util:time(ge_minimise:fsa_to_ge(Finals0,Trans0)),
    fsa_util:time(ge_minimise:minimise),
    fsa_util:time(ge_minimise:ge_to_finals(Finals)),
    fsa_util:time(ge_minimise:ge_to_trans(Trans)),
    fsa_util:time(ge_minimise:ge_to_start(Starts0,Starts)),
    fsa_util:time(fsa_data:fsa_construct_rename_states(Symbols,Starts,Finals,Trans,[],Fsa)).

fsa_to_ge(Finals,Transitions):-
    assert_finals(Finals),
    assert_transitions(Transitions).

ge_to_start([],[]).
ge_to_start([H|T],[C|R]):-
    state_class_ref(H,C),
    ge_to_start(T,R).

ge_to_finals(Finals):-
    findall(Final,a_final(Final),Finals0),
    sort(Finals0,Finals).

ge_to_trans(Trans):-
    findall(trans(P,Sym,Q),a_transition(P,Sym,Q),Trans0),
    sort(Trans0,Trans).

a_final(Final) :-
    final_(F),
    state_class_ref(F,Final).

a_transition(P,Sym,Q):-
    transition_(P0,Sym,Q0),
    state_class_ref(P0,P),
    state_class_ref(Q0,Q).

state_class_ref(A,B):-
    (	state_class_ref_(A,B,_)
    ->	true
    ;   A=B        % for certain borderline cases, epsilon, eset
    ).

assert_transitions([]).
assert_transitions([trans(P,Sym,Q)|T]):-
    assertz(transition_(P,Sym,Q)),
    assert_transitions(T).

assert_finals([]).
assert_finals([H|T]):-
    assertz(final_(H)),
    assert_finals(T).

%% minimise/0 should be supplied with:
%%
%% transition_/3 - transitions
%% final_/1 - final states

minimise :-
  minimise1, % set up state_from_state_/3
  minimise2, % set up class_state_/2, state_class_ref_/3 and class_/1
  minimise3, % set up class_count_/2 and class_to_refine_with_/1
  minimise_loop.


minimise1 :-
  retractall(state_from_state_(_,_,_)),
  transition_(From,Symbol,To),
  \+ state_from_state_(To,Symbol,From),
  assert(state_from_state_(To,Symbol,From)),
  fail.
minimise1.

minimise2 :-
  retractall(class_state_(_,_)),
  retractall(state_class_ref_(_,_,_)),
  retractall(class_(_)),
  assert(class_(2)),
  state_from_state_(State,_,_),
  \+ state_class_ref_(State,_,_),
  minimise2(State),
  fail.
minimise2 :-
  state_from_state_(_,_,State),
  \+ state_class_ref_(State,_,_),
  minimise2(State),
  fail.
minimise2.

minimise2(State) :-
  final_(State),
  !,
  assert(class_state_(1,State),Ref),
  assert(state_class_ref_(State,1,Ref)).
minimise2(State) :-
  assert(class_state_(0,State),Ref),
  assert(state_class_ref_(State,0,Ref)).

minimise3 :-
  retractall(class_count_(_,_)),
  findall([],class_state_(0,_),L0),
  length(L0,N0),
  assert(class_count_(0,N0)),
  findall([],class_state_(1,_),L1),
  length(L1,N1),
  assert(class_count_(1,N1)),
  N1 < N0,
  !,
  assert(class_to_refine_with_(0)), %???
  assert(class_to_refine_with_(1)).
minimise3 :-
  assert(class_to_refine_with_(0)),
  assert(class_to_refine_with_(1)). %???


% The minimisation algorithm itself, using:
%
% state_from_state_/3 - transitions recorded with the destination first
% class_/1 - count for numbering the equivalence classes
% state_class_ref_/3 - map from state number to equivalence class number
%                      plus identifier of corresponding class_state_/2 clause
% class_state_/2 - inverse of state_class_/2
% class_count_/2 - map from equivalence class to number of states it contains
% class_to_refine_with_/1 - asserted if class still needs to be used for splitting
% symbol_state_/2 - temporarily records Symbol-From pairs
% class_twin_/2 - temporarily records new classes made from old ones

:- dynamic symbol_state_/2.
:- dynamic class_twin_/2.
:- dynamic state_from_state_/3.
:- dynamic class_to_refine_with_/1.
:- dynamic state_class_ref_/3.
:- dynamic class_state_/2.
:- dynamic class_state_ref_/3.
:- dynamic class_count_/2.
:- dynamic class_/1.
:- dynamic transition_/3.
:- dynamic final_/1.

minimise_loop :-
  retract(class_to_refine_with_(Class)),
  !,
  refine_with_class(Class),
  minimise_loop.
minimise_loop.

refine_with_class(Class) :-
  setup_symbol_state(Class),
  refine_partition.

setup_symbol_state(Class) :-
  class_state_(Class,State),
  state_from_state_(State,Symbol,FromState),
  assert(symbol_state_(Symbol,FromState)),
  fail.
setup_symbol_state(_).

refine_partition :-
  symbol_state_(Symbol,_),
  !,
  refine_partition(Symbol),
  refine_partition.
refine_partition.

refine_partition(Symbol) :-
  retract(symbol_state_(Symbol,State)),
  retract(state_class_ref_(State,Class,Ref)), erase(Ref),
  class_twin(Class,NewClass),
  assert(class_state_(NewClass,State),NewRef), assert(state_class_ref_(State,NewClass,NewRef)),
  fail.
refine_partition(_) :-
  retract(class_twin_(Class1,Class2)),
  add_to_queue(Class1,Class2),
  fail.
refine_partition(_).

class_twin(Class,NewClass) :-
  class_twin_(Class,NewClass),
  !,
  retract(class_count_(Class,A)), A1 is A-1, assert(class_count_(Class,A1)),
  retract(class_count_(NewClass,B)), B1 is +(B,1), assert(class_count_(NewClass,B1)).
class_twin(Class,Twin) :-
  retract(class_(Twin)), Twin1 is +(Twin,1), assert(class_(Twin1)),
  assert(class_twin_(Class,Twin)),
  retract(class_count_(Class,A)), A1 is A-1, assert(class_count_(Class,A1)),
  assert(class_count_(Twin,1)).

add_to_queue(Class,NewClass) :-
  class_to_refine_with_(Class),
  !,
  assert(class_to_refine_with_(NewClass)).
add_to_queue(Class,NewClass) :-
  class_count_(Class,Count),
  class_count_(NewClass,NewCount),
  Count =< NewCount,
  !,
  assert(class_to_refine_with_(Class)).
add_to_queue(_,NewClass) :-
  assert(class_to_refine_with_(NewClass)).

