:- multifile rx/2.
:- multifile macro/2.


rx(baum(Expr,Sequence),Fa):-
    fsa_regex:rx(Expr,Fa0),
    baum(Fa0,Sequence,Fa).

rx(round(Expr),Fa):-
    fsa_regex:rx(Expr,Fa0),
    round(Fa0,Fa).

rx(tohmm(Expr),Fa):-
    fsa_regex:rx(Expr,Fa0),
    add_probs(Fa0,Fa).

rx(ex(N),Fa):-
    ex(N,Fa).

macro(vowel, {a,e,i,o,u}).
macro(cons,  {b,c,d,f,g,h,j,k,l,m,n,p,r,s,t,v,w,z}).

macro(words,[[cons*,vowel*,cons*],#]*).

rx(assigns(Expr,Symbols),Fa):-
    fsa_regex:rx(Expr,Fa),
    assigns(Fa,Symbols,Score),
    length(Symbols,L),
    Pwce is -(log(2,Score)/L),
    format(user_error,"assigns ~w (~w)~n",[Score,Pwce]).

%% charniak figure 4.1
ex(1, fa(r(fsa_preds,fsa_frozen),2,[0],[0,1],
	 [trans(0,a/0.4,0),
	  trans(0,a/0.3,1),
	  trans(0,b/0.2,0),
	  trans(0,b/0.1,1),
	  trans(1,a/0.2,0),
	  trans(1,a/0.2,1),
	  trans(1,b/0.5,1),
	  trans(1,b/0.1,0)
	 ],[])).

%% charniak figure 4.9.b
ex(2, fa(r(fsa_preds,fsa_frozen),2,[0],[0,1],
	 [trans(0,0/0.04,1),
	  trans(0,0/0.48,0),
	  trans(0,1/0.48,0),
	  trans(1,1/1.00,0)
	 ],[])).

ex(3, fa(r(fsa_preds,fsa_frozen),3,[0],[0,1,2],
	 [trans(0,a/0.15,2),
	  trans(0,a/0.6,1),
	  trans(0,b/0.1,2),
	  trans(0,b/0.15,1),
	  trans(1,a/0.05,0),
	  trans(1,a/0.05,1),
	  trans(1,a/0.03,2),
	  trans(1,b/0.07,2),
	  trans(1,b/0.8,0),
	  trans(2,a/0.7,0),
	  trans(2,b/0.03,0),
	  trans(2,b/0.17,2)
	 ],[])).


ex(4,fa(r(fsa_preds,fsa_frozen),6,[0],[0,1,2,3,4,5],
	Trans,[])) :-
    random_product([0,1,2,3,4,5],
		   [a,b,c,d,e,f],Trans0),
    sort(Trans0,Trans).
   

ex(5,fa(r(fsa_preds,fsa_frozen),4,[0],[0,1,2,3],
	Trans,[])) :-
    random_product([0,1,2,3],
		   [a,b,c,d],Trans0),
    sort(Trans0,Trans).
   
ex(6,fa(r(fsa_preds,fsa_frozen),2,[0],[0,1],
	Trans,[])):-
    random_product([0,1],
		   [a,b],Trans0),
    sort(Trans0,Trans).


ex(7,fa(r(fsa_preds,fsa_frozen),3,[0],[0,1,2],
	 [trans(0,0/0.5,1),
	  trans(0,1/0.5,2),
	  trans(1,0/0.99,1),
	  trans(1,1/0.01,1),
	  trans(2,0/0.01,2),
	  trans(2,1/0.99,2)
	 ],[])).

baum(Fa0,Symbols,Fa):-
    nl,nl,
    baum(Fa0,Symbols,Fa,0).

%% Fa0: string-to-weight transducer
%%      without jumps and []/..
%%      all states are final
baum(Fa0,Symbols,Fa,Score0) :-
    fsa_data:fsa_states_set(Fa0,States),
    %% forward pass
    alphas(Symbols,States,Fa0,Alphas),
    lists:append(_,[K],Alphas),
    sum_vals(K,0,Score),
    length(Symbols,L),
    Pwce is -(log(2,Score)/L),
    format(user_error,"assigns ~w (~w)~n",[Score,Pwce]),
    %% continue as long as there is improvement
    (   Score > Score0
    ->	fsa_data:fsa_copy_except(transitions,Fa0,Fa1,Trans0,NewTrans),
	betas(Symbols,States,Fa0,Betas),   %% backward pass
	transition_probabilities(Trans0,Alphas,Betas,Trans),
	estimate_parameters(Trans0,Trans,States,NewTrans),
	baum(Fa1,Symbols,Fa,Score)
    ;   Fa0=Fa
    ).

alphas(Symbols,States,Fa0,[PrevAlpha|Alphas]):-
    fsa_data:fsa_start_state(Fa0,Start),
    treat_start_state(States,Start,PrevAlpha),
    alphas(Symbols,States,PrevAlpha,Fa0,Alphas).

treat_start_state([],_,[]).
treat_start_state([H|T],Start,[H-Alpha|T2]) :-
    (   Start==H
    ->	Alpha=1
    ;   Alpha=0
    ),
    treat_start_state(T,Start,T2).

%% forward 
alphas([],_,_,_,[]).
alphas([Sym|Syms],States,Prev,Fa0,[Next|Tail]) :-
    alphas0(States,Sym,Prev,Next,Fa0),
    alphas(Syms,States,Next,Fa0,Tail).

alphas0([],_,_,[],_).
alphas0([State|States],Sym,Prevs,[State-Next|Nexts],Fa0):-
    %% get all transitions t on Sym ending in State
    %% sum over prev_alpha(source(t)).P(t)
    findall(P,(   fsa_data:fsa_transition(Fa0,Source,Sym/Score,State),
		  lists:memberchk(Source-Alpha1,Prevs),
		  P is *(Score,Alpha1)
	      ),Ps),
    sum(Ps,0,Next),
    alphas0(States,Sym,Prevs,Nexts,Fa0).

sum([],L,L).
sum([H|T],L0,L) :-
    L1 is +(H,L0),
    sum(T,L1,L).

%% backward 
betas(Symbols0,States,Fa0,Betas):-
    lists:reverse(Symbols0,Symbols),
    treat_last(States,PrevBeta),
    betas(Symbols,States,PrevBeta,Fa0,[PrevBeta],Betas).

treat_last([],[]).
treat_last([H|T0],[H-1|T]) :-
    treat_last(T0,T).

betas([],_,_,_,Bs,Bs).
betas([Sym|Syms],States,Prev,Fa0,Bs0,Bs) :-
    betas0(States,Sym,Prev,Next,Fa0),
    betas(Syms,States,Next,Fa0,[Next|Bs0],Bs).

betas0([],_,_,[],_).
betas0([State|States],Sym,Prevs,[State-Next|Nexts],Fa0):-
    %% get all transitions t on Sym starting in State
    %% sum over prev_beta(target(t)).P(t)
    findall(P,(   fsa_data:fsa_transition(Fa0,State,Sym/Score,Target),
		  lists:memberchk(Target-Beta1,Prevs),
		  P is *(Score,Beta1)
	      ),Ps),
    sum(Ps,0,Next),
    betas0(States,Sym,Prevs,Nexts,Fa0).

%% step 3
transition_probabilities(Trans0,[Alpha|Alphas],[_beta|Betas],Trans):-
    tp(Alphas,Alpha,Betas,Trans0,Trans).

tp([],_,_,_,[]).
tp([Ha|Ta],Alpha,[Beta|Betas],Trans0,[Tr0|Trans]):-
    tp1(Trans0,Alpha,Beta,Tr0),
    tp(Ta,Ha,Betas,Trans0,Trans).

tp1([],_,_,[]).
tp1([trans(P,_Sym/W,Q)|Trs0],Alpha,Beta,[P-NewW|Trs]):-
    lists:memberchk(P-W0,Alpha),
    lists:memberchk(Q-W1,Beta),
    NewW is *(*(W,W0),W1),
    tp1(Trs0,Alpha,Beta,Trs).

count([H|T],R):-
    count(H,T,R).
count([],_,[]).
count([Key-H|T],R,[Key-First|Rest]):-
    scalp([[Key-H|T]|R],Row,[Next0|Next]),
    sum(Row,0,First),
    count(Next0,Next,Rest).

%% scalp(In,Heads,Tails)
%% take all heads without key of a list of lists
scalp([],[],[]).
scalp([[_key-HdHd|HdTail]|Tail],[HdHd|T0],[HdTail|T1]):-
    scalp(Tail,T0,T1).

%% step 4
estimate_parameters(Trans0,Trans,States,NewTrans):-
    count(Trans,TrCount),  % per transition
    count_states(States,TrCount,StatesCount),
    ep(Trans0,TrCount,StatesCount,NewTrans).

count_states([],_,[]).
count_states([State|States],TrCount,[State-C|Tail]) :-
    findall(W,lists:member(State-W,TrCount),C0),
    sum(C0,0,C),
    count_states(States,TrCount,Tail).

ep([],_,_,[]).
ep([trans(A,B/_,C)|Tail],[A-W1|Tail1],StatesCount,[trans(A,B/W,C)|Tail2]):-
    lists:memberchk(A-W2,StatesCount),
    (	W1 =:= 0
    ->	W = 0
    ;	W is W1/W2
    ),
    ep(Tail,Tail1,StatesCount,Tail2).


%% end


%% other stuff
assigns(Fa0,Symbols,Score):-
    fsa_data:fsa_states_set(Fa0,States),
    alphas(Symbols,States,Fa0,Alphas),
    lists:append(_,[K],Alphas),
    sum_vals(K,0,Score).

sum_vals([],L,L).
sum_vals([_key-H|T],L0,L) :-
    L1 is +(H,L0),
    sum_vals(T,L1,L).



%% utils for testing

%% random_product(States,Symbols,Trans)
%% all transitions from states to states with symbols, s.t.
%% probabilities add up to one per source state.
random_product(States,Symbols,Trans):-
    findall(trans(St0,Sym/Random,St),
	    (	lists:member(St0,States),
		lists:member(St,States),
		lists:member(Sym,Symbols),
		random:random(0,100,Random)
	    ), Trans0),
    pen_per_source(States,Trans0,PerState),
    div_trans(Trans0,Trans,PerState).

pen_per_source([],_,[]).
pen_per_source([H|T],Trans,[H-Total|Totals]):-
    findall(W,lists:member(trans(H,_/W,_),Trans),Total0),
    sum(Total0,0,Total),
    pen_per_source(T,Trans,Totals).

div_trans([],[],_).
div_trans([trans(A,B/W0,C)|T0],[trans(A,B/W,C)|T],Totals):-
    lists:member(A-W1,Totals),
    W is W0/W1,
    div_trans(T0,T,Totals).


round(Fa0,Fa):-
    fsa_data:fsa_copy_except(transitions,Fa0,Fa,Trans0,Trans),
    round_tr(Trans0,Trans).

round_tr([],[]).
round_tr([H|T],Trans):-
    round_tr0(H,Trans,Trans0),
    round_tr(T,Trans0).

round_tr0(trans(P,L/R0,Q),Trans,Trans0):-
    R is round(*(R0,1000))/1000,
    (	R =:= 0
    ->	Trans=Trans0
    ;   Trans=[trans(P,L/R,Q)|Trans0]
    ).

add_probs(Fa0,Fa):-
    fsa_data:fsa_start_states(Fa0,Start),
    fsa_data:fsa_final_states(Fa0,Final),
    fsa_data:fsa_transitions(Fa0,Trans0),
    fsa_data:fsa_jumps(Fa0,Jumps),
    fsa_data:fsa_states_number(Fa0,States),
    fsa_data:fsa_states_set(Fa0,AllStates),
    add_probs_tr(Trans0,Trans,AllStates),
    fsa_data:fsa_construct(r(fsa_preds,fsa_frozen),
			   States,Start,Final,Trans,Jumps,Fa).

add_probs_tr(Trans0,Trans,States):-
    add_one(Trans0,Trans1),
    pen_per_source(States,Trans1,PerState),
    div_trans(Trans1,Trans,PerState).

add_one([],[]).
add_one([trans(A,B,C)|T0],[trans(A,B/1,C)|T]):-
    add_one(T0,T).

