/*----------------------------------------------------------------------
msol.pl

http://www.ling.ohio-state.edu/~nvail/msol.html

A monadic second-order logic front end for
Gertjan van Noord' Finite State Utilities


(c) Nathan Vaillette (nvail@ling.ohio-state.edu)

Last updated July 10, 01

Tested with the Finite State Utilities v. 6.234


               *             *               *
FIXED:

- msol_macro expansion properly names variables apart

- n-ary "and" and "or" (no longer balk when n > 2 and even ;)

- vacuous quantification now works right


ADDED

- "lambda abstracts"

- some more defined formulas

               *             *               *
	       
This program uses the Finite State Utilities (see
http://odur.let.rug.nl/~vannoord/Fsa) to build FSAs from expressions
of a monadic second order string logic (a variant of MSOL(Str)).  From
formulas of the language, multi-track automata are built with the
first track for alphabetic symbols and the other tracks encoding
boolean strings corresponding to the variables of the formula.  An
n-track machine is simulated on a 1-track machine by treating the
alphabet of the 1-track machine as ? x ... x ? (n times); a tuple
symbol for the n-track machine is thus represented sequentially on the
1-track one.

Variables are interpreted as ranging over sets of positions in a
string.  Free variables have existential import.  The semantics of the
atomic formulas are explained in the comments below.

The syntax of the logic is:

    Sym        <-  atomic Prolog term

    Var        ::= v(Sym) 

    Formula    ::= AtomForm | MolForm

    AtomForm   ::= symbol(Var,Sym) | singleton(Var) |
                   subset(Var,Var) | precedes(Var,Var)

    MolForm    ::= not(Form) | and(Form,Form) | or(Form,Form) | if(Form,Form) |
                   exists(Var,Form) | all(Var,Form)

    Expression ::= msol(Formula) | msol2(Formula)

msol2(Formula) compiles the formula into an automaton; msol(Formula)
additionally eliminates the variable tracks.  Only *expressions* of
the logic should be used as part of regular expressions.  (*Formulas*
are in fact also regular expression macros in this implementation, but
this is subject to change and should not be considered part of the
"public interface.")

Macros for logic formulas ("formula" in the sense given in the syntax)
may be declared with msol_macro/2.  Not that you should not use the
Utilities' macro/2 for this purpose ("macro(msol2(E0,E))" will produce
the same machine as "msol_macro(E0,E)", but only the latter can be
safely embedded in a larger msol expression as a subformula).  See the
"defined formulas" and "regular expression matching" sections below
for some extension to the syntax.

In addition to these syntactic devices, there is an extra-logical
device of "lambda abstracts," which are written as
lambda(LambdaParam,Term).  The "lambda parameter" LambdaParam is a
Prolog atom wrapped in l(...).  Term can be any syntactic element of
the logic, e.g. variables, symbols, regex's, formulas, as well as
lambda abstracts themselves.  Abstracts are "applied" with
beta_reduce(LambdaAbstract,Anything).

Lambda abstracts do not actually provide higher-order terms.  The
notation is merely suggestive; they perform a purely syntactic
replacement of one term by another in a third.  Replacement points are
marked by apply/2 makes the substitutions.
E.g. beta_reduce(lambda(l(p),foo(l(p))), v(x)) expands to foo(v(x)).
It is perfectly possible to use lambda abstracts to write "formulas"
whose lambda-expansions are not well-formed.

Since the replacement is purely syntactic, some points about the
syntax have to be mentioned.

- both the 2nd arg. of the abstract and the 2nd arg of apply undergo
  msol_macro expansion.  Note that this expansion names apart
  variables, but not lambda parameters.

- the replacement avoids accidental capturing of both variables, e.g.

  beta_reduce(lambda(l(l),exists(v(x),foo(v(x),l(l)))),v(x)) ==>

  exists(v(fresh_var(2)),foo(v(fresh_var(2)),v(x))) 

               *             *               *

TO DO:

- replace macros for formulas with a special recursive
  formula-to-regex compilation predicate for more efficient treatment
  of free variables

- caching subformulas
  
- (maybe, someday) Prolog implementation of BDDs a la MONA 
  (see http://www.brics.dk/mona) that can be called by the
  Utilities

----------------------------------------------------------------------*/

:- [expansion].				    
:- multifile macro/2.
:- multifile msol_macro/2.

/*----------------------------------------------------------------------

The double colon is used in the construction of multi-track machines
and interpreted as tuple formation.  For instance, a::b::c::d can be
though of as the singleton language over the alphabet (? x ? x ? x ?)
containing just the 4-tuple <a,b,c,d>.  However, since multi-track
machines are simulated on 1-track machines, "::" is literally just
concatenation.

----------------------------------------------------------------------*/

:- op(500,xfy,::).
macro(X::Y,[X,Y]).

macro(times(L,N),E) :- times(L,N,E).
times(_L,0,[]) :- !.
times(L,N,E) :-
	M is N-1,
	times(L,M,E1),
	E = L::E1.


macro(bool, {true,false}).

%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%   wrapper macro       %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%

macro(msol(F0),minimize(F)) :-
	expand_msol_macros(F0,F1),
	write(F1),nl,nl,
	free_vars(F1,_Vars,N),
	F = mt_domain(F1,N).

%% the first projection of the (N+1)-ary relation R:
macro(mt_domain(R,N), E) :-
	E = range(R o [id(?), times(bool:[],N)]*).

%% msol2: wrapper macro that does msol_macro
%% expansion, but leaves the free variable tracks on

macro(msol2(F0),F) :-
	expand_msol_macros(F0,F).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%   basic atomic formulas:  %%               
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

				    
%% symbol(v(X),S) means "every position in v(X) carries the symbol S".

macro(symbol(v(_),S),E) :-
    E = {
	? :: false,
	S :: true 
	}* .


%% singleton(v(X)) means "v(X) is a singleton set of positions"

macro(singleton(v(_)),E) :-
	E = [
	     (	 ? :: false )*,
	     (	 ? :: true  ),
	     (	 ? :: false )*
	    ].


%% subset(v(X),v(Y)) means "v(X) is a subset of v(Y)".

macro(subset(v(X),v(Y)), E) :-
	(
	    X == Y
	
	->  
	    
	    E = ( ? :: bool :: bool )*
	
	;   
	    
	    X @< Y
	
	->  
	    
	    E = {
		 ? :: false :: false,
		 ? :: false :: true ,
		 ? :: true  :: true 
		}*
	
	;   
	    
	    E = {
		 ? :: false :: false,
		 ? :: true  :: false,
		 ? :: true  :: true 
		}*
	).


%% precedes(v(X),v(Y)) means "the contiguous set of positions v(X) directly
%% precedes the contiguous set of positions v(Y)"

macro(precedes(v(X),v(Y)), E) :-
	(
	    X == Y
	->
	    E = (? :: false :: false)*
	;
	    X @< Y
	->
	    E = [
		 (   ? :: false :: false  )* ,
		 (   ? :: true  :: false  )* ,	   	
		 (   ? :: false :: true   )* ,	   	
		 (   ? :: false :: false  )*
		 ]

	;
	    E = [
		 (   ? :: false :: false  )* ,
		 (   ? :: false :: true   )* ,	   	
		 (   ? :: true  :: false  )* ,	   	
		 (   ? :: false :: false  )*
		 ]

	).


%%%%%%%%%%%%%%%%%%%%%%%%%%
%% non-atomic formulas  %%
%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Negation %%

macro(not(F), E) :-
	free_vars(F,_Vars,N),
	E = [?,times(bool,N)]* - F.

%% Existential quantification %%

%% problem with vacuous quantification!!

macro(exists(v(X),F),E) :-
	free_vars(F,Vars,N), %% returns sorted list
	                     %% of free variables
	(   nth(T,Vars,X)
	->  E = remove_track(F,T,N)
	;   %% v(X) does not occur free in F, so:
	    E = F
	).

%% remove the Tth variable track of the N variable tracks of L:

macro(remove_track(L,T,N),minimize(efree(E))) :-
	X is T-1,
	Y is N-T,
	E = range(L o [id(?), times(bool,X),bool x [],times(bool,Y)]*).

%% Conjunction %%

macro(and(F1,F2), E) :-
	!,
	free_vars(F1,Vars1,_),
	free_vars(F2,Vars2,_),

	redundant_tracks(Vars1,Vars2,Insertions1,Insertions2),

	E =
	efree(range(F1 o ([id(?)|Insertions1]* )))
	&
	efree(range(F2 o ([id(?)|Insertions2]* ))).
	
%% redundant_tracks/4
%% example:
%% | ?- redundant_tracks([x,z],[y],I1,I2).
%% I1 = [bool     , [] x bool, bool     ],
%% I2 = [[] x bool, bool     , [] x bool]


before(X,Y) :- X @< Y.
macro(insert(I),[] x I).

redundant_tracks([],[],[],[]).

redundant_tracks([],[_|T],[insert(bool)|FPs1],[id(bool)|FPs2]) :-
        redundant_tracks([],T,FPs1,FPs2).

redundant_tracks([_|T],[],[id(bool)|FPs1],[insert(bool)|FPs2]) :-
        redundant_tracks(T,[],FPs1,FPs2).

redundant_tracks([H|T1],[H|T2],[id(bool)|FPs1],[id(bool)|FPs2]) :-
        !,
        redundant_tracks(T1,T2,FPs1,FPs2).

redundant_tracks([H1|T1],[H2|T2],[id(bool)|FPs1],[insert(bool)|FPs2]) :-
        before(H1,H2),
        !,
        redundant_tracks(T1,[H2|T2],FPs1,FPs2).

redundant_tracks([H1|T1],[_H2|T2],[insert(bool)|FPs1],[id(bool)|FPs2]) :-
        %% before(H2,H1),
        redundant_tracks([H1|T1], T2, FPs1, FPs2). 

%% "and" with arbitrary arity > 1:

msol_macro(and(A),A).
msol_macro(E,and(and(A,B),Conj)) :-
	E =.. [and,A,B,C|Conjs],
	Conj  =.. [and,C|Conjs].




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% defined (definable) formulas:   %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

/*---------------------------------------------------------------------------

Many useful relations &c. could be defined in the logic with
msol_macro/2; however, some of them can be handled more efficiently by
directly specifying the regular relation they compile to.

The selection of defined formulas is a little sporadic and disorganized
at the moment.

---------------------------------------------------------------------------*/

%% Material implication %%

msol_macro(if(A,B),
	   not(
	       and(
		   A,
		   not(B)
		  )
	      )
	  ).

%% Bi-implication %%

msol_macro(iff(A,B),
	   and(if(A,B),if(B,A))).


%% Univeral quantification %%

msol_macro(all(v(X),F),
	   not(
	       exists( v(X),
		       not(F)
		     )
	      )
	  ).


%% Disjununction %%

% msol_macro(or(A,B),
% 	   not(
% 	       and(
% 		   not(A),
% 		   not(B)
% 		  )
% 	      )
% 	  ).

%% defining "or" directly in terms of union avoids the negations and
%% conjunction:

macro(or(F1,F2), E) :-
	free_vars(F1,Vars1,_N1),
	free_vars(F2,Vars2,_N2),

	redundant_tracks(Vars1,Vars2,Insertions1,Insertions2),

	E =
	{
	 efree(range(F1 o ([id(?)|Insertions1] *))),
	 efree(range(F2 o ([id(?)|Insertions2] *)))
	}.


%% "or" with arbitrary arity > 1:

msol_macro(or(A),A).
msol_macro(E,or(or(A,B),Disj)) :-
	E =.. [or,A,B,C|Disjs],
	Disj =.. [or,C|Disjs].


		  

%% Exclusive disjunction %%

msol_macro(xor(A,B),
	   and(
	       or(A,B),
	       not(
		   and(A,B)
		  )
	      )
	  ).


%% Universal quantification with restriction

%% this is more efficient than all(v(X),if(If,Then)) since
%% it avoids the double negation in the expansion

msol_macro(all_if(v(X),If,Then),
	   not(
	       exists(v(X),
		      and(
			  If,
			  not(Then)
			 )
		     )
	      )
	  ).


% msol_macro(tautology,
% 	   exists(v(x),
% 		  subset(v(x),v(x)))
%  	  ).

macro(tautology, ? *).

% msol_macro(contradiction,
% 	   exists(v(x),
% 		  and(
% 		      not(empty(v(x))),
% 		      precedes(v(x),v(x))))).
	   

macro(contradiction, {} ).

%% v(_) is a string:

% msol_macro(string(v(X)),
% 	   exists(v(y),
% 		  precedes(v(X),v(y)) %% v(y) may be empty
% 		 )
% 	  ).

macro(string(v(_)),E) :-
	E = [
	     (	 ? :: false )*,
	     (	 ? :: true  )*,
	     (	 ? :: false )*
	    ].

%% v(X) is empty:

% msol_macro(empty(v(X)),
% 	   not(
% 	       exists(v(y),
% 		      and(
% 			  singleton(v(y)),
% 			  subset(v(y),v(X))
% 			 )
% 		     )
% 	      )
% 	  ).

		     
macro(empty(v(X)),E) :-
	E = (? :: false )*
	.

%% the definition in the logic of the following relations is left
%% as an exercise to the user ;)

%% v(X) is a prefix of v(X)

macro(prefix(v(X),v(Y)), E) :-
	(
	    X == Y
	->
	    E = ( ? :: bool :: bool )*

	;
	    X @< Y
	->
	    E = [
		 (   ? :: false :: false  )* ,
		 (   ? :: true  :: true  )* ,	   	
		 (   ? :: false :: true   )* ,	   	
		 (   ? :: false :: false  )*
		 ]

	;
	    E = [
		 (   ? :: false :: false  )* ,
		 (   ? :: true  :: true   )* ,	   	
		 (   ? :: true  :: false  )* ,	   	
		 (   ? :: false :: false  )*
		 ]

	).

%% v(X) is a proper prefix of v(X)

macro(proper_prefix(v(X),v(Y)), E) :-
	(
	    X == Y
	->
	    E = ( ? :: bool :: bool )*

	;
	    X @< Y
	->
	    E = [
		 (   ? :: false :: false  )* ,
		 (   ? :: true  :: true   )* ,	   	
		 (   ? :: false :: true   )+ ,	   	
		 (   ? :: false :: false  )*
		 ]

	;
	    E = [
		 (   ? :: false :: false  )* ,
		 (   ? :: true  :: true   )* ,	   	
		 (   ? :: true  :: false  )+ ,	   	
		 (   ? :: false :: false  )*
		 ]

	).



%%% 7.10.01: fixed quantifiers  and added some new defined formulas:

%% 
%% allows exists(v(x),Form1,...,FormN) with Forms interpreted as conjoined
msol_macro(F,exists(v(V),Conj)) :-
	F =.. [exists,v(V),Arg1,Arg2|Args],
	!,
	Conj =.. [and,Arg1,Arg2|Args].



%% exists_elt(v(Sing),Set,Form1,...,Formn): exists an element X of Set
%% (really singleton subset)
msol_macro(F,E) :-
	F =.. [exists_elt,v(Sing),v(Set)|Args],
	!,
	E =.. [exists,v(Sing),singleton(v(Sing)),subset(v(Sing),v(Set))|Args].


%% every singleton element of a set (really singleton subset)
msol_macro(all_elt(v(Sing),v(Set),Arg),
	   all_if(v(Sing),
		  and(
		      singleton(v(Sing)),
		      subset(v(Sing),v(Set))
		     ),
		  %% then
		  Arg)
	  ).

%% denotes the empty string
msol_macro(empty_string,
	   all(v(x),empty(v(x)))
	  ).

%% true if R contains the empty string
msol_macro(contains_e(regex(R)),
	   exists(v(x),
		  empty(v(x)),
		  matches(v(x),regex(R))
		 )
	  ).

%% F is an initial string
msol_macro(first(v(F)),
	   exists(v(whole),
		  whole_string(v(whole)),
		  first(v(F),v(whole))
		 )
	  ).
		       

%% Whole is the whole string
msol_macro(whole_string(v(Whole)),
	   all(v(x),
	       subset(v(x),v(Whole))
	      )
	  ).

%% F is initial in Set
%% presupposes Set is a string 
msol_macro(first(v(F),v(Set)),
	   and(
	       subset(v(F),v(Set)),
	       all_elt(v(other),v(Set),
		       not(precedes(v(other),v(F)))
		      )
	      )
	  ).

	      
msol_macro(equals(v(X),v(Y)),
	   and(
	       subset(v(X),v(Y)),
	       subset(v(Y),v(X))
	      )
	  ).

%% string X occurs before string Y, with any distance between them
msol_macro(before(v(X),v(Y)),
	   or(
	      precedes(v(X),v(Y)),
	      exists(v(z),
		     not(empty(v(z))),  %% the empty set can be anywhere
		     precedes(v(X),v(z)),
		     precedes(v(z),v(Y))
		    )
	     )
	  ).

%% set comprehension with lambda abstracts:
msol_macro(setofall(v(Set),LambdaAbs),

	   all_if(v(x),
		  singleton(v(x)),
		  
		  iff(
		      subset(v(x),v(Set)),
		      beta_reduce(LambdaAbs,v(x))
		      )
		  )
	   ).




%% I don't think the following 3 predicates can be defined in the logic,
%% though they can be for any fixed integers of course.  They're useful
%% mostly for debugging.

%% the nth position in a string:
macro(nth_pos(v(_),N),

      [times(? :: false, M),
       ? :: true,
       (? :: false)*
      ]) :-
	M is N -1.

%% the substring of length Length starting at Start
macro(region(v(_),Start,Length),
      [times(? :: false, M),
       times(? :: true,Length),
       (   ? :: false)*
      ]) :-
	M is Start -1.

%% the substring starting at Start and extending to Finish
macro(region2(v(_),Start,Finish),
      [times(? :: false, M),
       times(? :: true,Length),
       (   ? :: false)*
      ]) :-
	M is Start -1,
	Length is +(1,(Finish-Start)).




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%   regular expression matching  %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


/*----------------------------------------------------------------------

The predicate "matches" allows reference to arbitrary regular
expressions within the logic.  Where Regex is a regular expression,
matches(v(X),regex(Regex)) means "v(X) is a substring and is a member
of the language denoted by Regex."  (The "regex(...)" wrapper is
necessary to prevent the collection of free variables and the
expansion of msol_macros from venturing down into the Regex argument.)

----------------------------------------------------------------------*/

                  
macro(matches(v(X),regex(L)),E) :-

	%% add extra "track" to L:

        L1 = range(L o [id(?),[] x true ]*),
	
	E = [ ( ? :: false )*, L1, ( ? :: false )* ]
	.

%% this lets you write matches without the regex wrapper, 
%% though it's not safe:

macro(matches(v(X),L),matches(v(X),regex(L))).


%% FIN

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



