/*----------------------------------------------------------------------
msol_rep.pl

http://www.ling.ohio-state.edu/~nvail/msol.html

Demonstration on a small, closed alphabet of the logical definition of
the conditional replacement operator

(c) Nathan Vaillette (nvail@ling.ohio-state.edu)

Last updated July 10, 01

Tested with the Finite State Utilities v. 6.234

               *             *               *

This file contains macros definining the conditional replacement
operator in monadic second order logic, in particular
replace_uo(Reln,Left,Right) and replace_ro(Reln,Left,Right).  These
undirectedly make the replacment specifed by Reln in the context
between Left and Right, and are *u*pward and *r*ightward oriented,
resp. (uo checks both contexts on the input side, ro checks left
context on the ouput side and right context on the input side.)
For more explanation of the implementation, see
http://www.ling.ohio-state.edu/~nvail/ftp/helsinki-abs.ps.gz

In this simple version, these operators restricted to a small, closed
alphabet, namely {a,b,c,d,e} plus the "blank" symbol o, and
internally, transducers are simulated as recognizers using caslr.pl
(*c*losed-*a*lphabet *s*ame-*l*ength *r*elations): e.g. the
input/output pair a:e is just represented as the symbol ae.  The Reln
argument of replace_uo/_ro however is specified as a real transducer
and is converted to a caslr automatically.

For example, try helsinki_ex, which is 

replace_ro(({[a,b], [b,a]} x c)+,b,{a,b})

i.e. it replaces any string of any number > 1 of occurances of "a b"
or "b a" with the same number of cs when the string occurs after an
output b and before an input a or b.  Apply it to "bbababb" for a nice
test.

One restriction: the Reln to be replaced should not contain the empty
string.  cond_insert_uo(Insertion, Left,Right) and cond_insert_ro are
defined separately for this purpose.


----------------------------------------------------------------------*/

:- multifile macro/2 .
:- multifile rx/2 .
:- multifile msol_macro/2 .

:- [msol].
:- write(hi),nl.
:- [caslr].


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



/*

We can define a conditional replacement operator (in the caslr
simulation) for same-length relations.  We will also see that it can
be extendended pretty straightforwardly to a replacement operator for
arbitrary relations (i.e. where input and output can be different
lengths).

Replacements will be specified by an arbitrary regular relation as in
Gerdemann and van Noord (1999), i.e. instead of input and output
arguments, the replace operation will have a single transducer
argument, allowing "backreferencing."

The operator we define first will do non-deterministic replacement of
all possibilities, but we will show that the definition can be
converted into a definition of a directed, left-to-right (or
right-to-left) longest (or shortest) match replacement operator with
virtually no extra effort.

We also want to allow various possibilities for conditioning context,
in the spirit of Karttunen's (1995) up-/down-/left-/rightward-oriented
replacement.  To do this, we will make the conditioning contexts just
arbitrary caslr relations.  If you want to condition the replacement
of the preceding input (i.e. with leftward orientation), say [a,b],
then you just make the left context [a,b] xx ? * (in the caslr
simulation, string_times_caslr([a,b])).  Most of the examples I give
will be of upward-oriented replacement (both conditions checked on the
input side).

 Some preliminaries needed for the replace operator:

*/

%% v(x) matches the input side of Reln
msol_macro(matches_input(v(X),Reln),
	   matches(v(X),non_det_range(Reln))
	  ).

msol_macro(matches_output(v(X),Reln),
	   matches(v(X),non_det_domain(Reln))
	  ).

%% v(X) is a replacement candidate, i.e.
%% v(X) is the right kind of thing in the right context

%% msol_macro(replacement_candidate(v(X),Relation,Left,Right),

msol_macro(to_replace(v(X),Relation,Left,Right),

	   and(
	       matches_input(v(X),Relation),
	       between(v(X),Left,Right)
	      )
	  ).


%% v(X) is directly preceded by Left and followed by Right:
msol_macro(between(v(X),Left,Right),
	   exists(v(a),
		  exists(v(b),
			 and(
			     matches(v(a),Left),
			     precedes(v(a),v(X)),
			     matches(v(b),Right),
			     precedes(v(X),v(b))
			    )
			)
		 )
	  ).




%% v(X) is to be preserved in the ouput, because it doesn't have anything
%% in it that is supposed to be replaced
msol_macro(to_preserve(v(X),T,Left,Right),
	   not(exists(v(y),
		      and(
	       
			  subset(v(y),v(X)),
			  
			  not(empty(v(y))),				     
			  
			  %% replacement_candidate(v(y),T,Left,Right)
			  to_replace(v(y),T,Left,Right)
			 )
		     )
	      )
	  ).





/*

The correct procedural intuition is that general replacement should go
through an input string, nondeterministically making allowed
replacements until no more replacement candidates remain.  In other
words, there is some parse of the input string partioning it into
allowable replacements and preserved regions.  Karttunen (1995)
gives a declarative statement of this notion for the *un*conditional
replacement operator by requiring that each string in the replacment
relation must be an alteration of allowed replacements with preserved
substrings, where the preserved substrings don't contain any
replacement candidates.

In general, we can write a regular expression for the alternation of A
and B as

[ A^ [B,A]* B^ ]

Karttunen's unconditional replacement operator uses essentially this
template.  Note that the non-determinism ("there is *some* parse of
the input string") just follows from the nondeterministic matching of
such regular expressions.

Defining unconditional replacement will prove a little harder in the
logic than with just regular relations.  However, the resulting
definition can be extended almost trivially to conditional replacment,
while Karttunen's regular expression has many more steps to go through
to become conditional.

First, for illustration, the basic alternation macro:
*/

%% alternating0(A,B) is the same as [ A^ [B,A]* B^ ],
%% that is the language of strings of alternating As and Bs (with no
%% restrictions on which comes first or last).

macro(alternating0(A,B),
      
      msol(
	   all_if(v(x),
		  singleton(v(x)),
		  or(
		     in_region0(v(x),regex(A),regex(B)),
		     in_region0(v(x),regex(B),regex(A))
		    )
		 )
		 )
	  ).


%% X is part of a string of As which is either a final string,
%% or is followed by a string of Bs

msol_macro(in_region0(v(X),A,B),
	   exists(v(s),
		  %%string(v(s)),
		  subset(v(X),v(s)),
		  matches(v(s),A),
		  or(
		     exists(v(s2),
			    matches(v(s2),B),
			    precedes(v(s),v(s2))),
		     final(v(s))
		    )
		 )
	  ).



%% (final(v(X)) presupposes X is a string)
msol_macro(final(v(X)),
	   not(exists(v(s),
		      not(empty(v(s))),
		      precedes(v(X),v(s))))).
	  

%% (initial(v(X)) presupposes X is a string)
msol_macro(initial(v(X)),
	   not(exists(v(s),
		      not(empty(v(s))),
		      precedes(v(s),v(X))))).


/*

  Now let's generalize alternating0 a little using the lambda-abstract
mechanism to paramaterize it.  In this version, A and B are lambda
abstracts, so e.g. alternating0(A,B) could be redefined as

alternating(
	    lambda(l(x), matches(l(x),A)),
	    lambda(l(y), matches(l(y),A))
	   )

*/



%% version with lambda-abstract arguments
macro(alternating(A,B),
      msol(
	   all_if(v(x),
		  singleton(v(x)),
		  or(
		     in_region(v(x),A,B),
		     in_region(v(x),B,A)
		    )
		 )
		 )
	  ).


msol_macro(in_region(v(X),A,B),
	   exists(v(s),
		  %%string(v(s)),
		  subset(v(X),v(s)),
		  beta_reduce(A,v(s)),
		  or(
		     exists(v(s2),
			    beta_reduce(B,v(s2)),
			    precedes(v(s),v(s2))),
		     final(v(s))
		    )
		 )
	  ).



/* now unconditional replacement a la Karttunen is easy: */


macro(uncond_replace(Reln),
      alternating(
		  lambda( l(x), and(
				    to_replace(l(x),Reln,[],[]),
				    %replacement_candidate(l(x),Reln,[],[]),
				                                    %% no context
				    matches(l(x),Reln)
				   )
			),

		  lambda( l(y), and(
				    to_preserve(l(y),Reln,[],[]), %% no context
				    matches(l(y),caslr_id *)
				   )
			)
		 )
     ).
		  

/*

Now, what Karttunen does to get conditional replacement is start with
the unconditional replacement relation, then devise a method to flag
contexts in a string and use unconditional replacement for the flagged
contexts.  This is where the cascaded composition with all the
brackets comes in, which is hard to interpret declaritively.  But we
can turn our unconditional replacement into a conditional one
trivially by just reintroducing the context into the specifications of
what is to be preserved and what is to be replaced:

*/

macro(cond_replace(Reln,Left,Right),
      alternating(
		  lambda( l(x), and(
				    to_replace(l(x),Reln,Left,Right),
				                                    %% context
				    matches(l(x),Reln)
				   )
			),

		  lambda( l(y), and(
				    to_preserve(l(y),Reln,Left,Right), %% context
				    matches(l(y),caslr_id *)
				   )
			)
		 )
     ).

macro(replace1(Reln,Left,Right),
      caslr_to_trans(cond_replace(Reln,Left,Right))).




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

msol_macro(matches_weak(v(X),L),
      matches(v(X), range(L o {?, [] x %%oo}*))).
			               input_o}*))).

macro(remove_input_os(L), range(L o {(alpha - input_o), input_o x []}*)).

%% X's domain is the same as the domain of a string from Relation
msol_macro(matches_input_weak(v(X),Relation),
	   matches_weak(v(X),non_det_range(
					   remove_input_os(
					   Relation
					 )
		       ))
	  ).


msol_macro(matches_input_weak2(v(X),Relation),
	   matches_weak(v(X),non_det_range(
				%	   remove_input_os(
					   Relation
					% )
		       ))
	  ).


%% v(X) is a replacement candidate, i.e.
%% v(X) is the right kind of thing in the right context



msol_macro(to_replace_weak(v(X),Relation,Left,Right),

	   and(
	       matches_input_weak(v(X),Relation),
	       between_weak(v(X),Left,Right)
	      )
	  ).

msol_macro(to_replace_weak2(v(X),Relation,Left,Right),

	   and(
	       matches_input_weak2(v(X),Relation),
	       between_weak(v(X),Left,Right)
	      )
	  ).


%% v(X) is directly preceded by Left and followed by Right:
msol_macro(between_weak(v(X),Left,Right),
	   exists(v(a),
		  exists(v(b),
			 and(
			     matches_weak(v(a),Left),
			     precedes(v(a),v(X)),
			     matches_weak(v(b),Right),
			     precedes(v(X),v(b))
			    )
			)
		 )
	  ).




%% v(X) is to be preserved in the ouput, because it doesn't have anything
%% in it that is supposed to be replaced
msol_macro(to_preserve_weak(v(X),T,Left,Right),
	   not(exists(v(y),
		      and(
	       
			  subset(v(y),v(X)),
			  
			  %% not(empty(v(y))),				     
			  
			  
			  to_replace_weak(v(y),T,Left,Right)

			 )
		     )
	      )
	  ).



macro(cond_replace_weak(Reln,Left,Right),
      alternating(
		  lambda( l(x), and(

				    %% seems to be nec. to use weak 2 here!
				    to_replace_weak(l(x),Reln,Left,Right),
				                                    %% context
				    matches(l(x),Reln)
				   )
			),

		  lambda( l(y), and(
				    to_preserve_weak(l(y),Reln,Left,Right), %% context
				    matches(l(y),caslr_id *)
				   )
			)
		 )
     ).

macro(replace2(Reln,Left,Right),
      {?, [] x o}*
      o
      caslr_to_trans(cond_replace_weak(Reln,Left,Right))
      o
      {? - o, o x []}*
).



%% upward-oriented undirectional conditional replacement.
%% Left and Right are normal languages; Reln must be specified as
%% a caslr

macro(replace_uo(Reln,Left,Right),

     %       [([] x o)+, [?,([] x o)+]* ]
     
      {? - o , [] x o}*
     o
     caslr_to_trans(cond_replace_weak(Caslr,LeftInput,RightInput))
     o
     {? - o, o x []}*
     ) :-
	Caslr = trans_to_caslr(Reln),
	LeftInput = string_times_caslr(Left),
	RightInput = string_times_caslr(Right).

macro(replace_ro(Reln,Left,Right),
      %[([] x o)+, [?,([] x o)+]* ]

      {? - o, [] x o}*
     o
     caslr_to_trans(cond_replace_weak(Caslr,LeftOutput,RightInput))
     o
     {? - o, o x []}*
     ) :-
	Caslr = trans_to_caslr(Reln),
	LeftOutput = caslr_times_string(Left),
	RightInput = string_times_caslr(Right).

%% ({[a,b], [b,a]} x c)+ / b ___ {a,b}
macro(helsinki_ex,
%      replace_lu({[ac,bo],[bc,ao]}+,b,{a,b})
      replace_ro(({[a,b], [b,a]} x c)+,b,{a,b})
      ).


macro(cond_insert_uo(Insert,Left,Right),

      [([] x o)+, [?,([] x o)+]* ]

     o
      cond_replace(Reln,Left,Right)
     o
     {? - o, o x []}*) :-
	
	Reln = caslr_times_string(Insert) & (input_o *),
	LeftInput = string_times_caslr(Left),
	RightInput = string_times_caslr(Right).


macro(cond_insert_ro(Insert,Left,Right),

      [([] x o)+, [?,([] x o)+]* ]

     o
      cond_replace(Reln,LeftOutput,RightInput)
     o
     {? - o, o x []}*) :-
	
	Reln = caslr_times_string(Insert) & (input_o *),
	LeftOutput= caslr_times_string(Left),
	RightInput = string_times_caslr(Right).

macro(cond_insertion(Insert,Left,Right),
      msol(
	   and(
	       all_if(v(x),
		      to_insert(v(x),Left,Right),
		      %% then
		      matches(v(x),caslr_times_string(Insert) & (input_o *))
		     ),
	      %% and
 	       all_if(v(x),
		      not_to_insert(v(x),Left,Right),

		      %% then
		      matches(v(x),caslr_id *)
 		     )
	       
% 	       and(not(exists(v(x),
% 			      matches(v(x),[Left,
% 					    ~(caslr_times_string(Insert) & (input_o *)),
% 					   Right])
% 			     ))
% 		  )
		      
		       
	     )
	   
	  )
     ).
				
msol_macro(to_insert(v(X),Left,Right),
	   and(matches(v(X),input_o *),
	       not(empty(v(X))),
	       between(v(X),Left,Right))).

msol_macro(not_to_insert(v(X),Left,Right),
	   and(
	       singleton(v(X)),
	       not(exists(v(y),
			  subset(v(X),v(y)),
			  to_insert(v(y),Left,Right))))).


			    


