%%           -*-Mode: prolog;-*-


user:re:- nl,fsa_regex:fsa_load_aux_file('chap3.pl').

:- multifile macro/2.
:- discontiguous macro/2.

:- ensure_loaded('combined.pl').
:- ensure_loaded('replace_nm.pl').
%% This file implements chapter 3 of Hammond, "The Phonology of
%% English". See below macro: chap3.



%% General Utilities:


% intro3, ignore 3: These macros are as in Kaplan & Kay, except that 
% they are modified to treat sequences of three symbols as if they
% were singletons. This is done so that a sequence of [d,t_,co] can
% be interpreted as a unit representing an underlying d, corresponding
% to a surface t in a coda. The underscore in 't_' is there to ensure
% that surface segments can be distinguished underlying ones. Surface
% segments could also perhaps be distinguished by position, since they
% are second in each triple. But distinguishing by position makes a
% number makes a number of operations more difficult. 

macro(intro3(Triple),[[[?,?,?]*,[]:Triple]*,[?,?,?]*]).
macro(ignore3(Expr,Triple),range(Expr o intro3(Triple))).




%% Alphabet symbol types:

% output:

% The optimality operator needs to know what symbols are added by
% gen. Specifically, the step 'add_violation' strips output symbols,
% then adds violation marks, and then reinserts the output symbols in
% arbitrary positions. For example, suppose that the input is 'abc'
% and the output symbol are {x,y}. Suppose the candidates are axbcy
% and abxyc. If constraint C marks these with:  axb*cy* and abx*yc,
% then stripping the output symbols yields: ab*c* and ab*c. Adding
% arbitrary positive numbers of markers to these yields an infinite
% set of alternatives: {a*b*c*,*a*b*c*,a*b*c,ab*c*,...}. Then randomly 
% reinserting output symbols, yields: {xxa*yxb*c*,*ax*b*c*,*ax*b*yc*,
% ax*b*cy,ayxb*c*,axb*cy*,...}. And then since the marked candidate
% axb*cy* is in this infinite set of alternatives, it is
% eliminated. The macros in matching.pl also include permutation to
% move marks into different positions in constructing the set of
% alternatives. 
%
% For this grammar, the output symbols are 1) the tags for marking
% syllable constituents as onset coda, etc, 2) the surface symbols
% defined in the macro for segement_, and 3) the special symbol ins,
% for insertions. The corresponding 'del' symbols is treated as a
% dummy syllable constituent tag. Using this dummy tag helps maintain 
% the pattern of treating alphabet triples as units. A deleted /t/,
% for example, is represented by the triple [t,del,del]. This will be
% distinguished from [t,xp,del], which is a /t/ deleted in the sense
% of an M-Parse violation.
macro(output, {tag,segment_,ins}). 


% Markers for onset, coda, unparsed and nucleus. 
macro(tag, {on,co,xp,nu,del}).

macro(segment, {vowel, consonant}).
macro(segment_, {vowel_, consonant_}).

macro(onset,[{segment,ins},segment_,on]).
macro(nucleus,[{segment,ins},segment_,nu]).
macro(coda,[{segment,ins},segment_,co]).
macro(unparsed,[{segment,ins},segment_,xp]).
macro(deleted,[segment,{del,xp},del]).





% Chapter 3 only deals with consonants, so this version of gen is
% designed not to do anything with vowels.
macro(gen,
  {[consonant, []:consonant_, []:{on,co,xp}],
   [vowel_io,[]:nu],
   [segment,[]:[del,del]],
   [[]:[ins,consonant_,{on,co,nu,xp}]]   % insertion (input is epsilon)
  }*
             o
   % The following allows segments to be deleted in two ways. Normal
   % deletions are marked with [del,del]. Deletions in the sense of
   % the "null parse" are marked as [xp,del]. If one segment is marked 
   % as [xp,del], then every segment must be so marked. 

  { ? *, [?, del:xp, del]*}   
             o
  % The following ensures that gen doesn't produce nonsense. Words,
  % can't, for example, begin with a coda or end with an onset.
  % Possibly this doesn't belong in gen; it may eventually be moved
  % into the constraints.
  ignore3([onset*,nucleus+,coda*]*,{deleted,unparsed})
  ).



macro(chap3,
    gen
    mc
   'Parse-POA-Sonorant'
    mc
   'Pre-Parse'
    mc
   'Max-IO'
    mc
   'Dep-IO'
    mc
   'Faith(MOA)' 
    mc
   'Faith(Voicing)' 
    mc
   'Faith(COR)' 
    mc
   'Faith(DOR)'
    mc
   'CyConstraint'
    mc 
    h_proscription
    mc
   'VoicingWithFricatives'
    mc                  
   'FricativeObstruent'
    mc
   'NoGeminates'   % Works okay, but commented out for efficiency
    mc
   'Onset/{N,Z}'
    mc
   'AffricateProscription'
    mc
   '*[sr]' 
    mc
   'LateralConstraint'
    mc
   'Nonhomorganic_NC_Codas'
    mc
   'NC_Codas_With_Voiced_Obstruent'
    mc
   'MaximalCoda'
    mc
   '*o/ao' mc '*o/an' mc '*o/no' mc '*o/na'
    mc
    '*o/nn' mc '*o/aa'
    mc
   '*o/stop-o' mc '*o/[f]-o' mc '*o/[T]-o' mc '*o/[S]-o' mc '*o/voiced-o' 
    mc
   '*o/stop-N' mc '*o/[f]-N' mc '*o/[T]-N' mc '*o/[S]-N' mc '*o/voiced-N' 
    mc
   '*c/stop-o' mc '*c/[f]-o' mc '*c/[T]-o' mc '*c/[S]-o' mc '*c/voiced_obst-o' 
    mc
   '*o/on'  % may come after M-Parse for some dialects   
    mc
   '*c/or'
    mc
   '*c/ol' mc '*c/nr'
    mc
   '*c/on' mc '*c/nl' mc '*c/lr'
    mc
   'Sc'
    lc
   'M-Parse'
    oo
   'Faith(POA)'
    oo
   '*o/oa'
    oo
%% The order of simple onset constraints and of simple coda
%% constraints are fixed by the sonority hierarchy. But in 
%% principle, these constraints could be shuffled in various ways.
  '*o/a' oo '*o/n' oo '*o/o'   % intrinsic onset constraints
    oo
  '*c/o' oo '*c/n' oo '*c/l' oo '*c/r' % intrinsic coda constraints
    oo
  '*c/no' oo '*c/ln' oo '*c/rl'
    oo
  '*c/lo' oo '*c/rn'
    oo
  '*c/ro'
    oo
  '*o/[s]-o'
    oo
  '*o/[s]-N'
    oo
  '*c/[s]-o'

).



% Parse: pp. 44, 51 : changed to Parse-POA & Parse-Sonorant on p. 98
% On p. 51 it says that Parse is unordered with respect to Faith. This 
% must be modified since Faith is split into a family of constraints, 
% which may even come after M-Parse, as explained by Michael Hammond
% in the following e-mail
% 
% My question:
% > > Here's the problem I see. First, the tableaux on p. 51 only makes
% > > sense with Faith(POA) >> M-Parse. If Faith(POA) came after
% > > M-Parse, then /Na/ could surface as na or ma.
% > >
% > > But the Faith(POA) >> M-Parse ordering implicitly assumed on p. 51 is
% > > contradicted by the ordering implicitly assumed on p. 103. On p. 103,
% > > there's a non-null successful candidate that violates Faith(POA). For
% > > Faith(POA) to be violable, it must be ordered after M-Parse.
% 
% Michael Hammond's response:
% > i think that's correct. one way to solve this would be to decompose
% > Faith(POA) into two separate constraints. One would be responsible for POA
% > differences among coronals, and the other would be responsible for POA
% > differences with the velars. They could then be ranked to either side of
% > MParse.
%
% A sonorant or noncoronal must be parsed. One question is whether or
% not an unparsed cor_dor should be a violation. In the text, Hammond
% says a noncoronal must be parsed, but in the formalization, he
% speaks of a "segment with a marked place of articulation". I assume
% that a cor_dor does have a marked place of articulation.
macro(mark_ot_constraint('Parse-POA-Sonorant',Mark),
   mark([{sonorant_, (segment_ - cor_)},xp],Mark)).

macro(mark_ot_constraint('Pre-Parse',Mark),
   mark([xp,segment,segment_,(? - xp)],Mark)).


% Max-IO (No Deletion): 
% This is one half of the Faith(Segments) contraint used by
% Hammond. 
macro(mark_ot_constraint('Max-IO',Mark), mark([del,del],Mark)).


% Dep-IO (No Insertion).
% Look for eps (epsilon), which is similar to the 0 as used in Kimmo.
% This would indicate an insertion site.
macro(mark_ot_constraint('Dep-IO',Mark),
   mark(ins,Mark)).



macro(mark_ot_constraint('Faith(MOA)',Mark),
   mark({[fricative,{affricate_,approximant_,stop_,nasal_}],
	 [affricate,{fricative_,approximant_,stop_,nasal_}],
	 [approximant,{fricative_,affricate_,stop_,nasal_}],
	 [stop,{fricative_,affricate_,approximant_,nasal_}],
	 [nasal,{fricative_,affricate_,approximant_,stop_}]
	},Mark)
           o
   mark({[l,r_],[r,l_]},Mark)
           o
   mark({[dist,not_dist_],[not_dist,dist_]},Mark)
).



% Faith(dist): Not in Hammond
% A constraint like this could be used for the /srink/ example on
% p. 102, 103. The book suggests that the output is SriNk, but nothing
% rules out TriNk as an alternative. Hammond, in an email writes:
% > 
% > <Trink> would be o.k., and as you suggest, maybe even the
% > best. Note that we really don't know how /srink/ would be
% > realized; there are no relevant alternations.
macro(mark_ot_constraint('Faith(dist)',Mark),
   mark({[dist,not_dist_],[not_dist,dist_]},Mark)).
      

% Underlying cor or cor_dor, must surface as cor or cor_dor
% What about COR (not distributed)??
macro(mark_ot_constraint('Faith(COR)',Mark),
   mark(unfaithful_cor, Mark)).

macro(unfaithful_cor,
   {  [{cor,cor_dor},{lab_,lab_dor_,dor_,h_}],
      [{lab,lab_dor,dor,h_},{cor_,cor_dor_}]}).

% Faith(DOR): Not in book, but suggested by Hammond in email.
%             See comment for Parse.

% The idea is that something that is underlying dorsal (velar) must
% surface as dorsal. Unfortunately, the email doesn't give
% details. Can a dor surface as cor_dor or lab_dor? I assume (without
% any particular evidence) that the answer is yes.
macro(mark_ot_constraint('Faith(DOR)',Mark),
   mark(unfaithful_dor, Mark)).

macro(unfaithful_dor,
   { [{dor,cor_dor,lab_dor},{lab_,cor_}],
     [{lab,cor},{dor,cor_dor,lab_dor}]}).

macro(mark_ot_constraint('Faith(Voicing)',Mark),
   mark({[voiced,voiceless_],[voiceless,voiced_]},Mark)).


macro(mark_ot_constraint(h_proscription,Mark),
   mark([h_,co],Mark)).

% p. 80: a "linear constraint", Cy constraint
macro(mark_ot_constraint('CyConstraint',Mark),
   mark([consonant_,
	 (tag - del),
	 deleted*,
	 segment,
	 y_,
	 (tag - del)],
	Mark)).

% NoGeminates: p. 73
macro(mark_ot_constraint('NoGeminates',Mark),
   mark(geminate,Mark)).

% FricativeObstruent: p. 76
% This is labeled as "preliminary", but I can't find where a final
% version is ever presented. There's also a problem with the features
% used to pick out the relevant class of fricatives, so I've just
% listed them.
macro(mark_ot_constraint('FricativeObstruent',Mark),
   mark([{'S_','Z_',v_,'T_','D_'},(? - del),
	 deleted*,
	 obstruent_],
	Mark)).


% VoicingWithFricatives: p. 77
macro(mark_ot_constraint('VoicingWithFricatives',Mark),
   mark([(voiceless_ & obstruent_), %% Formalization on p. 77 doesn't
	 (tag - del),               %%    mention obstruent. h is the
	 deleted*,                  %%    only voiceless non-obstruent
	 segment,
	 (voiced_ & {fricative_,affricate_}),
	 (tag - del)],
	Mark)).


% Onset/{N,Z}: p. 51
macro(mark_ot_constraint('Onset/{N,Z}',Mark),
   mark([{'N_','Z_'},on],Mark)).



% '*o/ao': p. 90
% Note that this and other derived ranking constraints are subject to
% the "Meta-constraint for [s]" on p. 98:
%   Constraints involving [s] are not subject to derived ranking
% Since s is an obstruent, this means that any derived constraint that 
% mentions obstruent_, should really mention (obstruent_ - s). But I
% think that this meta constraint is not quite correct as stated. It
% should refer to a peripheral [s]. The constraint against approximant 
% followed by obstruent should rule out onsets like [ls].
macro(mark_ot_constraint('*o/ao',Mark),
   mark([approximant_,
	 on,
	 deleted*,
	 segment,
	 obstruent_,
	 on],
	Mark)).

% '*o/an': p. 90
macro(mark_ot_constraint('*o/an',Mark),
   mark([approximant_,
	 on,
	 deleted*,
	 segment,
	 nasal_,
	 on],
	Mark)).

% '*o/no': p. 90
macro(mark_ot_constraint('*o/no',Mark),
   mark([nasal_,
	 on,
	 deleted*,
	 segment,
	 obstruent_,
	 on],
	Mark)).

% '*o/na': p. 90
macro(mark_ot_constraint('*o/na',Mark),
   mark([nasal_,
	 on,
	 deleted*,
	 segment,
	 approximant_,
	 on],
	Mark)).


% '*o/on': p. 90 (Rules out marginally possible knish, tmesis, knesset)
macro(mark_ot_constraint('*o/on',Mark),
   mark([(obstruent_ - s_),   % see comment above on meta-constraint
	 on,
	 deleted*,
	 segment,
	 nasal_,
	 on],
	Mark)).

% '*o/oa': p. 90
macro(mark_ot_constraint('*o/oa',Mark),
   mark([(obstruent_ - s_),
	 on,
	 deleted*,
	 segment,
	 approximant_,
	 on],
	Mark)).

% The following constraints are not in Hammond. But clearly the
% sonority hierarchy should rule out such sequences. Note that the
% constraint *o/oo is dealt with on p. 96

macro(mark_ot_constraint('*o/nn',Mark),
   mark([nasal_,
	 on,
	 deleted*,
	 segment,
	 nasal_,
	 on],
	Mark)).

macro(mark_ot_constraint('*o/aa',Mark),
   mark([approximant_,
	 on,
	 deleted*,
	 segment,
	 approximant_,
	 on],
	Mark)).





% Intrinsic onset and coda constraints. These are mostly important
% for the resulting derived constraints.

macro(mark_ot_constraint('*o/a',Mark),
   mark([approximant_,on],Mark)).

macro(mark_ot_constraint('*o/n',Mark),
   mark([nasal_,on],Mark)).

macro(mark_ot_constraint('*o/o',Mark),
   mark([obstruent_,on],Mark)).

macro(mark_ot_constraint('*c/o',Mark),
   mark([obstruent_,co],Mark)).

macro(mark_ot_constraint('*c/n',Mark),
   mark([nasal_,co],Mark)).

macro(mark_ot_constraint('*c/l',Mark),
   mark([l_,co],Mark)).

macro(mark_ot_constraint('*c/r' ,Mark),
   mark([r_,co],Mark)).


macro(mark_ot_constraint('*c/or',Mark),
   mark([obstruent_,
	 co,
	 deleted*,
	 segment,
	 r_,
	 co],
	Mark)).

macro(mark_ot_constraint('*c/ol',Mark),
   mark([obstruent_,
	 co,
	 deleted*,
	 segment,
	 l_,
	 co],
	Mark)).

macro(mark_ot_constraint('*c/nr',Mark),
   mark([nasal_,
	 co,
	 deleted*,
	 segment,
	 r_,
	 co],
	Mark)).

macro(mark_ot_constraint('*c/on',Mark),
   mark([obstruent_,
	 co,
	 deleted*,
	 segment,
	 nasal_,
	 co],
	Mark)).

macro(mark_ot_constraint('*c/nl',Mark),
   mark([nasal_,
	 co,
	 deleted*,
	 segment,
	 l_,
	 co],
	Mark)).

macro(mark_ot_constraint('*c/lr',Mark),
   mark([l_,
	 co,
	 deleted*,
	 segment,
	 r_,
	 co],
	Mark)).

macro(mark_ot_constraint('*c/no',Mark),
   mark([nasal_,
	 co,
	 deleted*,
	 segment,
	 (obstruent_ - s_),
	 co],
	Mark)).

macro(mark_ot_constraint('*c/ln',Mark),
   mark([l_,
	 co,
	 deleted*,
	 segment,
	 nasal_,
	 co],
	Mark)).

macro(mark_ot_constraint('*c/rl',Mark),
   mark([r_,
	 co,
	 deleted*,
	 segment,
	 l_,
	 co],
	Mark)).

macro(mark_ot_constraint('*c/lo',Mark),
   mark([l_,
	 co,
	 deleted*,
	 segment,
	 (obstruent_ - s_),
	 co],
	Mark)).

macro(mark_ot_constraint('*c/rn',Mark),
   mark([r_,
	 co,
	 deleted*,
	 segment,
	 nasal_,
	 co],
	Mark)).

macro(mark_ot_constraint('*c/ro',Mark),
   mark([r_,
	 co,
	 deleted*,
	 segment,
	 (obstruent_ - s_),
	 co],
	Mark)).

% MaximalCoda: p. 94 : Note the C+COR exception is dealt with p. 98
macro(mark_ot_constraint('MaximalCoda',Mark),
   mark([segment_,co,deleted*,
	 segment,segment_,co,deleted*,
	 segment,segment_,co],
	Mark)).

% Sc: p. 95
macro(mark_ot_constraint('Sc',Mark),
   mark((  [[segment,obstruent_,on],deleted*,[segment,obstruent_,on]]
	    -
           [[segment,s_,on],deleted*,[segment,stop_,on]]),
	Mark)).
	

% obstruent-obstruent constraints: p. 96

macro(mark_ot_constraint('*o/stop-o',Mark),
   mark([stop_,on,deleted*,segment,obstruent_,on],Mark)).

macro(mark_ot_constraint('*o/[f]-o',Mark),
   mark([f_,on,deleted*,segment,obstruent_,on],Mark)).

macro(mark_ot_constraint('*o/[T]-o',Mark),
   mark(['T_',on,deleted*,segment,obstruent_,on],Mark)).

macro(mark_ot_constraint('*o/[S]-o',Mark),
   mark(['S_',on,deleted*,segment,obstruent_,on],Mark)).

macro(mark_ot_constraint('*o/voiced-o',Mark),
   mark([voiced_,on,deleted*,segment,obstruent_,on],Mark)).

macro(mark_ot_constraint('*o/[s]-o',Mark),
   mark(['s_',on,deleted*,segment,obstruent_,on],Mark)).

macro(mark_ot_constraint('*o/stop-N',Mark),
   mark([stop_,on,deleted*,segment,nasal_,on],Mark)).

macro(mark_ot_constraint('*o/[f]-N',Mark),
   mark([f_,on,deleted*,segment,nasal_,on],Mark)).

macro(mark_ot_constraint('*o/[T]-N',Mark),
   mark(['T_',on,deleted*,segment,nasal_,on],Mark)).

macro(mark_ot_constraint('*o/[S]-N',Mark),
   mark(['S_',on,deleted*,segment,nasal_,on],Mark)).

macro(mark_ot_constraint('*o/voiced-N',Mark),
   mark([voiced_,on,deleted*,segment,nasal_,on],Mark)).

macro(mark_ot_constraint('*o/[s]-N',Mark),
   mark(['s_',on,deleted*,segment,nasal_,on],Mark)).

macro(mark_ot_constraint('*c/stop-o',Mark),
   mark([stop_,co,deleted*,segment,(obstruent_ - s_),co],Mark)).

macro(mark_ot_constraint('*c/[f]-o',Mark),
   mark([f_,co,deleted*,segment,(obstruent_ - s_),co],Mark)).

macro(mark_ot_constraint('*c/[T]-o',Mark),
   mark(['T_',co,deleted*,segment,(obstruent_ - s_),co],Mark)).

macro(mark_ot_constraint('*c/[S]-o',Mark),
   mark(['S_',co,deleted*,segment,(obstruent_ - s_),co],Mark)).

macro(mark_ot_constraint('*c/voiced_obst-o',Mark),
   mark([(voiced_ & obstruent_),co,deleted*,segment,(obstruent_ - s_),co],Mark)).

macro(mark_ot_constraint('*c/[s]-o',Mark),
   mark(['s_',co,deleted*,segment,(obstruent_ - s_),co],Mark)).


% M-Parse: p. 51
% I treat the "null parse" as a kind of deletion of the input. 
% But the deletion is distinguished from ordinary deletion, since the
% "null parse" is not supposed to violate any faithfulness
% constraints. It's unclear whether or not the null parse could
% violate any markedness constraints, but it appears from Hammond's
% book, that this is never the case.
% The approach used here is a kind of short-circuit approach. Gen is
% written so that either every or no segment is marked as [xp,del]. So 
% we can just look at the first segment to see if it's [xp,del] and
% introduce just one mark if it is.
macro(mark_ot_constraint('M-Parse',Mark),
   { [?,xp,del,[]:Mark,? *], [?, [?,?]-[xp,del],? *] }).
 

% Faith(POA):
% This is Faith(POA) with Faith(COR) and Faith(COR) factored out.
macro(mark_ot_constraint('Faith(POA)',Mark),
   mark(unfaithful_poa - {unfaithful_cor, unfaithful_dor}, Mark)).

macro(unfaithful_poa,
   { [lab,     {cor_,cor_dor_,dor_,lab_dor_}],
     [cor,     {lab_,cor_dor_,dor_,lab_dor_}],
     [cor_dor, {lab_,cor_,dor_,lab_dor_}],
     [dor,     {lab_,cor_,cor_dor_,lab_dor_}],
     [lab_dor, {lab_,cor_,cor_dor_,dor_}] }).

      

% Linear Restrictions: Sec 3.5

% AffricateProscription: p. 101
% Although in the section titled "Linear Restrictions," this is not a
% linear restriction.
% 
macro(mark_ot_constraint('AffricateProscription',Mark),
   replace([segment,affricate_,on,[]:Mark],
	   [onset,Mark^,deleted*],
	   [])
                      o
   replace([segment,affricate_,on,[]:Mark],
	   [],
	   [deleted*,onset])
                      o
     replace([Mark,Mark]:Mark)).
% mark_ot_constraint('AffricateProscription',violation_marker)  

% *[sr]: p. 102
macro(mark_ot_constraint('*[sr]',Mark),
   mark([s_,(tag - del),deleted*,segment,r_,(tag - del)],Mark)).

% LateralConstraint: p. 103
macro(mark_ot_constraint('LateralConstraint',Mark),
   mark([(cor_ & stop_),on,deleted*,segment,l_,on],Mark)).

% Nonhomorganic_NC_Codas: p. 103

% The POA features for this constraint are somewhat problematic. 
% The chart on p. 5 does not make clear whether or not cor_dor is
% intended to be both cor and dor, or whether it's a distinct place
% between cor and dor. I have assumed the latter interpretation. But
% this produces problems with words like 'bench'. Phonetically, the
% [n] here should be cor_dor, but there's no such nasal segment in our 
% inventory. So I assume that a nasal coronal may be followed by a
% cor_dor. But then why can't a nasal dorsal also be followed by a
% cor_dor? A similar problem occurs for the hypothetical coda sequence 
% mw. Here the /m/ should become lab_dor. It seems that this coda
% sequence should be ruled out by one of the sonority-based coda
% constraint on p. 91, but these constraints don't mention w.
% 
% Note also that this constraint is allowed to be violated when there
% is a morpheme boundary if there is a morpheme boundary after the
% nasal. But morpheme boundaries are not yet handled in this
% implementation. 
macro(mark_ot_constraint('Nonhomorganic_NC_Codas',Mark),
   mark({ [(nasal_ & lab_),co,deleted*,segment,(? - lab_),co],
	  [(nasal_ & cor_),co,deleted*,segment,(? - {cor_,cor_dor_}),co],
	  [(nasal_ & dor_),co,deleted*,segment,(? - dor_),co] }, Mark)).
	 
% 'NC_Codas_With_Voiced_Obstruent': p. 104 
macro(mark_ot_constraint('NC_Codas_With_Voiced_Obstruent',Mark),
   mark([nasal_,co,deleted*,segment,(voiced_ & obstruent_),co],Mark)).



%% More alphabet types 

macro(vowel, {   % The Sampa American Vowels
   'I',          %   pit                     pIt
   'E',          %   pet                     pEt
   '{',          %   pat                     p{t   }}
   'A',          %   pot                     pAt
   'V',          %   cut                     kVt
   'U',          %   put                     pUt
   'i',          %   ease                    iz
   'e',          %   raise                   rez
   'u',          %   lose                    luz
   'o',          %   nose                    noz
   'O',          %   cause                   kOz
   'aI',         %   rise                    raIz
   'OI',         %   noise                   nOIz
   'aU',         %   rouse                   raUz
   '3`',         %   furs                    f3`z
   '@' ,         %   allow                   @laU
   '@`'          %   corner                  kOrn@`
   }).


macro(consonant,  % The Sampa American Consonants (except y replaces j)
 { 
   p,               %    pin                     pIn
   b,               %    bin                     bIn
   t,               %    tin                     tIn
   d,               %    din                     dIn
   k,               %    kin                     kIn
   g,               %    give                    gIv
   tS,              %    chin                    tSIn
   dZ,              %    gin                     dZIn
   f,               %    fin                     fIn
   v,               %    vim                     vIm
   'T',             %    thin                    TIn
   'D',             %    this                    DIs
   s,               %    sin                     sIn
   z,               %    zing                    zIN
   'S',             %    shin                    SIn
   'Z',             %    measure                 "mEZ@` `"
   h,               %    hit                     hIt
   m,               %    mock                    mAk
   n,               %    knock                   nAk
   'N',             %    thing                   TIN
   r,               %    wrong                   rON
   l,               %    long                    lON
   w,               %    wasp                    wAsp
   y                %    yacht                   yAt
   }).

%% Consonant classes: Note l and r and not distinguished. And h fits
%% into no POA class at all. Hammond is unclear about whether or not a 
%% cor_dor counts as being both coronal and dorsal (same with
%% lab_dor). I assume that cor_dor is not cor, so if you really want
%% all coronals including the cor_dor class, you need to use the union: 
%% {cor, cor_dor}. I think that this is what Hammond means: cor_dor is
%% really a place between cor and dor and is not in either the
%% cor or the dor class. 

macro(voiceless, {p, t, k, tS, f, 'T', s, 'S', h}).

macro(obstruent, {stop, fricative, affricate}).

macro(voiced, {b, d, g, dZ, v, 'D', z, 'Z', m, n, 'N', r, l, w, y}).

macro(fricative, {f, v, 'T', 'D', s, z, 'S', 'Z'}).

macro(affricate, {tS, dZ}).

macro(approximant, {r, l, w, y, h}).

% sonorants also include vowels
macro(sonorant,{approximant,nasal}).

macro(stop, {p, t, k, b, d, g}).

macro(nasal, {m, n, 'N'}).

macro(lab, {p, b, f, v, m}). 

macro(cor, {t, d, 'T', 'D', s, z, n, r, l}).
macro(cor_dor, {tS, dZ, 'S', 'Z', y}).

% I include macros for both dist and not_dist since these cases are
% not exhaustive. The distinction is only relevant for COR.
macro(dist, {t, d, s, z, n, r, l}).
macro(not_dist, {'T', 'D'}).

macro(dor, {k, g, 'N'}).
macro(lab_dor, w).

macro(consonant_pair, {[p,p],[b,b],[t,t],[d,d],[k,k],[g,g],[tS,tS],
   [dZ,dZ],[f,f],[v,v],['T','T'],['D','D'],[s,s],[z,z],['S','S'],
   ['Z','Z'],[h,h],[m,m],[n,n],['N','N'],[r,r],[l,l],[w,w],[y,y]}).

macro(consonant_pair_, {[p_,p_],[b_,b_],[t_,t_],[d_,d_],[k_,k_],[g_,g_],
   [tS_,tS_],[dZ_,dZ_],[f_,f_],[v_,v_],['T_','T_'],['D_','D_'],[s_,s_],
   [z_,z_],['S_','S_'],['Z_','Z_'],[h_,h_],[m_,m_],[n_,n_],['N_','N_'],
   [r_,r_],[l_,l_],[w_,w_],[y_,y_]}).

%% Only consonant geminates
macro(geminate,
   range(consonant_pair_ o [([] : segment), segment_, []:{on,nu,co},
			    ([]:deleted*),
			    ([] : segment), segment_, []:{on,nu,co}])).



macro(vowel_, {   % The Sampa American Vowels
   'I_',          %   pit                     pIt
   'E_',          %   pet                     pEt
   '{_',          %   pat                     p{t   }}
   'A_',          %   pot                     pAt
   'V_',          %   cut                     kVt
   'U_',          %   put                     pUt
   'i_',          %   ease                    iz
   'e_',          %   raise                   rez
   'u_',          %   lose                    luz
   'o_',          %   nose                    noz
   'O_',          %   cause                   kOz
   'aI_',         %   rise                    raIz
   'OI_',         %   noise                   nOIz
   'aU_',         %   rouse                   raUz
   '3`_',         %   furs                    f3`z
   '@_' ,         %   allow                   @laU
   '@`_'          %   corner                  kOrn@`
   }).


macro(consonant_, {   % The Sampa American Consonants
   p_,               %    pin                     pIn
   b_,               %    bin                     bIn
   t_,               %    tin                     tIn
   d_,               %    din                     dIn
   k_,               %    kin                     kIn
   g_,               %    give                    gIv
   tS_,              %    chin                    tSIn
   dZ_,              %    gin                     dZIn
   f_,               %    fin                     fIn
   v_,               %    vim                     vIm
   'T_',             %    thin                    TIn
   'D_',             %    this                    DIs
   s_,               %    sin                     sIn
   z_,               %    zing                    zIN
   'S_',             %    shin                    SIn
   'Z_',             %    measure                 "mEZ@` `"
   h_,               %    hit                     hIt
   m_,               %    mock                    mAk
   n_,               %    knock                   nAk
   'N_',             %    thing                   TIN
   r_,               %    wrong                   rON
   l_,               %    long                    lON
   w_,               %    wasp                    wAsp
   y_                %    yacht                   yAt
   }).

%% Consonant classes: Note l and r and not distinguished. And h fits
%% into no POA class at all.

macro(voiceless_, {p_, t_, k_, tS_, f_, 'T_', s_, 'S_', h_}).

macro(obstruent_, {stop_, fricative_, affricate_}).

macro(voiced_,
      {b_, d_, g_, dZ_, v_, 'D_', z_, 'Z_', m_, n_, 'N_', r_, l_, w_, y_}).

macro(fricative_, {f_, v_, 'T_', 'D_', s_, z_, 'S_', 'Z_'}).

macro(affricate_, {tS_, dZ_}).

macro(approximant_, {r_, l_, w_, y_, h_}).

% sonorants also include vowels
macro(sonorant_,{approximant_,nasal_}).

macro(stop_, {p_, t_, k_, b_, d_, g_}).

macro(nasal_, {m_, n_, 'N_'}).

macro(lab_, {p_, b_, f_, v_, m_}).

macro(cor_, {t_, d_, 'T_', 'D_', s_, z_, n_, r_, l_}).
macro(cor_dor_, {tS_, dZ_, 'S_', 'Z_', y_}).

macro(dist_, {t_, d_, s_, z_, n_, r_, l_}).
macro(not_dist_, {'T_', 'D_'}).

macro(dor_, {k_, g_, 'N_'}).
macro(lab_dor_, w_).

% 
% vowel_io: A tagger that tags each vowel with an identical
% output vowel. This is used simply because this grammar implements
% chap 3 of Hammond, where vowels are not yet treated. So gen will
% have to uniformly map every vowel to an identical output vowel and
% mark it as nucleus.
macro(vowel_io,
   {['I',[]:'I_'],['E',[]:'E_'],['{',[]:'{_'],['A',[]:'A_'],
    ['V',[]:'V_'],['U',[]:'U_'],['i',[]:'i_'],['e',[]:'e_'],
    ['u',[]:'u_'],['o',[]:'o_'],['O',[]:'O_'],['aI',[]:'aI_'],
    ['OI',[]:'OI_'],['aU',[]:'aU_'],['3`',[]:'3`_'],['@',[]:'@_'],
    ['@`',[]:'@`_']
    }).




% testing macros

macro(get_onset,
   minimize(
   range(                  chap3
	                     o
	[(? * x []),[onset,{onset,deleted:[]}*],(? * x [])]
                             o
               [segment:[],remove_underscore,on:[]]*
     ))).


macro(get_coda,
   minimize(
   range(                  chap3
	                     o
	[(? * x []),[coda,{coda,deleted:[]}*],(? * x [])]
                             o
               [segment:[],remove_underscore,co:[]]*
     ))).


macro(get_medial_cluster,
   minimize(range(
		        ['A',consonant*,'A']
		              o
		            chap3
		              o
 [[?,?,?]:[],{deleted:[],[segment:[],remove_underscore,{co,on,xp}]}*,[?,?,?]:[]]
))).



macro(remove_underscore,
   {'I_':'I', 'E_':'E', '{_':'{', 'A_':'A', 'V_':'V', 'U_':'U',  i_: i, % }}
    e_: e,  u_: u,  o_: o, 'O_':'O',  aI_: aI, 'OI_':'OI',  aU_: aU,
    '3`_':'3`', '@_' :'@' , '@`_':'@`',  p_: p,  b_: b,  t_: t,  d_: d,
    k_: k,  g_: g,  tS_: tS,  dZ_: dZ,  f_: f,  v_: v, 'T_':'T', 'D_':'D',
    s_: s,  z_: z, 'S_':'S', 'Z_':'Z',  h_: h,  m_: m,  n_: n, 'N_':'N',
    r_: r,  l_: l,  w_: w,  y_: y}).


macro(remove_contained_input,
   { [merge_io_pair,tag],
     [ins:[],remove_underscore,tag],
     [?,?,del]:[] }*).

macro(merge_io_pair,
   {['I',('I_':[])],[(? - 'I'),('I_':'I')],
    ['E',('E_':[])],[(? - 'E'),('E_':'E')],
    ['{',('{_':[])],[(? - '{'),('{_':'{')],  % }
    ['A',('A_':[])],[(? - 'A'),('A_':'A')],
    ['V',('V_':[])],[(? - 'V'),('V_':'V')],
    ['U',('U_':[])],[(? - 'U'),('U_':'U')],
    [ i,(i_:[])],[(? -  i),(i_:i)],
    [ e,(e_:[])],[(? -  e),(e_:e)],
    [ u,(u_:[])],[(? -  u),(u_:u)],
    [ o,(o_:[])],[(? -  o),(o_:o)],
    ['O',('O_':[])],[(? - 'O'),('O_':'O')],
    [ aI,(aI_:[])],[(? -  aI),(aI_:aI)],
    ['OI',('OI_':[])],[(? - 'OI'),('OI_':'OI')],
    [ aU,(aU_:[])],[(? - aU),(aU_:aU)],
    ['3`',('3`_':[])],[(? - '3`'),('3`_':'3`')],
    ['@' ,('@_' :[])],[(? - '@' ),('@_' :'@' )],
    ['@`',('@`_':[])],[(? - '@`'),('@`_':'@`')],
    [ p,(p_:[])],[(? - p),(p_:p)],
    [ b,(b_:[])],[(? - b),(b_:b)],
    [ t,(t_:[])],[(? - t),(t_:t)],
    [ d,(d_:[])],[(? - d),(d_:d)],
    [ k,(k_:[])],[(? - k),(k_:k)],
    [ g,(g_:[])],[(? - g),(g_:g)],
    [ tS,(tS_:[])],[(? - tS),(tS_:tS)],
    [ dZ,(dZ_:[])],[(? - dZ),(dZ_:dZ)],
    [ f,(f_:[])],[(? - f),(f_:f)],
    [ v,(v_:[])],[(? - v),(v_:v)],
    ['T',('T_':[])],[(? - 'T'),('T_':'T')],
    ['D',('D_':[])],[(? - 'D'),('D_':'D')],
    [ s,(s_:[])],[(? - s),(s_:s)],
    [ z,(z_:[])],[(? - z),(z_:z)],
    ['S',('S_':[])],[(? - 'S'),('S_':'S')],
    ['Z',('Z_':[])],[(? - 'Z'),('Z_':'Z')],
    [ h,(h_:[])],[(? - h),(h_:h)],
    [ m,(m_:[])],[(? - m),(m_:m)],
    [ n,(n_:[])],[(? - n),(n_:n)],
    ['N',('N_':[])],[(? - 'N'),('N_':'N')],
    [ r,(r_:[])],[(? - r),(r_:r)],
    [ l,(l_:[])],[(? - l),(l_:l)],
    [ w,(w_:[])],[(? - w),(w_:w)],
    [ y,(y_:[])],[(? - y),(y_:y)]}).