%% (c) 2000-2002 Dale Gerdemann, Gertjan van Noord; available under GPL

%% This is a slightly modified version of matching.pl and counting.pl
%% from Examples/OptimalityTheory which is distributed with the FSA
%% Utilities: http://odur.let.rug.nl/~vannoord/fsa/.

%% The primary operator defined here is 'oo', the optimality operator
%% as described in: Gerdemann & van Noord, "Approximation and
%% Exactness in Finite State Optimality Theory," SIGPHON 2000, Fifth
%% Meeting of the ACL Special Interest Group in Computational
%% Phonology, COLING 2000, http://roa.rutgers.edu.

%% Additionally a separate operator 'lc' is defined for lenient
%% composition. This operator is basically as described by Karttunen
%% in "The Proper Treatment of Optimality in Computational
%% Phonology'. in: FSMNLP 98, http://roa.rutgers.edu. The difference
%% is that it is modified to apply constraints that insert constraint
%% violation markers as defined with mark_ot_constraint/2.
%%
%% The set of mark_ot_constraint/2 macros must be defined by the user
%% for the particular OT theory. For a constraint C, the macro
%% mark_ot_constraint(C,M) should be defined to expand to a transducer
%% that inserts the marker M after each violation of the constraint C.
%% Such a transducer can be defined in any way that the user finds
%% convenient. One generally useful macro, mark(Pattern,Marker), is
%% tentatively included in this file. This macro produces a transducer
%% that introduces a Marker after every instance of Pattern, as long
%% as the instances of Pattern do not overlap.
%%
%% Both the 'oo' and 'lc' operator can be supplied with an extra
%% parameter for precision. For example: gen oo 2 :: <ConstraintName>, 
%% or: gen lc 3 :: <ConstraintName>. For the optimality operator,
%% precision means the allowable amount of permutation. For the
%% lenient composition operator, precision means the depth bound for
%% counting constraint violations. For details, see the Gerdemann &
%% van Noord SIGPHON paper.
%%
%% A third operator, 'mc', is defined here for "merciless
%% composition". In Karttunen's (FSMNLP 98) approach, merciless
%% composition is simply ordinary composition with an identity
%% transducer. This is changed here so that the right argument of 'mc' 
%% is a constraint name defined with mark_ot_constraint/2. The effect, 
%% however, is the same as in Karttunen's approach. A candidate which
%% violates the constraint at all is eliminated even if this
%% eliminates every candidate for some inputs.
%%
%% The three operators, 'oo', 'lc' and 'mc' are all defined with the
%% same precedence and with left associativity so that they can be
%% combined in a sequence of constraint applications. Normally, 'oo'
%% is the only operator which is needed. However, in some cases 'lc'
%% or 'mc' can be substituted for efficiency. For example, if a
%% constraint is applied in a context where you can be certain that
%% there will always be at least one non-violating candidate, then
%% merciless composition can be applied. Some constraints are such
%% that they will be violated either just once or not at all. For such
%% constraints, the 'lc' operator is sufficient. For a constraint that
%% can be violated at most 2 times (perhaps a constraint referring to
%% the edges of a word), the 'lc 2 ::' operator construction could be
%% used.

:- op(403, yfx, 'oo').
:- op(403, yfx, 'lc').
:- op(403, yfx, 'mc').
:- op(400, yfx, '::').
:- op(400,yfx,(^>)).
:- op(403,yfx,(lc_base)).


:- multifile macro/2.
:- discontiguous macro/2.


%% The violation marker:
%% Note that the violation marker can be changed. But it  must be a
%% distinct symbol. Perhaps some ASCII representation for phonemes
%% uses '*' to represent a particular segment. If you use such a
%% representation, then the violation marker must be changed to
%% something else.
macro(violation_marker, '*').



%% optimality operator
macro(Input oo Precision :: Constraint,oo(Input,Precision,Constraint)).
macro(Input oo              Constraint,oo(Input,0,        Constraint)).

macro(oo(Input,Precision,Constraint),
      pragma([V-(Input o mark_ot_constraint(Constraint,violation_marker))],
                    V
                    o
             ~ range(  V
		       o
		  add_violation(Precision)
		    )
                    o
             { violation_marker : [],? -  violation_marker }*		 
		    )).


macro(add_violation(N), 
      {output : [], ? - output}*
                     o
           [[? *,[]: violation_marker ]+, ? *]
                     o
             permute_marker(N)
                     o
      {[] : output, ? - output}*
     ).

macro(permute_marker(0), ? *).
macro(permute_marker(N), cache(permute_marker) o permute_marker(N1)) :-
    N>0, N1 is N-1.

macro(permute_marker,
      [{[? *,  violation_marker : [], ? *, [] : violation_marker ],
	[? *, [] : violation_marker , ? *,  violation_marker : []]}*, ? *]).







%% Lenient composition
macro(Input lc Precision :: Constraint, lc(Input,Precision,Constraint)).
macro(Input lc              Constraint, lc(Input,0        ,Constraint)).

%% apply Constraint to Input with Precision
macro(lc(Input,Precision,Constraint),
                   Body
                    o
 { violation_marker : [], ? - violation_marker }*
     ) :-
    get_constraint(Precision,
		   Input
		   o
		   mark_ot_constraint(Constraint,violation_marker), Body).

get_constraint(Precision,B0,B) :-
    (	Precision =< 0
    ->	B = (B0 lc_base ~ ($ violation_marker))
    ;   B = (B1 lc_base ~ ($ violation_marker ^> Precision)),
	Precision1 is Precision-1,
	get_constraint(Precision1,B0,B1)
    ).


%% As in Karttunen, the (A ^> N) operator means more than N iterations
%% of A.
macro(A ^> N,[AN,A+]) :-
   make_list(N,A,AN).

%% make_list(+Length,+ListMember,?List)
%% Construct a list full of ListMember
make_list(N,A,List) :-
    (	N=:=0
    ->	List=[]
    ;	List=[A|List0],
	N1 is N-1,
	make_list(N1,A,List0)
    ).

%%macro(priority_union(Q,S), {Q, ~domain(Q) o S}).
%%macro(R lc_base C, priority_union(R o C,R)).

macro(R0 lc_base C0, pragma([R-R0,C-C0], {R o C, ~domain(R o C) o R} )).




%% Merciless composition
macro((Input mc Constraint),
      Input
        o
     domain( mark_ot_constraint(Constraint,violation_marker)
	                       o
                    ~ $ violation_marker)).




%% mark(Pattern,Marker) : Insert Marker after instances of Pattern
%% 
%% To be used in marking constraint violations.  This is a simple
%% approach that does not handle overlaps. For example, if Pattern
%% matches from string positions 1-4, 2-6 and 5-8, then only the
%% matches from 1-4 and 5-8 would be marked.
%% 
%% The macro is implemented using a standard idiom, familiar from the
%% Xerox' bracketing operator: X -> L ... R. The term (~ $ Pattern) can
%% be read as "does not contain Pattern. The term [Pattern, [] x Marker]
%% matches one instance of Pattern and then inserts an instance of
%% Marker.  Note here, that [] indicates epsilon, the empty string, and
%% 'x' indicates cross-product, so ([] x Marker) corresponds to a
%% transducer that maps the empty string to a marker. This expression
%% could also be written as ([]:Marker). The Kleene star allows the
%% Pattern to be marked repeatedly in the input string.

macro(mark(Pattern,Marker),
      [[~ $ Pattern, [Pattern, [] x Marker]]*, ~ $ Pattern]).

