:- module( fsa_preds_bbb, []).

:- multifile user:help_info/4.

user:help_info(module,fsa_preds_bbb,"Predicates on symbols: fsa_preds_bbb",
"This module is one of the standard `predicate' modules. It provides a 
consistent set of predicate_declarations. 

In this predicate module, it is assumed that labels in a finite
automaton are associated with sets of symbols (represented with
in(List) and with negated sets of symbols, represented with
not_in(List). Singleton sets in([El]) are represented with El.

The negated sets are useful to provide a treatment of the any symbol
?/0 operator, in a way which is functionally equivalent to the
treatment of ?/0 in FSA5. For instance, the expression ? - a will
result in an automaton with a transition over not_in([a]). 

The module has a full set of predicate_declarations and can be used as 
a model for new, user-defined task-specific predicate modules.
").

true(not_in(E)) :-
    set_bbbtree:set_bbbtree__empty(E).

regex_notation_to_predicate(Atom,in(Tree)) :-
    set_bbbtree:set_bbbtree__init(Tree0),
    set_bbbtree:set_bbbtree__insert(Tree0,Atom,Tree).

evaluate_predicate(in(List),Symbol) :-
    set_bbbtree:set_bbbtree__member(Symbol,List).

evaluate_predicate(not_in(Tree),Symbol) :-
    set_bbbtree:set_bbbtree__to_sorted_list(Tree,List),
    dif_list(List,Symbol).

dif_list([],_).
dif_list([H|T],Symbol) :-
    dif(H,Symbol),
    dif_list(T,Symbol).

talks_about(in(List),Symbol) :-
    set_bbbtree:set_bbbtree__member(Symbol,List).

talks_about(not_in(List),Symbol) :-
    talks_about(in(List),Symbol).

disjunction(not_in(A),B,C):-
    disjunction_not_in(B,A,C).

disjunction(in(A),B,C):-
    disjunction_in(B,A,C).

disjunction_not_in(not_in(B),A,not_in(C)) :-
    set_bbbtree:set_bbbtree__intersect(A,B,C).
disjunction_not_in(in(B),A,not_in(C)):-
    set_bbbtree:set_bbbtree__difference(A,B,C).

disjunction_in(in(B),A,in(C)):-
    set_bbbtree:set_bbbtree__union(A,B,C).
disjunction_in(not_in(A),B,not_in(C)):-
    set_bbbtree:set_bbbtree__difference(A,B,C).

conjunction(not_in(A),B,C):-
    conjunction_not_in(B,A,C).

conjunction(in(A),B,C):-
    conjunction_in(B,A,C).

conjunction_not_in(not_in(B),A,not_in(C)):-
    set_bbbtree:set_bbbtree__union(A,B,C).
conjunction_not_in(in(B),A,in(C)):-
    set_bbbtree:set_bbbtree__non_empty(C),
    set_bbbtree:set_bbbtree__difference(B,A,C).

conjunction_in(in(B),A,in(C)):-
    set_bbbtree:set_bbbtree__non_empty(C),
    set_bbbtree:set_bbbtree__intersect(A,B,C).  
conjunction_in(not_in(A),B,in(C)):-
    set_bbbtree:set_bbbtree__non_empty(C),
    set_bbbtree:set_bbbtree__difference(B,A,C).

negation(in(P),not_in(P)).
negation(not_in(P),in(P)) :-
    set_bbbtree:set_bbbtree__non_empty(P).

display_predicate(in(Tree),Conj):-
    set_bbbtree:set_bbbtree__to_sorted_list(Tree,List),
    fsa_visualization:abbreviate_label0(List,[H|T]),
    display_predicate_in(T,H,Conj).

display_predicate(not_in(Tree),Not):-
    set_bbbtree:set_bbbtree__to_sorted_list(Tree,List),
    fsa_visualization:abbreviate_label0(List,List1),    
    display_predicate_not_in(List1,Not).

display_predicate_in([],H,H).
display_predicate_in([H|T],A,{Conj}):-
    list_to_conj([H|T],A,Conj).

display_predicate_not_in([],?).
display_predicate_not_in([H|T],? - Conj):-
    display_predicate_in(T,H,Conj).

list_to_conj([],A,A).
list_to_conj([H|T],A,(A,B)):-
    list_to_conj(T,H,B).

prepare_complement_of_set(_fa,_).

complement_of_set(Preds,_,Comps):-
    true(True),
    (	conjoin_negated(Preds,True,Comp)
    ->	Comps=[Comp]
    ;   Comps=[]
    ).

conjoin_negated([],Pred,Pred).
conjoin_negated([P0|Preds],P1,Pred) :-
    negation(P0,NotP0),
    conjunction(P1,NotP0,P2),
    conjoin_negated(Preds,P2,Pred).

determinize_preds(U0,U):-
    fsa_determinizer:determinize_preds(U0,U,fsa_preds_bbb).

t_determinize_preds(U0,U):-
    fsa_t_determinizer:t_determinize_preds(U0,U,fsa_preds_bbb).

%% identity
identity(Pred,'$@'(Pred)).

class_to_pred(List,[in(Tree)]):-
    set_bbbtree:set_bbbtree__init(Tree0),
    set_bbbtree:set_bbbtree__insert_list(Tree0,List,Tree).

negated_class_to_pred(List,[not_in(Tree)]):-
    set_bbbtree:set_bbbtree__init(Tree0),
    set_bbbtree:set_bbbtree__insert_list(Tree0,List,Tree).

cleanup([],[]).
cleanup([H|T],[Pred]) :-
    cleanup(T,H,Pred).

cleanup([],P,P).
cleanup([H|T],Pred0,Pred) :-
    disjunction(H,Pred0,Pred1),
    cleanup(T,Pred1,Pred).

