%%% Toy analysis of Dutch plurals
%%%
%%%  R. Malouf (malouf@let.rug.nl), 23 Apr 2000
%%%
%%% This defines two top-level transducers.  'lex1' uses a list of
%%% both regular and irregular nouns and can only handle nouns on that
%%% list.  'lex2' only uses a list of irregular nouns, and guesses about
%%% the stems of regular nouns. 
%%%  
%%% To build and test:
%%%
%%%   fsa -aux twolevel.pl -aux dutch.pl -r 'inverse(lex1)'  >dutch.fsa
%%%   fsa -ta dutch.fsa <dutch.in >dutch.test
%%%   diff -c dutch.out dutch.test

:- multifile macro/2.

%% set this to `on' if you have lots of memory; otherwise it's 
%% probably best to set it to off, for a slower but more memory-modest
%% implementation.
%% :- fsa_globals:fsa_global_set(regex_cache,on).

%% Some useful recognizers

macro(letter, a..z).
macro(vowel, {a,e,i,o,u}).
macro(cons, letter - vowel).
macro(longv, {[vowel, vowel], [i,j]}).
macro(morph, {'+N','+Sg','+Pl','^s','^d','^r','^n','^i','^a','^z'}).

%% Lexical transducers

macro(regular,
  words([auto,bakker,bezem,boer,boot,brief,catalogus,deken,firma,fles,foto,
         gymnasium,historicus,krant,leraar,museum,museum,musicus,paraplu,
         prijs,provincie,tafel,woord])).

macro(plural_s, 
  words([garage,leraar,museum,oom,roman,tram,zoon])).

macro(irreg_plural,
  words([bad:baad,been,blad:blaad,dag:daag,dak:daak,gat:gaat,glas:glaas,
         leraar,schip:scheep,stad:steed,weg:weeg])).

macro(plural_eren,
  words([been:beend,blad,ei,kind,lied,volk])).

macro(other,
  words([motor,professor])).

%% Word structure 

% We define here three lexical transducers: one has just the irregular
% forms, the second has both irregular and listed regular forms, and the
% third allows anything to be a noun

macro(nn, ['+N', {'+Sg','+Pl'}]).
macro(nls, ['+N', '+Sg']).
macro(nlp, ['+N', '+Pl']).
macro(ns, [[]:'^s', nn]).  % ^s marks an unpredictable plural in -s

macro(irreg_lexicon, {
  % Nouns with an irregular plural
  [domain(irreg_plural), nls],
  [irreg_plural, []:'^n', nlp], % '^n' blocks the addition of different morph
  % Nouns with a plural in -s
  [plural_s, ns],
  % Nouns with a plural in -eren
  [domain(plural_eren), nls],
  [plural_eren, []:'^r', nlp], % ^r mark a plural in -eren
  % Other nouns
  [other, []:'^z', ([]:'^s')^, nn]}).

macro(full_lexicon, {[regular, nn], irreg_lexicon}).

% Some general phonotactic constraints would be helpful here

macro(anything, [letter+, nn]).

%% Select an ending

macro(pairs3, {letter, morph-'^n', 0:'^s', 0:'^i', 0:'^a', '^n':0}).

macro(reg_s_plural,
  comp(pairs3, 0:'^s', [cons, words([el,em,en,er,aar,ie,o,a,u])], ['+N'])).

macro(reg_i_plural,
  comp(pairs3, 0:'^i', [cons, word(us)], ['+N'])).

macro(reg_a_plural,
  comp(pairs3, 0:'^a', [word(um)], ['+N'])).

macro(select_morph, 
  twolevel(pairs3, [reg_s_plural,reg_i_plural,reg_a_plural])).

%% Rule transducer

macro(defpairs, 
  {a..z,vowel x '^x',s:z,f:v,0:'^d',0:'\'','^a':0,'^i':0,'^s':0,'+N':0,'+Sg':0,'+Pl':s,'+Pl':a,u:0,s:0,m:0,'+Pl':i,'^z':0}).

macro(pl1, {'+Pl':e, 0:n}).
macro(pl2, {'^r':e, 0:r}).

macro(pairs, {defpairs, pl1, pl2}).

macro(s_plural, 
  comp(pairs, '+Pl':s, ['^s':0, (0:'\'')^, '+N' o pairs], [])).

macro(a_plural, 
  comp(pairs, '+Pl':a, ['^a':0,'+N' o pairs], [])).

macro(i_plural, 
  comp(pairs, '+Pl':i, ['^i':0,'+N' o pairs], [])).

macro(latin_plural1, 
  comp(pairs, u:0, [], [{s,m} o pairs,{'^a','^i'} x 0,'+N' o pairs,'+Pl' o pairs])).

macro(latin_plural2, 
  comp(pairs, {s,m}:0, [u:0], [{'^a','^i'} x 0,'+N' o pairs,'+Pl' o pairs])).

macro(en_plural1,
  comp(pairs, '+Pl':e, [], [0:n])).

macro(en_plural2,
  opt(pairs, 0:n, ['+Pl':e], [])). 

%macro(eren_plural, 
%  comp(pairs, '+Pl':word(eren), ['^r':0,'+N' o pairs], [])).

macro(eren_plural1,
   comp(pairs, [ '^r':e ], [], [ 0:r, '+N' o pairs])).

macro(eren_plural2,
   comp(pairs, [ 0:r ], ['^r':e ], [ '+N' o pairs])).

macro(apostrophe,
  comp(pairs, 0:'\'', [{a,o,u}, '^s':0], ['+N' o pairs, '+Pl':s])).

% '^d' marks a consonant to be doubled

macro(short,
  comp(pairs, 0:'^d', [cons x cons, vowel x vowel, cons x cons], 
                      ['+N' o pairs, pairs o vowel, pairs o cons])).

% '^x' is something to be deleted

macro(long,
 var(V, vowel, 
    comp(pairs, V:'^x', [pairs o cons, V], 
                 [pairs o cons, '+N' o pairs, pairs o vowel, pairs o cons]))).

macro(voice,
  comp(pairs, {s,f} x {z,v}, [longv o [pairs, pairs]], 
                             ['+N' o pairs, pairs o vowel, pairs o cons])).

%macro(rules, 
%  twolevel(pairs, [a_plural,i_plural,latin_plural1,latin_plural2,s_plural,
%                   eren_plural,apostrophe,long,short,voice])).

macro(rules, 
  twolevel(pairs, [a_plural,i_plural,latin_plural1,latin_plural2,s_plural,
                   en_plural1,en_plural2,eren_plural1,eren_plural2,apostrophe,long,short,voice])).

%% Doubler (this stuff is separate to avoid conflicts)

macro(pairs2, {letter, '\'', '^d' x cons, '^x':0}).

macro(double, 
  var(C, cons, comp(pairs2, '^d':C, [C], []))).

macro(rules2, twolevel(pairs2, [double])).

%% The whole thing

macro(all_rules, select_morph o rules o rules2).

macro(irregular, irreg_lexicon o all_rules).
macro(guess, anything o all_rules).

macro(lex1, full_lexicon o all_rules).
macro(lex2, {irregular, guess o ~range(irregular)}).
