%% (c) Gertjan van Noord, 2000; GPL;  macro's for same-length transducers.

:- multifile macro/2.
:- multifile rx/2.

:- discontiguous macro/2.
:- discontiguous rx/2.
:- discontiguous trans_reachables/4.
:- discontiguous is_final_reachable/4.

:- op(800, fx, lt).
lt Call :-
    fsa_regex:ltime(2,fsa_regex_aux:Call).

%% the interesting operators are:
%%
%% sl_transducer/1: constructs equivalent transducer for the transduction
%% specified by its argument, but ensures that each of the transitions has
%% a single predicate on the input and on the output side.
%%
%% sl_intersection/2: constructs transducer for the intersection of the
%% same-length transductions specified as its arguments.
%%
%% sl_complement/1: constructs transducer for the complement of
%% the same-length transduction specified by its argument. 
%% NB: this is not treated in full generality. In particular, an exception
%% will be raised in case a transition occurs in the input in which an
%% identity is expressed over a predicate which is true of an unbounded
%% number of symbols. An example is the expression sl_complement(?).
%%
%% sl_complement/2: sl_complement(Pairs,Expr) is equivalent to the expression
%%                  sl_intersection(Pairs*,sl_complement(Expr))
%% but more efficient (the idea is that the first argument, Pairs, describes
%% the set of all possible pairs from which we subtract the pairs specified
%% by Expr. It's more efficient since in the construction of sl_complement
%% we can ignore pairs of predicates which are incompatible with any of
%% the pairs of predicates occurring in Pairs.

%% don't use cleanup here!
macro(mh(L),
      melt(mh0(coaccessible(determinize0(freeze(L,Flag)))),Flag)).


%% constructs same-length transducer for X, if possible. Raises
%% exception if it's not a same-length transduction.
macro(sl_transducer(X),sl_transducer0(letter_transducer(X,silent))).

rx(sl_transducer0(X),Fa) :-
    fsa_regex:rx(X,Fa0),
    user:catch(fsa_regex_aux:check_sl(Fa0,Fa),
	       not_same_length,
	       user:throw(fsa_error(
                 "Not a same-length transduction: ~w~n",[X]))
	       ).

check_sl(Fa0,Fa) :-
    fsa_data:fsa_jumps(Fa0,Jumps),
    check_sl(Jumps,Fa0,Fa).

check_sl([_|_],Fa0,Fa) :-
    fsa_regex:rx(efree(fa(Fa0)),Fa1),
    fsa_data:fsa_transitions(Fa1,Trans),
    check_sl_trans(Trans,Fa1,Fa).
check_sl([],Fa0,Fa) :-
    fsa_data:fsa_transitions(Fa0,Trans),
    check_sl_trans(Trans,Fa0,Fa).

check_sl_trans(Trans,Fa0,Fa) :-
    (	check_sl_trans(Trans)
    ->	Fa0=Fa
    ;   make_sl(Fa0,Fa)
    ).

check_sl_trans([]).
check_sl_trans([trans(_,L/R,_)|T]) :-
    L \== [],
    R \== [],
    check_sl_trans(T).

make_sl(Fa0,Fa) :-
    fsa_regex:all_components_reachable(fsa_regex_aux:sl_transducer,Fa0,Fa).

prepare_reachable(sl_transducer,Fa0,Fa0,StartStates,Sigma):-
    fsa_data:fsa_symbols_decl(Fa0,Sigma),
    fsa_data:fsa_start_states(Fa0,StartStates).

is_final_reachable(sl_transducer,Fa0,P,Bool) :-
    (	fsa_data:fsa_final_state(Fa0,P)
    ->	Bool=yes
    ;	Bool=no
    ).

trans_reachables(sl_transducer,P,Fa0,Transitions) :-
    integer(P),
    findall(trans(P,Pair,Q),sl_trans_reachable(P,Fa0,Pair,Q),Transitions).

trans_reachables(sl_transducer,p(Ls,Rs,Q),_Fa0,[trans(p(Ls,Rs,Q),Pair,NQ)]) :-
    create_transition(Ls,Rs,Q,Pair,NQ).

sl_trans_reachable(P,Fa,Pair,Q) :-
    sl_path(P,[],0,[],0,Fa,Pair,Q,[]).


%% sl_path(P,Left,LeftL,Right,RightL,Fa,Pair,Q,History)

create_transition([L],[R],P,L/R,P).
create_transition([L,L1|Ls],[R,R1|Rs],P,L/R,p([L1|Ls],[R1|Rs],P)).

%% case 1: we have a same-length non-empty path
sl_path(P,Left,LeftL,Right,RightL,_Fa,Pair,Q,_):-
    LeftL =:= RightL,
    LeftL > 0,
    !,
    create_transition(Left,Right,P,Pair,Q).

%% case 2: a non-same-length path.
%% Check for cycles and for final state
sl_path(P0,Left0,LeftL0,Right0,RightL0,Fa,Pair,Q,History) :-
    (	lists:member(P0,History)
    ->	user:throw(not_same_length)
    ;	true
    ),
    (	fsa_data:fsa_final_state(Fa,P0),
	(   LeftL0 > 0
	;   RightL0 > 0
	)
    ->	user:throw(not_same_length)
    ;	true
    ),
    fsa_data:fsa_transition(Fa,P0,L/R,P),
    (	L==[]
    ->	Left0=Left, LeftL0=LeftL
    ;	lists:append(Left0,[L],Left),
	LeftL is +(LeftL0,1)
    ),
    (	R==[]
    ->	Right0=Right, RightL0=RightL
    ;	lists:append(Right0,[R],Right),
	RightL is +(RightL0,1)
    ),
    sl_path(P,Left,LeftL,Right,RightL,Fa,Pair,Q,[P0|History]).



%% X and Y are same-length transducers. This yields the same-length
%% intersection, taking into account the predicates on both side of
%% the symbol pair. 
macro(sl_intersection(X,Y),
      cleanup(m(sl_intersection0(sl_transducer(X),
				 sl_transducer(Y))))
     ).

rx(sl_intersection0(X,Y),Fa):-
    fsa_regex:rx(X,FaX),
    fsa_regex:rx(Y,FaY),
    sl_intersect_fa(FaX,FaY,Fa).

sl_intersect_fa(FaA,FaB,Fa):-
    fsa_regex:all_components_reachable(fsa_regex_aux:sl_intersect,FaA,FaB,Fa).

prepare_reachable(sl_intersect,FaA,FaB,
	     d(FinalsA,FinalsB,TransA,TransB,ModA,ModB),
		  StartStates,Sigma):-
    fsa_regex:coerce_sigma(FaA,FaB,Sigma),
    fsa_data:fsa_symbols_decl(FaA,t(ModA,ModB)),
    fsa_regex:pairs_of_start_states(FaA,FaB,StartStates),
    fsa_data:index_finals(FaA,FinalsA),
    fsa_data:index_finals(FaB,FinalsB),
    fsa_data:index_transitions(FaA,TransA),
    fsa_data:index_transitions(FaB,TransB).

trans_reachables(sl_intersect,Pa/Pb,d(_,_,Tra,TrB,ModA,ModB),Trans):-
    intersect_trans(Pa,Pb,Tra,TrB,ModA,ModB,Trans,[]).

sl_conjunction(M,M,'$@'(P)/'$@'(P),'$@'(Q)/'$@'(Q),S/S):-
    !,
    fsa_pred_module:conjunction(M,P,Q,R),
    fsa_pred_module:identity(M,R,S).

sl_conjunction(M,M,Qa/Qb,'$@'(P)/'$@'(P),S/S):-
    !,
    fsa_pred_module:conjunction(M,P,Qa,R0),
    fsa_pred_module:conjunction(M,R0,Qb,R),
    fsa_pred_module:identity(M,R,S).

sl_conjunction(M,M,'$@'(P)/'$@'(P),Qa/Qb,S/S):-
    !,
    fsa_pred_module:conjunction(M,P,Qa,R0),
    fsa_pred_module:conjunction(M,R0,Qb,R),
    fsa_pred_module:identity(M,R,S).

sl_conjunction(Ma,Mb,Pa/Pb,Qa/Qb,Sa/Sb):-
    fsa_pred_module:conjunction(Ma,Pa,Qa,Sa),
    fsa_pred_module:conjunction(Mb,Pb,Qb,Sb).

intersect_trans(Pa,Pb,TrA0,TrB0,ModA,ModB,Trans0,Trans) :-
    fsa_array:fsa_array_access(Pa,TrA,[],TrA0), 
    fsa_array:fsa_array_access(Pb,TrB,[],TrB0),
    trans_intersect(TrA,TrB,ModA,ModB,Pa/Pb,Trans0,Trans).

trans_intersect([],_,_,_,_,T,T).
trans_intersect([Sym-Qs|Tail],TrB,ModA,ModB,Src,Trans0,Trans):-
    trans_intersect0(TrB,ModA,ModB,Src,Sym,Qs,Trans0,Trans1),
    trans_intersect(Tail,TrB,ModA,ModB,Src,Trans1,Trans).

trans_intersect0([],_,_,_,_,_,T,T).
trans_intersect0([SymB-Qbs|Tail],ModA,ModB,Src,SymA,Qas,Trans0,Trans):-
    (	sl_conjunction(ModA,ModB,SymA,SymB,Sym)
    ->	trans_intersect1(Qas,Src,Sym,Qbs,Trans0,Trans1)
    ;   Trans0=Trans1
    ),
    trans_intersect0(Tail,ModA,ModB,Src,SymA,Qas,Trans1,Trans).

trans_intersect1([],_,_,_,T,T).
trans_intersect1([Qa|Qas],Src,Sym,Qbs,Trans0,Trans):-
    trans_intersect2(Qbs,Src,Sym,Qa,Trans0,Trans1),
    trans_intersect1(Qas,Src,Sym,Qbs,Trans1,Trans).

trans_intersect2([],_,_,_,T,T).
trans_intersect2([Qb|Qbs],Src,Sym,Qa,[trans(Src,Sym,Qa/Qb)|Trans0],Trans):-
    trans_intersect2(Qbs,Src,Sym,Qa,Trans0,Trans).

is_final_reachable(sl_intersect,d(FinalsA,FinalsB,_,_,_,_),A/B,Bool):-
    fsa_data:indexed_final(A,FinalsA,Bool1),
    fsa_data:indexed_final(B,FinalsB,Bool2),
    fsa_regex:and(Bool1,Bool2,Bool).

macro(sl_complement(E),sl_complement(? : ?,E)).

macro(sl_complement(Pairs,E),
      sl_intersection(
	Pairs*,
	swap_final_non_final(
	   cleanup(sl_complete0(
		    sl_transducer(Pairs),
		    m(expand_non_overlapping_predicates(sl_transducer(E)))
			       )
		  )
			    )
		     )
     ).

macro(sl_complete(A),
      cleanup(sl_complete0(sl_transducer(? : ?),sl_transducer(A)))).
macro(sl_complete(A,B),
      cleanup(sl_complete0(sl_transducer(A),sl_transducer(B)))).

rx(sl_complete0(Pairs,A),Fa) :-
    fsa_regex:rx(A,Fa0),
    fsa_data:fsa_symbols_decl(Fa0,t(PmA,PmB)),
    get_pairs(Pairs,Preds,PmA,PmB),
    fsa_globals:fsa_global_get(debug,Debug),
    fsa_regex:ltime(Debug,fsa_regex_aux:sl_complete_fa(Fa0,Fa,Preds)).

get_pairs(Pairs,Preds,PmA,PmB) :-
    fsa_regex:rx(Pairs,Fa),
    (	fsa_data:fsa_components(t(PmA,PmB),_,_,_,Trans,_,Fa)
    ->	findall(Sym,lists:member(trans(_,Sym,_),Trans),Preds0),
	sort(Preds0,Preds)
    ;   throw(fsa_error("sl_complete: improper `feasible pairs' declaration~n",
			[]))
    ).

sl_complete_fa(Fa0,Fa,Pairs):-
    fsa_data:fsa_copy_except(states,Fa0,Fa1,States0,States),
    fsa_data:fsa_copy_except(transitions,Fa1,Fa,Trans,CompletedTrans),
    States is +(States0,1),
    fsa_data:fsa_symbols_decl(Fa0,t(Ma,Mb)),
    complete_trans(0,Trans,CompletedTrans0,[],States0,Ma,Mb,Pairs),
    sort(CompletedTrans0,CompletedTrans).

complete_trans(State,Trans0,Ts0,Ts,Sink,Ma,Mb,Pairs) :-
    (	State > Sink
    ->	Ts0=Ts
    ;   complete_trans_states_symbols(State,Trans0,Trans,Ts0,Ts1,
				      Sink,Ma,Mb,Pairs),
	State2 is +(State,1),
	complete_trans(State2,Trans,Ts1,Ts,Sink,Ma,Mb,Pairs)
    ).

complete_trans_states_symbols(State,Trans0,Trans,Ts0,Ts,Sink,Ma,Mb,Pairs) :-
    get_all_preds(Trans0,Trans,Ts0,Ts1,State,Preds0),
    expand_ids(Preds0,Ma,Mb,Preds),
    complement_of_set(Preds,Ma,Mb,Complements,Pairs),
    complement_transitions(Complements,State,Sink,Ts1,Ts).

get_all_preds([],[],Ts,Ts,_,[]).
get_all_preds([trans(P,Pred,Q)|Trans0],Trans,Ts0,Ts,State,Preds) :-
    (	State==P
    ->	Preds=[Pred|Preds0],
	Ts0=[trans(P,Pred,Q)|Ts1],
	get_all_preds(Trans0,Trans,Ts1,Ts,State,Preds0)
    ;   Trans=[trans(P,Pred,Q)|Trans0],
	Ts0=Ts,
	Preds=[]
    ).

expand_ids([],_,_,[]).
expand_ids([Pred|Preds0],Ma,Mb,Preds) :-
    expand_id(Pred,Ma,Mb,Preds,Preds1),
    expand_ids(Preds0,Ma,Mb,Preds1).

expand_id('$@'(P)/'$@'(P),M,M,Preds0,Preds) :-
    !,
    findall(Q,expand_pred(P,M,Q),Preds0,Preds).
expand_id(Pred,_,_,[Pred|Preds],Preds).

expand_pred(P,M,Q/Q) :-
    fsa_pred_module:evaluate_predicate(M,P,Sym),
    (	var(Sym)
    ->	throw(fsa_error(
	  "sl_complement: evaluate_predicate yields variable: ~w~n",[P]))
    ;	true
    ),
    fsa_pred_module:regex_notation_to_predicate(M,Sym,Q).

complement_of_set([],_Ma,_Mb,Pairs,Pairs).
complement_of_set([H|T],Ma,Mb,Compls,Pairs) :-
    %% terms:term_hash([H|T],Index),
    (	%%fsa_regex_cache:cache(Index,cos([H|T],Ma,Mb,Pairs),Compls)
        fsa_regex:in_cache(on,_,cos([H|T],Ma,Mb,Pairs),Compls)
    ->	true
    ;	negation_of_each([H|T],Ma,Mb,Pairs,Compls0),
	sort(Compls0,Compls),
	%%assertz(fsa_regex_cache:cache(Index,cos([H|T],Ma,Mb,Pairs),Compls))
        fsa_regex:add_cache(on,_,cos([H|T],Ma,Mb,Pairs),Compls)
    ).

negation_of_each([H|T],Ma,Mb,Pairs,Compls) :-
    each_negated([H|T],Ma,Mb,Compls0),
    non_overlapping_preds([Pairs|Compls0],Compls1,Ma,Mb),
    sort(Compls1,Compls2),
    findall(Pred,each_member_conj(Compls2,Pred),Compls).

each_member_conj([H|T],A/B) :-
    each_member_conj_list(T,H,Lists),
    lists:member(Alist/Blist,Lists),
    lists:member(A,Alist),
    lists:member(B,Blist).

each_member_conj_list([],Lists,Lists).
each_member_conj_list([H1|Tail],H0,Lists) :-
    each_each(H0,H1,H20),
    sort(H20,H2),
    each_member_conj_list(Tail,H2,Lists).

each_each([],_,[]).
each_each([H|T],List0,List) :-
    each_each_(List0,H,List,List1),
    each_each(T,List0,List1).

each_each_([],_,L,L).
each_each_([H2|T],H1,L0,L):-
    each_each__(H1,H2,L0,L1),
    each_each_(T,H1,L1,L).

each_each__(Alist0/Blist0,Alist1/Blist1,L0,L):-
    ordsets:ord_intersection(Alist0,Alist1,Alist),
    ordsets:ord_intersection(Blist0,Blist1,Blist),
    each_each___(Alist,Blist,L0,L).

each_each___([],_,L,L).
each_each___([H|T],Blist,L0,L) :-
    each_each____(Blist,[H|T],L0,L).

each_each____([],_,L,L).
each_each____([B|Bs],As,[As/[B|Bs]|L],L).

pairs_to_lists([],[]).
pairs_to_lists([P/Q|T0],[A/B|T]) :-
    (	P = '$@'(P1)
    ->	A=P1, B=P1
    ;   A=P, B=Q
    ),
    pairs_to_lists(T0,T).

check_pair(Pred,Pairs,Ma,Mb) :-
    lists:member(Pred1,Pairs),
    sl_conjunction(Ma,Mb,Pred,Pred1,_),
    !.

non_overlapping_preds(LList0,LList,Ma,Mb) :-
    collect_preds(LList0,PredsA0,PredsB0),
    sort(PredsA0,PredsA1),
    sort(PredsB0,PredsB1),
    fsa_cleanup:get_small_predicates(PredsA1,Ma,PredsA2),
    fsa_cleanup:get_small_predicates(PredsB1,Mb,PredsB2),
    fsa_cleanup:get_hash(PredsA2,AllowedA),
    fsa_cleanup:get_hash(PredsB2,AllowedB),
    expand_all(LList0,AllowedA,AllowedB,LList).

expand_all([],_,_,[]).
expand_all([H|T],As,Bs,[NH|NT]) :-
    expand_all1(H,As,Bs,NH),
    expand_all(T,As,Bs,NT).

expand_all1([],_,_,[]).
expand_all1([A/B|T0],As,Bs,[Alist/Blist|T]) :-
    fsa_hash:fsa_hash_access(A,Alist,[],As),
    fsa_hash:fsa_hash_access(B,Blist,[],Bs),
    expand_all1(T0,As,Bs,T).

collect_preds([],[],[]).
collect_preds([H|T],As,Bs) :-
    collect_preds_(H,As,As1,Bs,Bs1),
    collect_preds(T,As1,Bs1).

collect_preds_([],As,As,Bs,Bs).
collect_preds_([A/B|T],[A|As0],As,[B|Bs0],Bs) :-
    collect_preds_(T,As0,As,Bs0,Bs).
    
each_negated([],_,_,[]).
each_negated([H|T],A,B,[NH|NT]) :-
    findall(Neg,sl_negation(H,A,B,Neg),NH0),
    sort(NH0,NH),
    each_negated(T,A,B,NT).

sl_negation(P/_,A,B,NotP/True) :-
    fsa_pred_module:negation(A,P,NotP),
    fsa_pred_module:true(B,True).
sl_negation(_/Q,A,B,True/NotQ) :-
    fsa_pred_module:negation(B,Q,NotQ),
    fsa_pred_module:true(A,True).

complement_transitions([],_,_,Ts,Ts).
complement_transitions([C|Cs],P,Q,[trans(P,C,Q)|Ts0],Ts):-
    complement_transitions(Cs,P,Q,Ts0,Ts).


