%% (c) Gertjan van Noord
%% Arrays and hashes, implemented using O'Keefe's n+k trees.
:- module(fsa_array, [%% non-updatable arrays
		      fsa_array_new/1,
		      fsa_array_new/2,
		      fsa_array_access/3,
		      fsa_array_access/4,
		      fsa_array_get/3,
		      fsa_array_to_list/2
		     ]).


#swi# :- style_check([-atom]).

:- discontiguous help_info/4.
:- public help_info/4.

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

user:help_info(module,fsa_array,"Non-updatable Arrays (127+32 trees)",
"This module provides a non-updatable array data-structure. Accessing \
 \n individual items in the array is very efficient. The arrays are implemented \
 \n using O'Keefe's N+K trees, with N=127 and K=32.\
 \n \
 \n NB. Array indices start at 0: so 0 refers to the first element of\
 \n the array.\
 \n \
 \n Here's an overview of the predicates provided:  \
 \n \
 \n * fsa_array_new/[1,2]         create a new non-updatable array\
 \n * fsa_array_access/[3,4]      access a value in a non-updatable array\
 \n * fsa_array_get/3             get a value in a non-updateble array\
 \n * fsa_array_to_list/2         conversion of array -> list\
 \n \
 \n The N+K tree data-structure is described in The Craft of Prolog, by \
 \n Richard A. O'Keefe, MIT Press, 1990, chapters 4.5. and 4.6. It is\
 \n similar to the tries described in section 2 of Tarjan and Yao, Storing\
 \n a Sparse Table, CACM, 1979, 606-611; also refer to Knuth, The Art of\
 \n Computing Vol 3. ").


help_info(class,pred,"List of Predicates",
"This section lists the predicates defined by this module.").

help_info(pred,fsa_array_new,"fsa_array_new(-FsaArray[,?Size])",
"Initializes FsaArray as a new array. In this implementation of\
 \n arrays the optional second argument is not used."). 

fsa_array_new(_).

fsa_array_new(_,_).

help_info(pred,fsa_array_access,
"fsa_array_access(+Index,?Val[,?Default],+FsaArray)",
"Val is unified with the Index'th entry of FsaArray. This predicate\
 \n thus subsumes setting and reading of a value in the array. Remember\
 \n that you can't change values of an array (except by further instantiation).\
 \n For the 4-ary form, if the Index'th entry was not yet defined, then Val \
 \n is unified with Default (and not with the Index'th entry)."). 
fsa_array_access(Index0,Val,Thing):-
    Index is Index0+1,
    fsa_array_access1(Index,Val,Thing).

fsa_array_access(Index0,Val,Default,Thing):-
    Index is Index0+1,
    fsa_array_access1(Index,Val0,Thing),
    (	var(Val0)
    ->	Val=Default
    ;   Val=Val0
    ).

%% fsa_array_access(+In,?Val,?Tree)
fsa_array_access1(Index,Val,Tree):-
    functor(Tree,' nk ',127),
    (	Index < 96
    ->	arg(Index,Tree,Val)
    ;   K is (Index /\ 31)+96,    %% mod
	J is Index >> 5,          %% divide by 32
	arg(K,Tree,Son),
	fsa_array_access1(J,Val,Son)
    ).

help_info(pred,fsa_array_get,
"fsa_array_get(+Index,?Val,+FsaArray)",
"Val is unified with the Index'th entry of FsaArray. That entry must\
 \n not be variable. This predicate is different from fsa_array_access/3 in\
 \n that it can fail."). 
fsa_array_get(Index0,Val,Thing):-
    Index is Index0+1,
    fsa_array_get1(Index,Val,Thing).

%% fsa_array_access(+In,?Val,?Tree)
fsa_array_get1(Index,Val,Tree):-
    nonvar(Tree),
    (	Index < 96
    ->	arg(Index,Tree,Val),
	nonvar(Val)
    ;   K is (Index /\ 31)+96,    %% mod
	J is Index >> 5,          %% divide by 64
	arg(K,Tree,Son),
	fsa_array_get1(J,Val,Son)
    ).

:- multifile user:portray/1.

user:portray(X) :-
    functor(X,' nk ',127),
    !,
    write('<FsaArray>').

fsa_array_to_list(Thing,List):-
    fsa_array_to_list(Thing,List,[]).

fsa_array_to_list(Thing,List0,List):-
    (	var(Thing)
    ->	List0=List
    ;   functor(Thing,F,Arity),
	(   F == ' nk '
	->  fsa_array_to_list(Arity,Thing,List0,List)
	;   List0 = [Thing|List]
	)
    ).

fsa_array_to_list(Arity,Thing,List0,List):-
    (	Arity =:= 0
    ->	List0=List
    ;   arg(Arity,Thing,Arg),
	fsa_array_to_list(Arg,List0,List1),
	Arity1 is Arity-1,
	fsa_array_to_list(Arity1,Thing,List1,List)
    ).
