:- module(fsa_cleanup, []).

#!swi#:- use_module(library(terms), [ term_hash/2 ]).
#swi#:- use_module(terms, [ term_hash/2 ]).
#yap# :- use_module(library(lists),   [ member/2, append/3, select/3 ]).
#sic# :- use_module(library(lists),   [ select/3 ]).
#!swi# :- use_module(library(ordsets), [ ord_union/3 ]).
#swi# :- use_module(ordsets, [ ord_union/3 ]).

#!yap# :- dynamic fsa_regex_cache:mclt/8.
#!yap# :- dynamic fsa_regex_cache:mclwt/8.

cleanup_fa(Fa0,Fa):-
    fsa_data:fsa_type(Fa0,Type),
    cleanup_fa(Type,Fa0,Fa).

cleanup_fa(transducer(Sub),Fa0,Fa):-
    cleanup_fa_transducer(Sub,Fa0,Fa).

cleanup_fa(w_transducer(Sub),Fa0,Fa):-
    cleanup_fa_w_transducer(Sub,Fa0,Fa).

%% replace transitions trans(p,t1,q) and trans(p,t2,q)
%% to trans(p,t1 or t2,q)
cleanup_fa(recognizer,Fa0,Fa):-
    fsa_data:fsa_copy_except(transitions,Fa0,Fa,Trans0,Trans),
    fsa_data:fsa_symbols_decl(Fa,r(Mod)),
    (	Mod == fsa_frozen
    ->	Fa0 =  Fa
    ;	cleanup_transitions(Trans0,Trans,Mod,r)
    ).

cleanup_fa(w_recognizer,Fa0,Fa):-
    fsa_data:fsa_copy_except(transitions,Fa0,Fa,Trans0,Trans),
    fsa_data:fsa_symbols_decl(Fa,r(Mod,_)),
    cleanup_transitions(Trans0,Trans,Mod,w).

cleanup_fa_transducer(letter,Fa0,Fa):-
    fsa_data:fsa_copy_except(transitions,Fa0,Fa,Trans0,Trans),
    fsa_data:fsa_symbols_decl(Fa,t(ModA,ModB)),
    cleanup_transitions(Trans0,Trans,ModA/ModB,t).

%% for sequence transducers we *ignore* identities, since we
%% don't know synchronisation
cleanup_fa_transducer(sequence,Fa0,Fa):-
    fsa_data:fsa_copy_except(transitions,Fa0,Fa,Trans0,Trans),
    fsa_data:fsa_symbols_decl(Fa,t(ModA,ModB)),
    cleanup_transitions(Trans0,Trans,ModA/ModB,st).

cleanup_fa_w_transducer(letter,Fa0,Fa):-
    fsa_data:fsa_copy_except(transitions,Fa0,Fa,Trans0,Trans),
    fsa_data:fsa_symbols_decl(Fa,t(ModA,ModB,fsa_frozen)),
    cleanup_transitions(Trans0,Trans,ModA/ModB,wt).

cleanup_fa_w_transducer(sequence,Fa0,Fa):-
    fsa_data:fsa_copy_except(transitions,Fa0,Fa,Trans0,Trans),
    fsa_data:fsa_symbols_decl(Fa,t(ModA,ModB,fsa_frozen)),
    cleanup_transitions(Trans0,Trans,ModA/ModB,wst).

%% collect all transitions for each pair of states, and apply cleanup
%% to that set of transitions. In this sense, cleanup applies locally.
cleanup_transitions(TransIn,Trans,Mod,Type) :-
    fsa_globals:fsa_global_get(cleanup_list_cache,Cache),
    collect_tr_per_pair(TransIn,Trans0,[],Mod,Type,Cache),
    sort(Trans0,Trans).

collect_tr_per_pair([],Tr,Tr,_,_,_).
collect_tr_per_pair([trans(P,S,Q)|T0],Trs0,Trs,Mod,Type,Cache) :-
    collect_tr_per_pair(T0,P,[[Q-[S]]],Trs0,Trs,Mod,Type,Cache).

collect_tr_per_pair([],P,Ppairs,Trs0,Trs,Mod,Type,Cache) :-
    keymerge(Ppairs,Kvs),
    cleanup_per_pair(Kvs,P,Trs0,Trs,Mod,Type,Cache).

collect_tr_per_pair([trans(P,S,Q)|Tail],P0,Ppairs,Trs0,Trs,Mod,Type,C) :-
    (	P0 == P
    ->	Ppairs1=[[Q-[S]]|Ppairs],
	Trs0=Trs1
    ;   Ppairs1=[[Q-[S]]],
	keymerge(Ppairs,Kvs),
	cleanup_per_pair(Kvs,P0,Trs0,Trs1,Mod,Type,C)
    ),
    collect_tr_per_pair(Tail,P,Ppairs1,Trs1,Trs,Mod,Type,C).


%% cleanup_per_pair(Kv0,P,T0,T,Mod,Type)
%% Kv0 is a key-value list where the key is a target state
%% and the value is a list of symbols. We must cleanup that
%% list. The result is T0-T. P is source, Mod is pred-module
%% information, Type is type of automaton.
cleanup_per_pair([],_,T,T,_,_Type,_Cache).
cleanup_per_pair([Q-List|Kv],P,T0,T,Mod,Type,Cache):-
    cleanup_list(Type,Mod,P,List,Q,T0,T1,Cache),
    cleanup_per_pair(Kv,P,T1,T,Mod,Type,Cache).

%% treat each of the cases in turn
:- discontiguous cleanup_list/8.

%%%%%%%%%%%%%%%%%
%% RECOGNIZERS %%
%%%%%%%%%%%%%%%%%

cleanup_list(r,Mod,P,List,Q,T0,T,_) :-
    fsa_pred_module:cleanup(Mod,List,NList),
    cleaned_up_tr(NList,P,Q,T0,T).

cleaned_up_tr([],_,_,T,T).
cleaned_up_tr([Pr|Prs],P,Q,[trans(P,Pr,Q)|T0],T):-
    cleaned_up_tr(Prs,P,Q,T0,T).

%%%%%%%%%%%%%%%%%%%%%%%%%%
%% WEIGHTED RECOGNIZERS %%
%%%%%%%%%%%%%%%%%%%%%%%%%%

cleanup_list(w,Mod,P,List0,Q,T0,T,_C) :-
    weight_sym_pairs(List0,List1,P,Q,T0,T1),
    keymerge(List1,List),
    cleanup_list_w(List,Mod,P,Q,T1,T).

weight_sym_pairs([],[],_,_,Tr,Tr).
weight_sym_pairs([S/W|T0],Kv0,P,Q,Tr0,Tr) :-
    (	S==[]
    ->	Tr0=[trans(P,[]/W,Q)|Tr1],
	Kv0=Kv1
    ;   Kv0=[[W-[S]]|Kv1],
	Tr0=Tr1
    ),
    weight_sym_pairs(T0,Kv1,P,Q,Tr1,Tr).

cleanup_list_w([],_,_,_,T,T).
cleanup_list_w([W-List0|T0],Mod,P,Q,Tr0,Tr):-
    fsa_pred_module:cleanup(Mod,List0,List),
    cleaned_up_tr(List,P,W,Q,Tr0,Tr1),
    cleanup_list_w(T0,Mod,P,Q,Tr1,Tr).

cleaned_up_tr([],_,_,_,T,T).
cleaned_up_tr([Pr|Prs],P,W,Q,[trans(P,Pr/W,Q)|T0],T):-
    cleaned_up_tr(Prs,P,W,Q,T0,T).

%%%%%%%%%%%%%%%%%%%%%%%%
%% LETTER TRANSDUCERS %%
%%%%%%%%%%%%%%%%%%%%%%%%

cleanup_list(t,Ms,P,List0,Q,Tr0,Tr,Cache) :-
    cleanup_list_t(Cache,Ms,P,List0,Q,Tr0,Tr).

cleanup_list_t(off,Ma/Mb,P,List0,Q,Tr0,Tr) :-
    no_cache_cleanup_list_t(Ma,Mb,P0,List0,Q0,Tr0,Tr),
    P=P0, Q=Q0.

cleanup_list_t(on,Ma/Mb,P,List0,Q,Tr0,Tr) :-    
    cache_cleanup_list_t(Ma,Mb,P0,List0,Q0,Tr0,Tr),
    P=P0, Q=Q0.

cache_cleanup_list_t(Ma,Mb,P,List0,Q,Tr0,Tr) :-
    term_hash(mclt(Ma,Mb,List0),Index),
#!yap#     (   fsa_regex_cache:mclt(Index,Ma,Mb,List0,P,Q,Tr0,Tr)
#yap#     (   recorded(fsa_regex_cache,mclt(Index,Ma,Mb,List0,P,Q,Tr0,Tr),_)
    ->	true
    ;	cleanup_transitions_t(List0,List,Ma,Mb),
	list_to_trans(List,P,Q,Tr0,Tr),
#!yap#         assertz(fsa_regex_cache:mclt(Index,Ma,Mb,List0,P,Q,Tr0,Tr))
#yap#         catch(recordz(fsa_regex_cache,mclt(Index,Ma,Mb,List0,P,Q,Tr0,Tr),_),P,format(user_error,"Caching disabled for large term (~w)~n",[P]))
    ).

no_cache_cleanup_list_t(Ma,Mb,P,List0,Q,Tr0,Tr) :-
   cleanup_transitions_t(List0,List,Ma,Mb),
   list_to_trans(List,P,Q,Tr0,Tr).

list_to_trans([],_,_,Tr,Tr).
list_to_trans([Sym|Syms],P,Q,[trans(P,Sym,Q)|Tr0],Tr):-
    list_to_trans(Syms,P,Q,Tr0,Tr).


%% prefer to combine identities first
cleanup_transitions_t(List0,List,ModA,ModB):-
    expand_list_t(List0,List1,ModA,ModB),
    sort(List1,List2),
    (	ModA==ModB
    ->	cleanup_trans_id(List2,ModA,List3)
    ;	List1=List2
    ),
    all_cleanup_trans_dom(List3,ModA,List4),
    all_cleanup_trans_ran(List4,ModB,List5),
    sort(List5,List).

/*
%% prefer to combine domain/range parts first
cleanup_transitions_t(List0,List,ModA,ModB):-
    expand_list_t(List0,List1,ModA,ModB),
    sort(List1,List2),
    all_cleanup_trans_dom(List2,ModA,List3),
    all_cleanup_trans_ran(List3,ModB,List4),
    (	ModA==ModB
    ->	cleanup_trans_id(List4,ModA,List5)
    ;	List4=List5
    ),
    sort(List5,List).
*/

cleanup_trans_id(Pairs0,Mod,Pairs) :-
    ident_pairs(Pairs0,Ident0,Pairs,Pairs1,Mod),
    fsa_pred_module:cleanup(Mod,Ident0,Ident),
    add_idents(Ident,Mod,Pairs1).

ident_pairs([],[],NI,NI,_M).
ident_pairs([D/R|T],Ident,NonIdent0,NonIdent,Mod):- 
    (	D==R
    ->	(   D='$@'(D0)
	->  Ident=[D0|OT],
	    NonIdent0=NonIdent1
	;   fsa_pred_module:identity(Mod,D,D1),
	    (	D==D1
	    ->	Ident=[D|OT],
		NonIdent0=NonIdent1
	    ;	NonIdent0=[D/R|NonIdent1],
		Ident=OT
	    )
	)
    ;   NonIdent0=[D/R|NonIdent1],
	OT=Ident
    ),
    ident_pairs(T,OT,NonIdent1,NonIdent,Mod).

add_idents([],_,[]).
add_idents([Sym0|T0],Mod,[Sym1/Sym1|T]) :-
    fsa_pred_module:identity(Mod,Sym0,Sym1),
    add_idents(T0,Mod,T).

expand_list_t([],[],_,_).
expand_list_t([H|T],List,Ma,Mb) :-
    expand_list_t_(T,H,List,Ma,Mb).

expand_list_t_([],H,[H],_,_).
expand_list_t_([List0H|List0T],H,List,Ma,Mb) :-
    Ma \== Mb, !,
    list_of_predicates_t([H|[List0H|List0T]],ListA00,ListB00),
    sort(ListA00,ListA0),
    get_small_predicates(ListA0,Ma,ListA1),
    get_hash(ListA1,AllowedA),
    sort(ListB00,ListB0),
    get_small_predicates(ListB0,Mb,ListB1),
    get_hash(ListB1,AllowedB),
    expand_non_overlapping_t([H|[List0H|List0T]],List,AllowedA,AllowedB).
expand_list_t_([List0H|List0T],H,List,M,M) :-
    list_of_predicates_ti([H|[List0H|List0T]],ListP0),
    sort(ListP0,ListP1),
    get_small_predicates(ListP1,M,ListP),
    get_hash(ListP,AllowedP),
    expand_non_overlapping_ti([H|[List0H|List0T]],List,AllowedP,M).

expand_non_overlapping_t([],[],_,_).
expand_non_overlapping_t([L/R|T0],List,AllowedA,AllowedB) :-
    findall(Sym,expand_non_overlap(L,AllowedA,R,AllowedB,Sym),List,List1),
    expand_non_overlapping_t(T0,List1,AllowedA,AllowedB).

expand_non_overlapping_ti([],[],_,_).
expand_non_overlapping_ti([L/R|T0],List,Allowed,Pm) :-
    findall(Sym,expand_i(L/R,Allowed,Sym,Pm),List,List1),
    expand_non_overlapping_ti(T0,List1,Allowed,Pm).

list_of_predicates_t([],[],[]).
list_of_predicates_t([A/B|T],As0,Bs0):-
    (	A == []
    ->	As0 = As
    ;   As0 = [A|As]
    ),
    (	B == []
    ->	Bs0 = Bs
    ;   Bs0 = [B|Bs]
    ),
    list_of_predicates_t(T,As,Bs).

list_of_predicates_ti([],[]).
list_of_predicates_ti([P0a/P0b|Ps0],Ps1):-
    (	P0a = '$@'(P0),
	P0b = '$@'(P0)
    ->	Ps1=[P0|Ps]
    ;	(   P0a == []
	->  Ps1=Ps2
	;   Ps1=[P0a|Ps2]
	),
	(   P0b == []
	->  Ps2=Ps
	;   Ps2=[P0b|Ps]
	)
    ),
    list_of_predicates_ti(Ps0,Ps).

all_cleanup_trans_dom(List0,Mod,List) :-
    output_sym_pairs(List0,Kv0,Rems,[]),
    keymerge(Kv0,Kv),
    cleanup_list_t(Kv,Mod,List1),
    back_to_list_t(List1,List,Rems).

output_sym_pairs([],[],Tr,Tr).
output_sym_pairs([S/W|T0],Kv0,Tr0,Tr) :-
    (	S==[]
    ->	Tr0=[[]/W|Tr1],
	Kv0=Kv1
    ;   Kv0=[[W-[S]]|Kv1],
	Tr0=Tr1
    ),
    output_sym_pairs(T0,Kv1,Tr1,Tr).

back_to_list_t([],L,L).
back_to_list_t([R-Dlist|Tail],L0,L) :-
    back_to_list_t_(Dlist,R,L0,L1),
    back_to_list_t(Tail,L1,L).

back_to_list_t_([],_,L,L).
back_to_list_t_([H|T],R,[H/R|L0],L):-
    back_to_list_t_(T,R,L0,L).

all_cleanup_trans_ran(List0,Mod,List) :-
    inv_output_sym_pairs(List0,Kv0,Rems,[]),
    keymerge(Kv0,Kv),
    cleanup_list_t(Kv,Mod,List1),
    inv_back_to_list_t(List1,List,Rems).

inv_output_sym_pairs([],[],Tr,Tr).
inv_output_sym_pairs([S/W|T0],Kv0,Tr0,Tr) :-
    (	W==[]
    ->	Tr0=[S/W|Tr1],
	Kv0=Kv1
    ;   Kv0=[[S-[W]]|Kv1],
	Tr0=Tr1
    ),
    inv_output_sym_pairs(T0,Kv1,Tr1,Tr).

cleanup_list_t([],_,[]).
cleanup_list_t([R-Dlist0|Tail0],Mod,[R-Dlist|Tail]):-
    (	R = '$@'(_)
    ->	Dlist0=Dlist
    ;	fsa_pred_module:cleanup(Mod,Dlist0,Dlist)
    ),
    cleanup_list_t(Tail0,Mod,Tail).

inv_back_to_list_t([],L,L).
inv_back_to_list_t([R-Dlist|Tail],L0,L) :-
    inv_back_to_list_t_(Dlist,R,L0,L1),
    inv_back_to_list_t(Tail,L1,L).

inv_back_to_list_t_([],_,L,L).
inv_back_to_list_t_([H|T],R,[R/H|L0],L):-
    inv_back_to_list_t_(T,R,L0,L).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% WEIGHTED LETTER TRANSDUCERS %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

cleanup_list(wt,Ms,P,List0,Q,Tr0,Tr,Cache) :-
    cleanup_list_wt(Cache,Ms,P,List0,Q,Tr0,Tr).

cleanup_list_wt(off,Ma/Mb,P,List0,Q,Tr0,Tr) :-
    no_cache_cleanup_list_wt(Ma,Mb,P0,List0,Q0,Tr0,Tr),
    P=P0,  Q=Q0.

cleanup_list_wt(on,Ma/Mb,P,List0,Q,Tr0,Tr) :-
    cache_cleanup_list_wt(Ma,Mb,P0,List0,Q0,Tr0,Tr),
    P=P0,  Q=Q0.

cache_cleanup_list_wt(Ma,Mb,P,List0,Q,Tr0,Tr) :-
    term_hash(mclwt(Ma,Mb,List0),Index),
#!yap#     (   fsa_regex_cache:mclwt(Index,Ma,Mb,List0,P,Q,Tr0,Tr)
#yap#     (   recorded(fsa_regex_cache,mclwt(Index,Ma,Mb,List0,P,Q,Tr0,Tr),_)
    ->	true
    ;	cleanup_transitions_wt(List0,List,Ma,Mb),
	list_to_trans(List,P,Q,Tr0,Tr),
#!yap#         assertz(fsa_regex_cache:mclwt(Index,Ma,Mb,List0,P,Q,Tr0,Tr))
#yap#         catch(recordz(fsa_regex_cache,mclwt(Index,Ma,Mb,List0,P,Q,Tr0,Tr),_),P,format(user_error,"Caching disabled for large term (~w)~n",[P]))
    ).

%% in yap, caching is buggy
no_cache_cleanup_list_wt(Ma,Mb,P,List0,Q,Tr0,Tr) :-
    cleanup_transitions_wt(List0,List,Ma,Mb),
    list_to_trans(List,P,Q,Tr0,Tr).

/*
cleanup_transitions_wt(List0,List,ModA,ModB):-
    expand_list_wt(List0,List1,ModA,ModB),
    sort(List1,List2),
    all_cleanup_trans_dom(List2,ModA,List3),
    findall(Dis,w_cleanup_trans_ran(List3,ModB,Dis),List4),
    (	ModA==ModB
    ->	findall(Dis,w_cleanup_trans_id(List4,ModA,Dis),List5)
    ;	List4=List5
    ),
    sort(List5,List).
*/
cleanup_transitions_wt(List0,List,ModA,ModB):-
    expand_list_wt(List0,List1,ModA,ModB),
    sort(List1,List2),
    (	ModA==ModB
    ->	findall(Dis,w_cleanup_trans_id(List2,ModA,Dis),List3)
    ;	List2=List3
    ),
    all_cleanup_trans_dom(List3,ModA,List4),
    findall(Dis,w_cleanup_trans_ran(List4,ModB,Dis),List5),
    sort(List5,List).

w_cleanup_trans_id(List,Mod,Sym):-
    setof(D/R,member(D/(R/W),List),Pairs),
    ident_pairs(Pairs,Ident,NonIdent,[],Mod),
    fsa_pred_module:cleanup(Mod,Ident,DisList),
    (	member(A/B,NonIdent),
	Sym=(A/(B/W))
    ;   member(Sym0,DisList),
	fsa_pred_module:identity(Mod,Sym0,Sym1),
	Sym=Sym1/(Sym1/W)
    ).

w_cleanup_trans_ran(List,_Mod,Domain/([]/W)) :-
    member(Domain/([]/W),List).
w_cleanup_trans_ran(List,_Mod,Domain/('$@'(Pred)/W)) :-
    member(Domain/('$@'(Pred)/W),List).
w_cleanup_trans_ran(List,Mod,Domain/(Dis/W)):-
    setof(Pred,w_cleanup_trans_ran_member(Domain,Pred,W,List),Preds),
    fsa_pred_module:cleanup(Mod,Preds,DisList),
    member(Dis,DisList).

w_cleanup_trans_ran_member(Domain,Pred,W,List) :-
    member(Domain/(Pred/W),List),
    Pred \== [],
    Pred \= '$@'(_).

expand_list_wt([],[],_,_).
expand_list_wt([H|T],List,Ma,Mb) :-
    expand_list_wt_(T,H,List,Ma,Mb).

expand_list_wt_([],H,[H],_,_).
expand_list_wt_([List0H|List0T],H,List,Ma,Mb) :-
    Ma \== Mb, !,
    list_of_predicates_wt([H|[List0H|List0T]],ListA00,ListB00),
    sort(ListA00,ListA0),
    get_small_predicates(ListA0,Ma,ListA1),
    get_hash(ListA1,AllowedA),
    sort(ListB00,ListB0),
    get_small_predicates(ListB0,Mb,ListB1),
    get_hash(ListB1,AllowedB),
    expand_non_overlapping_wt([H|[List0H|List0T]],List,AllowedA,AllowedB).
expand_list_wt_([List0H|List0T],H,List,M,M) :-
    list_of_predicates_wti([H|[List0H|List0T]],ListP0),
    sort(ListP0,ListP1),
    get_small_predicates(ListP1,M,ListP),
    get_hash(ListP,AllowedP),
    expand_non_overlapping_wti([H|[List0H|List0T]],List,AllowedP,M).

expand_non_overlapping_wt([],[],_,_).
expand_non_overlapping_wt([L/(R/W)|T0],List,AllowedA,AllowedB) :-
    findall(Sa/(Sb/W),
	    expand_non_overlap(L,AllowedA,R,AllowedB,Sa/Sb),List,List1),
    expand_non_overlapping_wt(T0,List1,AllowedA,AllowedB).

expand_non_overlapping_wti([],[],_,_).
expand_non_overlapping_wti([L/(R/W)|T0],List,Allowed,Pm) :-
    findall(Sa/(Sb/W),expand_i(L/R,Allowed,Sa/Sb,Pm),List,List1),
    expand_non_overlapping_wti(T0,List1,Allowed,Pm).

list_of_predicates_wt([],[],[]).
list_of_predicates_wt([A/(B/_)|T],As0,Bs0):-
    (	A == []
    ->	As0 = As
    ;   As0 = [A|As]
    ),
    (	B == []
    ->	Bs0 = Bs
    ;   Bs0 = [B|Bs]
    ),
    list_of_predicates_wt(T,As,Bs).

list_of_predicates_wti([],[]).
list_of_predicates_wti([P0a/(P0b/_)|Ps0],Ps1):-
    (	P0a = '$@'(P0),
	P0b = '$@'(P0)
    ->	Ps1=[P0|Ps]
    ;	(   P0a == []
	->  Ps1=Ps2
	;   Ps1=[P0a|Ps2]
	),
	(   P0b == []
	->  Ps2=Ps
	;   Ps2=[P0b|Ps]
	)
    ),
    list_of_predicates_wti(Ps0,Ps).

%%%%%%%%%%%%%%%%%%%%%%%%%%
%% SEQUENCE TRANSDUCERS %%
%%%%%%%%%%%%%%%%%%%%%%%%%%

cleanup_list(st,Ma/Mb,P,List0,Q,Tr0,Tr,_C) :-
    cleanup_transitions_st(List0,List,Ma,Mb),
    list_to_trans(List,P,Q,Tr0,Tr).

cleanup_transitions_st(List0,List,ModA,ModB):-
    all_cleanup_trans_dom(List0,ModA,List1),
    findall(Dis,cleanup_trans_ran_list(List1,ModB,Dis),List).

cleanup_trans_ran_list(List0,Mod,Domain/Dis) :-
    setof(Pred,member(Domain/Pred,List0),Preds),
    simplify_pred_lists(Preds,SimplerPreds,Mod),
    member(Dis,SimplerPreds).

simplify_pred_lists(L0,L,Mod) :-
    (	simplify_pred_lists0(L0,L1,Mod)
    ->	simplify_pred_lists(L1,L,Mod)
    ;   L0=L
    ).

similar_with_hole([H1|T],[H2|T],[H|T],H1,H2,H).
similar_with_hole([H|T1],[H|T2],[H|T],H1,H2,H) :-
    H \= '$@'(_),
    similar_with_hole(T1,T2,T,H1,H2,H).

simplify_pred_lists0(List0,List,Mod) :-
    append(Pre,[One|Tail],List0),
    similar_with_hole(One,Two,New,H1,H2,H),
    select(Two,Tail,Tail2),
    fsa_pred_module:cleanup(Mod,[H1,H2],[H]),
    !,
    append(Pre,[New|Tail2],List).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% WEIGHTED SEQUENCE TRANSDUCERS %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

cleanup_list(wst,Ma/Mb,P,List0,Q,Tr0,Tr,_C) :-
    cleanup_transitions_wst(List0,List,Ma,Mb),
    list_to_trans(List,P,Q,Tr0,Tr).

cleanup_transitions_wst(List0,List,ModA,ModB):-
    all_cleanup_trans_dom(List0,ModA,List1),
    findall(Dis,w_cleanup_trans_ran_list(List1,ModB,Dis),List).

w_cleanup_trans_ran_list(List0,Mod,Domain/(Dis/W)) :-
    setof(Pred,member(Domain/(Pred/W),List0),Preds),
    simplify_pred_lists(Preds,SimplerPreds,Mod),
    member(Dis,SimplerPreds).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% EXPAND INTO NON OVERLAPPING PREDICATES %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

expand_non_overlapping_fa(recognizer,Fa0,Fa):-
    expand_non_overlapping_fa(Fa0,Fa).

expand_non_overlapping_fa(w_recognizer,Fa0,Fa):-
    expand_non_overlapping_fa(Fa0,Fa).

%% make a synchronized transducer first, otherwise it is too hard
%% to expand $@ predicates consistently on the domain and range side...
expand_non_overlapping_fa(transducer(Sub),Fa0,Fa):-
    fsa_regex:to_letter_if_sequence(Sub,silent,_,Fa0,Fa1),
    expand_non_overlapping_fa(Fa1,Fa).

%% make a synchronized transducer first, otherwise it is too hard
%% to expand $@ predicates consistently on the domain and range side...
expand_non_overlapping_fa(w_transducer(Sub),Fa0,Fa):-
    fsa_regex:w_to_letter_if_sequence(Sub,silent,_,Fa0,Fa1),
    expand_non_overlapping_fa(Fa1,Fa).

%% expand_non_overlapping_fa/2 (if transducer, then letter)
expand_non_overlapping_fa(Fa0,Fa):-
    fsa_data:fsa_copy_except(transitions,Fa0,Fa,Trans0,Trans),
    fsa_data:fsa_symbols_decl(Fa0,Sigma),
    collect_predicates(Sigma,Trans0,AllowedPreds),
    expand_non_overlapping(Trans0,Sigma,AllowedPreds,Trans1),
    sort(Trans1,Trans).

collect_predicates(r(Sigma),Trans,Allowed) :-
    list_of_predicates(Trans,List00),
    sort(List00,List0),
    get_small_predicates(List0,Sigma,List1),
    get_hash(List1,Allowed).

collect_predicates(r(Sigma,_),Trans,Allowed) :-
    list_of_predicates_d(Trans,List00),
    sort(List00,List0),
    get_small_predicates(List0,Sigma,List1),
    get_hash(List1,Allowed).

collect_predicates(t(SigmaA,SigmaB),Trans,AllowedA/AllowedB) :-
    SigmaA \== SigmaB,!,
    list_of_predicates_d(Trans,ListA00),
    list_of_predicates_r(Trans,ListB00),
    sort(ListA00,ListA0),
    get_small_predicates(ListA0,SigmaA,ListA1),
    get_hash(ListA1,AllowedA),
    sort(ListB00,ListB0),
    get_small_predicates(ListB0,SigmaB,ListB1),
    get_hash(ListB1,AllowedB).

collect_predicates(t(Sigma,Sigma),Trans,Allowed) :-
    list_of_predicates_i(Trans,List00),
    sort(List00,List0),
    get_small_predicates(List0,Sigma,List1),
    get_hash(List1,Allowed).

collect_predicates(t(SigmaA,SigmaB,_),Trans,AllowedA/AllowedB) :-
    SigmaA \== SigmaB,!,
    list_of_predicates_d(Trans,ListA00),
    list_of_predicates_wr(Trans,ListB00),
    sort(ListA00,ListA0),
    get_small_predicates(ListA0,SigmaA,ListA1),
    get_hash(ListA1,AllowedA),
    sort(ListB00,ListB0),
    get_small_predicates(ListB0,SigmaB,ListB1),
    get_hash(ListB1,AllowedB).

collect_predicates(t(Sigma,Sigma,_),Trans,Allowed) :-
    list_of_predicates_wi(Trans,List00),
    sort(List00,List0),
    get_small_predicates(List0,Sigma,List1),
    get_hash(List1,Allowed).

list_of_predicates([],[]).
list_of_predicates([trans(_,P,_)|Ps0],[P|Ps]):-
    list_of_predicates(Ps0,Ps).

list_of_predicates_d([],[]).
list_of_predicates_d([trans(_,P0/_,_)|Ps0],Ps1):-
    (	P0 == []
    ->	Ps1=Ps
    ;   Ps1=[P0|Ps]
    ),
    list_of_predicates_d(Ps0,Ps).

list_of_predicates_r([],[]).
list_of_predicates_r([trans(_,_/P0,_)|Ps0],Ps1):-
    (	P0 == []
    ->	Ps1=Ps
    ;   Ps1=[P0|Ps]
    ),
    list_of_predicates_r(Ps0,Ps).

list_of_predicates_wr([],[]).
list_of_predicates_wr([trans(_,_/(P0/_),_)|Ps0],Ps1):-
    (	P0 == []
    ->	Ps1=Ps
    ;   Ps1=[P0|Ps]
    ),
    list_of_predicates_wr(Ps0,Ps).

list_of_predicates_i([],[]).
list_of_predicates_i([trans(_,P0a/P0b,_)|Ps0],Ps1):-
    (	P0a = '$@'(P0),
	P0b = '$@'(P0)
    ->	Ps1=[P0|Ps]
    ;	(   P0a == []
	->  Ps1=Ps2
	;   Ps1=[P0a|Ps2]
	),
	(   P0b == []
	->  Ps2=Ps
	;   Ps2=[P0b|Ps]
	)
    ),
    list_of_predicates_i(Ps0,Ps).

list_of_predicates_wi([],[]).
list_of_predicates_wi([trans(_,P0a/(P0b/_),_)|Ps0],Ps1):-
    (	P0a = '$@'(P0),
	P0b = '$@'(P0)
    ->	Ps1=[P0|Ps]
    ;	(   P0a == []
	->  Ps1=Ps2
	;   Ps1=[P0a|Ps2]
	),
	(   P0b == []
	->  Ps2=Ps
	;   Ps2=[P0b|Ps]
	)
    ),
    list_of_predicates_wi(Ps0,Ps).

get_small_predicates(L0,Pm,L) :-
    kv_list(L0,L1),
    fsa_pred_module:determinize_preds(Pm,L1,L).

kv_list([],[]).
kv_list([P|T0],[P-[P]|T]):-
    kv_list(T0,T).


get_hash(KeyList0,Hash) :-
    inverse_keylist_m(KeyList0,KeyList1),
    keymerge(KeyList1,KeyList),
    fsa_hash:fsa_hash_new(Hash),
    keylist_to_hash(KeyList,Hash).

%% as union/2, but we have keylists, and we merge the lists associated
%% with identical keys as well.  
keymerge([], []).
keymerge([Set|Sets], Union) :-
    length([Set|Sets], NumberOfSets),
    keymerge_all(NumberOfSets, [Set|Sets], Union, []).

keymerge_all(N,Sets0,Union,Sets) :-
    (  N=:=1  -> Sets0=[Union|Sets]
    ;  N=:=2  -> Sets0=[Set1,Set2|Sets], 
                 keymerge(Set1,Set2,Union)
    ;  A is N>>1,
       Z is N-A,
       keymerge_all(A, Sets0, X, Sets1),
       keymerge_all(Z, Sets1, Y, Sets),
       keymerge(X, Y, Union)
    ).

%% much like keymerge, but pairs with the same key merge their values 
%% (which are ordered lists).
keymerge([],S,S).
keymerge([O|Os], Ns, Set) :- 
    keymerge(Ns, O, Os, Set).

keymerge([], O, Os, [O|Os]).
keymerge([N-Nlist|Ns], O-Olist, Os, Set) :-
    compare(C, O, N), 
    keymerge(C, O-Olist, Os, N-Nlist, Ns, Set).

keymerge(<, O1, Os, N, Ns, [O1|Set]) :- 
    keymerge(Os, N, Ns, Set).
keymerge(=, _-Olist, Os, N-Nlist, Ns, [N-Xlist|Set]) :- 
    ord_union(Olist,Nlist,Xlist),
    keymerge(Os,Ns,Set).
keymerge(>, O, Os, N1, Ns, [N1|Set]) :- 
    keymerge(Ns, O, Os, Set).


inverse_keylist_m([],[]).
inverse_keylist_m([K-V|T0],List) :-
    inverse_keylist_ml(V,K,List,List1),
    inverse_keylist_m(T0,List1).

inverse_keylist_ml([],_,L,L).
inverse_keylist_ml([H|T],Key,[[H-[Key]]|L0],L) :-
    inverse_keylist_ml(T,Key,L0,L).

keylist_to_hash([],_).
keylist_to_hash([K-V|T],Hash) :-
    fsa_hash:fsa_hash_access(K,V,Hash),
    keylist_to_hash(T,Hash).

expand_non_overlapping([],_Sigma,_,[]).
expand_non_overlapping([Trans|T],Sigma,Allowed,Ts) :-
    expand_non_overlapping_(Sigma,Trans,Allowed,Ts,Ts0),
    expand_non_overlapping(T,Sigma,Allowed,Ts0).

expand_non_overlapping_(r(_),trans(P,Pred,Q),Allowed,Ts,Ts0):-
    fsa_hash:fsa_hash_access(Pred,Preds,[],Allowed),
    findall(trans(P,NPred,Q),member(NPred,Preds),Ts,Ts0).

expand_non_overlapping_(r(_,_),trans(P,Pred/W,Q),Allowed,Ts,Ts0):-
    fsa_hash:fsa_hash_access(Pred,Preds,[],Allowed),
    findall(trans(P,NPred/W,Q),member(NPred,Preds),Ts,Ts0).

expand_non_overlapping_(t(PmA,PmB),trans(P,Pa/Pb,Q),Allowed,Ts,Ts0):-
    PmA==PmB,!,
    findall(trans(P,NPred,Q),expand_i(Pa/Pb,Allowed,NPred,PmA),Ts,Ts0).

expand_non_overlapping_(t(PmA,PmB),trans(P,L/R,Q),AllowedA/AllowedB,Ts,Ts0):-
    PmA \== PmB,
    findall(trans(P,NPred,Q),expand_non_overlap(L,AllowedA,R,AllowedB,NPred),
	    Ts,Ts0).

expand_non_overlapping_(t(PmA,PmB,_),trans(P,Pa/(Pb/W),Q),Allowed,Ts,Ts0):-
    PmA==PmB,!,
    findall(trans(P,X/(Y/W),Q),expand_i(Pa/Pb,Allowed,X/Y,PmA),Ts,Ts0).

expand_non_overlapping_(t(PmA,PmB,_),trans(P,L/(R/W),Q),AllA/AllB,Ts,Ts0):-
    PmA \== PmB,
    findall(trans(P,X/(Y/W),Q),expand_non_overlap(L,AllA,R,AllB,X/Y),Ts,Ts0).

expand_i(Pa/Pb,Allowed,NPred,PM) :-
    (	Pa='$@'(P),
	Pb='$@'(P)
    ->	NPred=Predi/Predi,
	fsa_hash:fsa_hash_access(P,Preds,[],Allowed),
	member(Pred,Preds),
	fsa_pred_module:identity(PM,Pred,Predi)
    ;   NPred=Left/Right,
	expand_left(Pa,Allowed,Left),
	expand_right(Pb,Allowed,Right)
    ).

expand_non_overlap(L,AllowedA,R,AllowedB,Left/Right) :-
    expand_left(L,AllowedA,Left),
    expand_right(R,AllowedB,Right).

expand_left(L,AllowedA,Left) :-
    (	L==[]
    ->	Left=[]
    ;   fsa_hash:fsa_hash_access(L,Lefts,[],AllowedA),
	member(Left,Lefts)
    ).

expand_right(L,AllowedA,Left) :-
    (	L==[]
    ->	Left=[]
    ;   fsa_hash:fsa_hash_access(L,Lefts,[],AllowedA),
	member(Left,Lefts)
    ).


