%%% -*- Module: fsa_determinizer; -*-
:- module(fsa_determinizer,[]).

#!swi# :- use_module(library(ordsets),  [ ord_union/2 ]).
#swi# :- use_module(ordsets,  [ ord_union/2 ]).
#yap# :- use_module(library(lists), [ member/2 ]).
#!swi#:- use_module(library(terms), [ term_hash/2 ]).
#swi#:- use_module(terms, [ term_hash/2 ]).
#swi# :- style_check([-atom]).

#!yap# :- dynamic fsa_regex_cache:detpreds/4.

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

prepare_data(per_graph,FsaIn,Starts,Finals,TransT,nojumps):-
    efree(FsaIn,FsaOut),
    fsa_data:fsa_start_states(FsaOut,Starts),
    fsa_data:fsa_final_states(FsaOut,Finals),
    fsa_data:index_transitions(FsaOut,TransT).

prepare_data(per_inverted_graph,FsaIn,Starts,Finals,TransT,nojumps) :-
    inverted_efree(FsaIn,FsaOut),
    fsa_data:fsa_start_states(FsaOut,Starts),
    fsa_data:fsa_final_states(FsaOut,Finals),
    fsa_data:index_transitions(FsaOut,TransT).

prepare_data(per_reachable_graph,FsaIn,Starts,Finals,TransT,nojumps) :-
    reachable_efree(FsaIn,FsaOut),
    fsa_data:fsa_start_states(FsaOut,Starts),
    fsa_data:fsa_final_states(FsaOut,Finals),
    fsa_data:index_transitions(FsaOut,TransT).

prepare_data(per_co_reachable_graph,FsaIn,Starts,Finals,TransT,nojumps) :-
    co_reachable_efree(FsaIn,FsaOut),
    fsa_data:fsa_start_states(FsaOut,Starts),
    fsa_data:fsa_final_states(FsaOut,Finals),
    fsa_data:index_transitions(FsaOut,TransT).
    
prepare_data(per_subset,FsaIn,Starts,Finals,TransT,JumpTable) :-
    fsa_data:fsa_start_states(FsaIn,Starts),
    fsa_data:fsa_final_states(FsaIn,Finals),
    fsa_data:index_transitions(FsaIn,TransT),
    index_jumps(FsaIn,JumpTable).

prepare_data(per_state,FsaIn,Starts,Finals,TransT,Data):-
    fsa_data:fsa_start_states(FsaIn,Starts),
    fsa_data:fsa_final_states(FsaIn,Finals),
    fsa_data:index_transitions(FsaIn,TransT),
    fsa_data:fsa_states_number(FsaIn,Size),
    index_jumps(FsaIn,Table),
    (	Table==nojumps
    ->	Data=nojumps
    ;   Data=j(Table,PrivTable),
	fsa_array:fsa_array_new(PrivTable,Size)
    ).

:- public determinize/3, determinize/4, determinize_with_map/4,
          fsa_deterministic/1,
          inverted_efree/2, efree/2, determinize_preds/3.

determinize(Fsa0,Fsa,Variant):-
    determinize(Fsa0,Fsa,Variant,_Map).

determinize(FsaIn,FsaOut,Variant,Done) :-
    fsa_data:fsa_symbols_decl(FsaIn,r(Mod)),
    fsa_data:fsa_states_number(FsaIn,Size),
    fsa_data:fsa_components(r(Mod),NrS,[0],Fins,Trs,[],FsaOut),
    fsa_hash:fsa_hash_new(JST),
    prepare_data(Variant,FsaIn,Start0,FT,TransT,JT),
    fsa_hash:fsa_hash_new(Done),
    treat_start_state(Start0,Start,Done,Fins,Fins0,JT,JST,Variant,FT,Size),
    subset_c(s(0),[Start|Q]-Q,0,Done,Trs,Fins0,TransT,JT,JST,Variant,FT,1,NrS,Mod,Size).

%% Map is a keylist with Integer-Subset pairs indicating the subset
%% that a state name is associated with.
determinize_with_map(FsaIn,FsaOut,Variant,Map) :-
    determinize(FsaIn,FsaOut,Variant,Done),
    fsa_hash:fsa_hash_to_keylist(Done,Map0),
    inverse_keylist(Map0,Map).

inverse_keylist([],[]).
inverse_keylist([Key-Val|Tail0],[Val-Key|Tail]):-
    inverse_keylist(Tail0,Tail).

treat_start_state(Start0,Start,Hash,Finals0,Finals,JT,JST,JF,FT,Size) :-
    jump_reachables(JF,Start0,Start,JT,JST,Size),
    fsa_hash:fsa_hash_access(Start,0,Hash),
    add_final(Start,0,FT,Finals0,Finals).

%% subset_c(AgC,Ag,NrQ0,Tg,Trs,Fs,TT,JT,JsT,JF,FT,NrS0,NrS)
%% AgC: number of items in Agenda
%% Ag: Agenda (dlist of subsets). Agenda is a dlist since new subsets are
%%     added to the end, to make sure that transitions are constructed in
%%     order. That way there's no need to sort afterwards.
%% NrQ0: state number of current state
%% TG: table for subset -> state number mapping
%% Trs: resulting ordered list of transitions (trans(Int,Atom,Int))
%% Fs: resulting ordered list of final states (Int)
%% TT: indexed table of transitions
%% JT: indexed table of jumps
%% JsT: table for memo-ing transitive closure of jumps computation
%% JF: flag to indicate mode of jumps treatment
%% NrS0: currently available number for new state,
%% NrS: eventually the total number of states
subset_c(0,_,_,_,[],[],_,_,_,_,_,S,S,_Mod,_).
subset_c(s(AgC0),[H|Ag0]-Ag1,NrQ0,Tg,Trs,Fs,TT,JT,JsT,JF,FT,NrS0,NrS,Mod,Sz) :-
    %% this is called 'Instruction Computation' in Johnson and Wood,
    %% Instruction Computation in Subset Construction, in: Raymond, Wood,
    %% Yu (eds), Automata Implementation. Springer Lecture Notes in
    %% Computer Science 1260.
    %%%%bb_put(user:dlast,NrQ0), % for debugging only!
    collect_ts(H,List0,TT),
    merge_transitions(List0,List1),
    fsa_pred_module:determinize_preds(Mod,List1,List),
    add_jumps(List,Trs,Trs0,JT,JsT,JF,Tg,Fs,Fs0,FT,NrQ0,NrS0,NrS1,
	                         AgC0,Ag0-Ag1,AgC,Ag,Sz),
    NrQ1 is NrQ0+1,
    subset_c(AgC,Ag,NrQ1,Tg,Trs0,Fs0,TT,JT,JsT,JF,FT,NrS1,NrS,Mod,Sz).

%% add_jumps(Pairs,Trs0,Trs,JT,JsT,JF,NK,TG,Fs0,Fs,FT,NrQ0,NrS0,NrS)
%% Pairs: list of symbol-subset pairs
%% Trs0-Trs: dlist of resulting transitions (list of trans(Num,Sym,Num))
%% FS0-Fs: dlist of resulting final states (list of state numbers)
%% others are as in subset_c(..)
add_jumps([],Trs,Trs,_,_,_,_,Fs,Fs,_,_,NrS,NrS,AgC,Ag,AgC,Ag,_Sz).
add_jumps([Sy-End0|Tail0],[trans(NrQ0,Sy,NrEnd)|Trs0],Trs,JT,JsT,JF,
	 TG,Fs0,Fs,FT,NrQ0,NrS0,NrS,AgC0,Ag0,AgC,Ag,Sz) :-
    jump_reachables(JF,End0,End,JT,JsT,Sz),     %% epsilon-closure(End0)=End
    fsa_hash:fsa_hash_access(End,NrEnd,TG),     %% get state number if old

    (	var(NrEnd)                              %% var: it's a new state 
    ->	NrEnd is NrS0,                          %% assign number to state
	NrS1 is NrS0+1,                         %% update next state number
	add_queue(End,AgC0,Ag0,AgC1,Ag1),       %% update agenda
	add_final(End,NrEnd,FT,Fs0,Fs1)         %% maybe add to final states

    ;	Fs0 = Fs1,                              %% it was not a new state
	NrS0=NrS1,                              %% hence everything remains
	AgC0=AgC1,                              %% whatever it was
	Ag0=Ag1                                 %% previously
    ),
    add_jumps(Tail0,Trs0,Trs,JT,JsT,JF,TG,Fs1,Fs,FT,NrQ0,NrS1,NrS,
                          AgC1,Ag1,AgC,Ag,Sz).  %% do the rest

add_queue(End,AgC0,Ag0-[End|Ag],s(AgC0),Ag0-Ag).

collect_ts([],[],_).
collect_ts([P|Tail],L0,TT) :-
    fsa_array:fsa_array_access(P,L1,TT),
    (	nonvar(L1)
    ->	L0=[L1|L2]
    ;	L2 = L0
    ),
    collect_ts(Tail,L2,TT).


add_final(State,NState,FT,Finals0,Finals) :-
    fsa_util:ord_intersect(State,FT,Yes),
    add_final0(Yes,NState,Finals0,Finals).

add_final0(yes,NState,[NState|Finals],Finals).
add_final0(no,_,F,F).

jump_reachables(Flag,S0,S,Data,Memo,Sz):-
    (	Data==nojumps
    ->	S0=S
    ;   fsa_hash:fsa_hash_access(S0,S,Memo),
	(   nonvar(S)
	->  true
	;   jump_reachables0(Flag,S0,S,Data,Sz)
	)
    ).

jump_reachables0(per_subset,States0,States,JTable,Sz) :-
    jump_reachables_per_subset(States0,States,JTable,Sz).
jump_reachables0(per_state,States0,States,j(JTable,Memo),Sz) :-
    jump_reachables_per_state(States0,States1,JTable,Memo,Sz),
    ord_union(States1,States).

jump_reachables_per_state([],[],_,_,_).
jump_reachables_per_state([H|T],[S1|S],JT,JsT,Sz) :-
    jump_reachable_per_state(H,S1,JT,JsT,Sz),
    jump_reachables_per_state(T,S,JT,JsT,Sz).

jump_reachable_per_state(H,States,JT,JsT,Sz) :-
    fsa_array:fsa_array_access(H,States,JsT),
    (	nonvar(States)
    ->	true
    ;	fsa_array:fsa_array_new(Done,Sz),
	fsa_array:fsa_array_access(H,[],Done),
	jump_reachable([H],States1,[H],Done,JT),
	sort(States1,States)
    ).

jump_reachables_per_subset(States0,States,JTable,Sz) :-
    fsa_array:fsa_array_new(Done,Sz),
    fsa_array_access_list(States0,Done),
    jump_reachable(States0,States1,States0,Done,JTable),
    sort(States1,States).

fsa_array_access_list([],_Done).
fsa_array_access_list([H|T],Done) :-
    fsa_array:fsa_array_access(H,[],Done),
    fsa_array_access_list(T,Done).

%% jump_reachable(+Ag,Results,?DoneTable,AgLength,TreatedL0,TreatedL)
jump_reachable([],R,R,_Done,_Table).
jump_reachable([H|T],Results0,Results,Done,JTable) :-
    fsa_array:fsa_array_access(H,Next0,[],JTable),
    add_if_new(Next0,T,Agenda,Results0,Results1,Done),
    jump_reachable(Agenda,Results1,Results,Done,JTable).

add_if_new([],Ag,Ag,R,R,_Done).
add_if_new([H|T],Ag0,Ag,Res0,Res,Done) :-
    fsa_array:fsa_array_access(H,XX,Done),
    (   var(XX)
    ->  Ag1=[H|Ag0],
        XX=[],
        Res0=[H|Res1]
    ;   Ag1=Ag0,
        Res0=Res1
    ),
    add_if_new(T,Ag1,Ag,Res1,Res,Done).

index_jumps(Fa,JT) :-
    fsa_data:fsa_jumps(Fa,Jumps),
    fsa_data:fsa_states_number(Fa,Size),
    index_jumps0(Jumps,JT,Size).

index_jumps0([],nojumps,_).
index_jumps0([H|T],JT,Sz):-
    fsa_array:fsa_array_new(JT,Sz),
    fsa_data:jumps_to_thing([H|T],JT).

%% as union/2, but we have keylists, and we merge the lists associated
%% with identical keys as well.
%% NB: merge_transitions is used by fsa_t_determinizer module as well!
%% NB: merge_transitions is used by fsa_minimizer module as well!
merge_transitions([], []).
merge_transitions([Set|Sets], Union) :-
    length([Set|Sets], NumberOfSets),
    merge_ts_all(NumberOfSets, [Set|Sets], Union, []).

merge_ts_all(N,Sets0,Union,Sets) :-
    (  N=:=1  -> Sets0=[Union|Sets]
    ;  N=:=2  -> Sets0=[Set1,Set2|Sets], 
                 merge_ts(Set1,Set2,Union)
    ;  A is N>>1,
       Z is N-A,
       merge_ts_all(A, Sets0, X, Sets1),
       merge_ts_all(Z, Sets1, Y, Sets),
       merge_ts(X, Y, Union)
    ).

%% much like keymerge, but pairs with the same key merge their values 
%% (which are ordered lists).
merge_ts([],S,S).
merge_ts([O|Os], Ns, Set) :- 
    merge_ts(Ns, O, Os, Set).

merge_ts([], O, Os, [O|Os]).
merge_ts([N-Nlist|Ns], O-Olist, Os, Set) :-
    compare(C, O, N), 
    merge_ts(C, O-Olist, Os, N-Nlist, Ns, Set).

merge_ts(<, O1, Os, N, Ns, [O1|Set]) :- 
    merge_ts(Os, N, Ns, Set).
merge_ts(=, _-Olist, Os, N-Nlist, Ns, [N-Xlist|Set]) :- 
    my_merge(Olist,Nlist,Xlist),
    merge_ts(Os,Ns,Set).
merge_ts(>, O, Os, N1, Ns, [N1|Set]) :- 
    merge_ts(Ns, O, Os, Set).

my_merge([], Set, Set).
my_merge([O|Os], Ns, Set) :- 
    my_merge(Ns, O, Os, Set).

my_merge([], O, Os, [O|Os]).
my_merge([N|Ns], O, Os, Set) :-
    compare(C, O, N), 
    my_merge(C, O, Os, N, Ns, Set).

my_merge(<, O1, Os, N, Ns, [O1|Set]) :- 
    my_merge(Os, N, Ns, Set).
my_merge(=, _, Os, N, Ns, [N|Set]) :- 
    my_merge(Os, Ns, Set).
my_merge(>, O, Os, N1, Ns, [N1|Set]) :- 
    my_merge(Ns, O, Os, Set).



inverted_efree(FsaIn,FsaOut) :-
    fsa_data:fsa_components(Syms,States,Starts,Finals0,_,Jumps0,FsaIn),
    (	Jumps0==[]
    ->	fsa_util:debug_message(3,"efree: already epsilon-free!~n",[]),
	FsaIn=FsaOut
    ;	fsa_data:fsa_components(Syms,States,Starts,Finals,Trans,[],FsaOut),
	transitive_closure_and_invert(Jumps0,JumpTable,States),
	lookup_each(Finals0,JumpTable,Finals1),
	ord_union([Finals0|Finals1],Finals),
	extend_trans_inverted(FsaIn,JumpTable,Trans1),
	sort(Trans1,Trans)
    ).

extend_trans_inverted(FsaIn,Table,Trans):-
    transitions_to_graph(FsaIn,Graph0),
    new_sources(Graph0,Table,Graph1),
    graph_to_transitions(Graph1,Trans).

new_sources([],_,[]).
new_sources([Key-Val|T0],Table,G):-
    fsa_array:fsa_array_access(Key,Keys,[],Table),
    new_sources0([Key|Keys],Val,G,G0),
    new_sources(T0,Table,G0).

new_sources0([],_,G,G).
new_sources0([P|Ps],V,[P-V|G0],G):-
    new_sources0(Ps,V,G0,G).

graph_to_transitions([],[]).
graph_to_transitions([P-Val|T0],Ts):-
    graph_to_transitions0(Val,P,Ts,Ts0),
    graph_to_transitions(T0,Ts0).

graph_to_transitions0([],_,Ts,Ts).
graph_to_transitions0([Sym-Qs|T0],P,Ts0,Ts):-
    graph_to_transitions00(Qs,P,Sym,Ts0,Ts1),
    graph_to_transitions0(T0,P,Ts1,Ts).

graph_to_transitions00([],_,_,Ts,Ts).
graph_to_transitions00([Q|Qs],P,Sym,[trans(P,Sym,Q)|Ts0],Ts):-
    graph_to_transitions00(Qs,P,Sym,Ts0,Ts).

efree(FsaIn,FsaOut) :-
    fsa_data:fsa_components(Syms,States,Starts0,Finals,_,Jumps0,FsaIn),
    (	Jumps0==[]
    ->	fsa_util:debug_message(3,"efree: already epsilon-free!~n",[]),
	FsaIn=FsaOut
    ;	fsa_data:fsa_components(Syms,States,Starts,Finals,Trans,[],FsaOut),
	transitive_closure(Jumps0,JumpTable,States),
	fsa_hash:fsa_hash_new(Memo),
	lookup_each_and_union(Starts0,JumpTable,Starts,Memo),
	extend_trans(FsaIn,JumpTable,Trans,Memo)
    ).

lookup_each_and_union(States,Table,Result,Memo):-
    fsa_hash:fsa_hash_access(States,Result,Memo),
    (	var(Result)
    ->	lookup_each(States,Table,Result1),
	ord_union([States|Result1],Result)
    ;   true
    ).

extend_trans(FsaIn,JumpTable,Trans,Memo) :-
    transitions_to_graph(FsaIn,Graph),
    graph_to_trans_with_jumps(Graph,JumpTable,Memo,Trans,[]).

graph_to_trans_with_jumps([],_,_,Tr,Tr).
graph_to_trans_with_jumps([Source-G0|Gs],Table,M,Tr0,Tr) :-
    graph_to_trans_with_jumps_(G0,Source,Table,M,Tr0,Tr1),
    graph_to_trans_with_jumps(Gs,Table,M,Tr1,Tr).

graph_to_trans_with_jumps_([],_,_,_,Tr,Tr).
graph_to_trans_with_jumps_([Sym-List|Tail],Source,Table,Memo,Tr0,Tr) :-
    lookup_each_and_union(List,Table,Union,Memo),
    list_to_transitions(Union,Source,Sym,Tr0,Tr1),
    graph_to_trans_with_jumps_(Tail,Source,Table,Memo,Tr1,Tr).

list_to_transitions([],_,_,Tr,Tr).
list_to_transitions([H|T],Source,Sym,[trans(Source,Sym,H)|Tr0],Tr) :-
    list_to_transitions(T,Source,Sym,Tr0,Tr).

lookup_each([],_,[]).
lookup_each([Q0|T],Table,[Qs|Tlist]):-
    fsa_array:fsa_array_access(Q0,Qs,[],Table),
    lookup_each(T,Table,Tlist).

%% nb. from now on we allow that some vertices are not mentioned at all, 
%% indicating the empty set of edges.
%% transitive_closure(+ListOfJumps,-ThingOfJumps,+NumberOfStates)
transitive_closure(Jumps,ClosureThing,Sz) :-
    fsa_hash:fsa_hash_new(Done),
    fsa_array:fsa_array_new(BasicThing,Sz),
    fsa_array:fsa_array_new(ClosureThing,Sz),
    prepare_jumps_efree(Jumps,BasicThing,Pairs,Done),
    closure(Pairs,TrPairs0,Pairs,BasicThing,Done),
    sort(TrPairs0,TrPairs),
    edges_to_thing(TrPairs,ClosureThing).

%% nb. from now on we allow that some vertices are not mentioned at all, 
%% indicating the empty set of edges.
%% transitive_closure(+ListOfJumps,-ThingOfJumps,+NumberOfStates)
transitive_closure_and_invert(Jumps,ClosureThing,Sz) :-
    fsa_hash:fsa_hash_new(Done),
    fsa_array:fsa_array_new(BasicThing,Sz),
    fsa_array:fsa_array_new(ClosureThing,Sz),
    prepare_jumps_efree(Jumps,BasicThing,Pairs,Done),
    closure(Pairs,TrPairs0,Pairs,BasicThing,Done),
    invert_key_val(TrPairs0,TrPairs1),
    sort(TrPairs1,TrPairs),
    edges_to_thing(TrPairs,ClosureThing).

invert_key_val([],[]).
invert_key_val([Key-Val|T0],[Val-Key|T]):-
    invert_key_val(T0,T).

%% prepare_jumps_efree(+JumpsList,-Thing,-ListOfPairs,-HashedSet)
%% constructs three different representations for list of jump(A,B)
%% Thing is used for fast lookup of jumps in closure
%% Done is used to initialize table of pairs already treated in closure
%% Pairs is used to initialize resulting list of pairs, and also to
%% initialize the Agenda (although the Agenda could also be in jump(A,B)
%% format of course)
prepare_jumps_efree([],_Thing,[],_Done).
prepare_jumps_efree([jump(P,Q)|Tail],Thing,[P-Q|Pairs],Done):-
    fsa_array:fsa_array_access(P,[Q|Targets],Thing),
    fsa_hash:fsa_hash_access(P-Q,[],Done),
    prepare_jumps1(Tail,P,Targets,Thing,Pairs,Done).

prepare_jumps1([],_,[],_Thing,[],_Done).
prepare_jumps1([jump(P0,Q)|Tail],P,Targets,Thing,[P0-Q|Pairs],Done) :-
    fsa_hash:fsa_hash_access(P0-Q,[],Done),
    (	P0==P
    ->	Targets=[Q|Targets1],
	prepare_jumps1(Tail,P,Targets1,Thing,Pairs,Done)
    ;	Targets=[],
	fsa_array:fsa_array_access(P0,[Q|Targets1],Thing),
	prepare_jumps1(Tail,P0,Targets1,Thing,Pairs,Done)
    ).

edges_to_thing([],_Thing).
edges_to_thing([P-Q|Tail],Thing) :-
    fsa_array:fsa_array_access(P,[Q|Targets],Thing),
    edges_to_thing(Tail,P,Targets,Thing).

edges_to_thing([],_,[],_).
edges_to_thing([P0-Q|Tail],P,Targets,Thing) :-
    (	P0==P
    ->	Targets=[Q|Targets1],
	edges_to_thing(Tail,P,Targets1,Thing)
    ;	Targets=[],
	fsa_array:fsa_array_access(P0,[Q|Targets1],Thing),
	edges_to_thing(Tail,P0,Targets1,Thing)
    ).

%% closure(Agenda,Database0,Database,BasicPairsThing,DoneHash)
closure([],T,T,_,_).
closure([A-B|Agenda0],New0,New,Basic,Done) :-
    fsa_array:fsa_array_access(B,Neighbors,[],Basic),
    add_pairs(Neighbors,A,Agenda,Agenda0,New0,New1,Done),
    closure(Agenda,New1,New,Basic,Done).

add_pairs([],_,Ag,Ag,N,N,_).
add_pairs([C|Tail],A,Ag0,Ag,N0,N,Done) :-
    fsa_hash:fsa_hash_access(A-C,Value,Done),
    (	var(Value)
    ->	Ag0 = [A-C|Ag1],
        N0 = [A-C|N1],
	Value=[]
    ;	Ag0=Ag1,
        N0=N1
    ),
    add_pairs(Tail,A,Ag1,Ag,N1,N,Done).

%% transitions_to_graph(Fa,Graph)
%% Graph is a list of State-Xs
%%               where Xs is a list of Sym-Targets
%% representing the transitions of Fa,
%% e.g. trans(0,a,1)
%%      trans(0,a,2)
%%      trans(0,b,1)
%%      trans(1,a,2)
%% has graph [0-[a-[1,2],
%%               b-[1]
%%              ],
%%            1-[]
%%           ]
%% transitions must be ordered already
transitions_to_graph(Fa,Graph) :-
    fsa_data:fsa_transitions(Fa,Trans),
    transitions_to_graph0(Trans,Graph).

transitions_to_graph0([],[]).
transitions_to_graph0([trans(P,Sym,Q)|Tail],[P-[Sym-[Q|Targets]|Syms]|Gr]) :-
    transitions_to_graph1(Tail,P,Sym,Targets,Syms,Gr).

transitions_to_graph1([],_,_,[],[],[]).
transitions_to_graph1([trans(P0,Sym0,Q)|Tail],P,Sym,Targets,Syms,Graph):-
    (	P0 == P
    ->	(  Sym0==Sym
	->  Targets=[Q|Targets1],
	    transitions_to_graph1(Tail,P,Sym,Targets1,Syms,Graph)
	;   Targets=[],
	    Syms=[Sym0-[Q|Targets1]|Syms1],
	    transitions_to_graph1(Tail,P,Sym0,Targets1,Syms1,Graph)
	)
    ;	Targets=[],
	Syms=[],
	Graph=[P0-[Sym0-[Q|Targets1]|Syms1]|Graph1],
	transitions_to_graph1(Tail,P0,Sym0,Targets1,Syms1,Graph1)
    ).


%%% tools to be used by predicate modules
/*  this one if you expect many overlapping results
determinize_preds(Kv0,Kv,Module) :-
    all_exclusive_preds(Kv0,Kv1,Module),
    remove_epsilon_target_state(Kv1,Kv).

remove_epsilon_target_state([],[]).
remove_epsilon_target_state([K-V|Tail0],R):-
    remove_ets(V,K,R,R0),
    remove_epsilon_target_state(Tail0,R0).

remove_ets([],_,R,R).
remove_ets([H|T],K,[K-[H|T]|R],R).

all_exclusive_preds([],[],_).
all_exclusive_preds([Key0-Val0|T],KvList,M) :-
    findall(Key-Val,take_or_leave(Key0,Key,Val0,Val,M),KvList0),
    sort(KvList0,KvList1),
    all_exclusive_preds1(T,KvList1,KvList,M).

all_exclusive_preds1([],Kv,Kv,_).
all_exclusive_preds1([Key0-Target0|T],Kv0,Kv,M):-
    findall(Key-Target,
	    add_exclusive_pred(Key0,Target0,Kv0,Key,Target,M),
	    Kv1
	   ),
    sort(Kv1,Kv2),
    all_exclusive_preds1(T,Kv2,Kv,M).

add_exclusive_pred(Key0,Target0,KvList0,Key,Target,M) :-
    member(Key1-Target1,KvList0),
    take_or_leave(Key0,Key1,Key,Target1,Target0,Target,M).

%% take
take_or_leave(Pred,Pred,Target,Target,_).
%% leave
take_or_leave(Pred0,Pred,_,[],M):-
    fsa_pred_module:negation(M,Pred0,Pred).

%% take
take_or_leave(Pred0,Pred1,Pred2,Target,Targets0,Targets,M) :-
    fsa_pred_module:conjunction(M,Pred0,Pred1,Pred2),
    ord_union(Targets0,Target,Targets).

%% leave
take_or_leave(Pred0,Pred1,Pred2,T,_,T,M) :-
    fsa_pred_module:negation(M,Pred0,NPred0),
    fsa_pred_module:conjunction(M,NPred0,Pred1,Pred2).

*/

%%%% determinize_preds(Kv0,Kv,Module)
%%%%
%%%% distinguish
%%%% 1) -> computation of exclusive preds
%%%% 2) -> computation wrt associated values
%%%% this allows caching of step 1)
%%%%

determinize_preds(Un0,Un,Module) :-
    fsa_globals:fsa_global_get(determinize_preds_cache,Cache),
    determinize_preds(Cache,Un0,Un,Module).

determinize_preds(on,Un0,Un,Module) :-
    determinize_preds_with_caching(Un0,Un,Module).
determinize_preds(off,Un0,Un,Module) :-
    determinize_preds_without_caching(Un0,Un,Module).

determinize_preds_with_caching(Un0,Un,Module):-
    kvs(Un0,Un1,Keys),
    term_hash(determinize_preds(Keys,Module),Index),
#!yap#     (   fsa_regex_cache:detpreds(Index,Keys,Module,Un2)
#yap#     (   recorded(fsa_regex_cache,detpreds(Index,Keys,Module,Un2),_)
    ->	true
    ;	findall(Pred-Tg0-Un1,exclusive_pred(Un1,Pred,Tg0,Module),Un2),
#!yap#         assertz(fsa_regex_cache:detpreds(Index,Keys,Module,Un2))
#yap#         catch(recordz(fsa_regex_cache,detpreds(Index,Keys,Module,Un2),_),P,format(user_error,"Caching disabled for large term (~w)~n",[P]))
    ),
    ord_union_each_val(Un2,Un0,Un).

ord_union_each_val([],_,[]).
ord_union_each_val([K-V0-Union0|T0],Union0,[K-V|T]) :-
    ord_union(V0,V),
    ord_union_each_val(T0,Union0,T).

%% provided as a tool; can be used if your predicate module
%% has negation/3 with at most one result.
determinize_preds_without_caching(Union0,Union,Module):-
    findall(Pred-Targets,exclusive_pred_nc(Union0,Pred,Targets,Module),Union).

exclusive_pred_nc(List,Pred,Targets,M) :-
    exclusive_pred(List,Pred,Targets0,M),
    ord_union(Targets0,Targets).

kvs([],[],[]).
kvs([K-_|T],[K-_|T1],[K|Ks]):-
    kvs(T,T1,Ks).

exclusive_pred([],Pred,Pred,[],_,F,F).
exclusive_pred([Pred1-Target|Rest],Pred0,Pred,Targets,M,F0,F):-
    take_or_leave(Pred0,Pred1,Pred2,Target,Targets,Targets0,M,F0,F1),
    exclusive_pred(Rest,Pred2,Pred,Targets0,M,F1,F).

exclusive_pred([First-Target0|Union0],Pred,Targets,M):-
    take_or_leave0(First,Pred0,Target0,Targets,Targets2,M,F1),
    exclusive_pred(Union0,Pred0,Pred,Targets2,M,F1,1).

%% take
take_or_leave0(Pred,Pred,Target,[Target|Targets],Targets,_,1).
%% leave
take_or_leave0(Pred0,Pred,_,T,T,M,0):-
    fsa_pred_module:negation(M,Pred0,Pred).

%% take
take_or_leave(Pred0,Pred1,Pred2,Target,[Target|Targets],Targets,M,_,1) :-
    fsa_pred_module:conjunction(M,Pred0,Pred1,Pred2).
%% leave
take_or_leave(Pred0,Pred1,Pred2,_,T,T,M,F,F) :-
    fsa_pred_module:negation(M,Pred1,Npred1),
    fsa_pred_module:conjunction(M,Pred0,Npred1,Pred2).

reachable_efree(FsaIn,FsaOut):-
    fsa_data:fsa_jumps(FsaIn,Jumps),
    reachable_efree0(Jumps,FsaIn,FsaOut).

reachable_efree0([],Fa,Fa):-
    fsa_util:debug_message(3,"efree: already epsilon-free!~n",[]).
reachable_efree0([_|_],FsaIn,FsaOut):-
    fsa_regex:all_components_reachable(efree,FsaIn,FsaOut).

co_reachable_efree(FsaIn,FsaOut):-
    fsa_data:fsa_jumps(FsaIn,Jumps),
    co_reachable_efree0(Jumps,FsaIn,FsaOut).

co_reachable_efree0([],Fa,Fa):-
    fsa_util:debug_message(3,"efree: already epsilon-free!~n",[]).
co_reachable_efree0([_|_],FsaIn,FsaOut):-
    fsa_regex:all_components_co_reachable(efree,FsaIn,FsaOut).

fsa_deterministic(Fa) :-
    fsa_data:fsa_components(r(fsa_frozen),_Nr,[_],_,Transitions,[],Fa),
    is_deterministic(Transitions).

is_deterministic([]).
is_deterministic([trans(P,Sym,_)|Tail]) :-
    is_deterministic(Tail,P,Sym).

is_deterministic([],_,_).
is_deterministic([trans(P0,Sym0,_)|Tail],P,Sym):-
    (	P0 == P
    ->	Sym0 \== Sym
    ;   true
    ),
    is_deterministic(Tail,P0,Sym0).

