%%% -*- Module: fsa_dict; -*-
:- module(fsa_dict,
	  [ fsa_dict_to_perfect_hash_file/2,
	    fsa_dict_to_perfect_hash/2,
	    fsa_dict_to_fsa/2,
	    fsa_dict_to_fsa_file/2,
	    fsa_dict_to_trie_file/2,
	    fsa_dict_to_weighted_trie_file/2,
	    fsa_dict_to_trie/2,
	    fsa_dict_to_weighted_trie/2
	  ]).

#swi# :- style_check([-atom]).

:- use_module(fsa_util).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

:- discontiguous help_info/4.

user:help_info(xpred,fsa_dict_to_perfect_hash,
"fsa_dict_to_perfect_hash(+ListOfStrings,-Fa)","\
 \n A string-to-weight transducer Fa will be constructed implementing the\
 \n perfect hash for the ListOfStrings; i.e. the transducer maps each\
 \n string to its rank (in alphabetic order), and does not accept any\
 \n string not listed in ListOfStrings. The transducer is (w_)minimal."). 

fsa_dict_to_perfect_hash(Lines,Fa):-
    reverse_dict_ph(Lines,Fa0),		% as dict, but adds number of output
    %% dict_ph results in a determinisic string to weight transducer
    %% in which scores cannot be further pushed, so w_minimize can be
    %% skipped!
    %% by construction this has low density. To-do: skip sorting, and
    %% first reversal.
    fsa_regex:fsa_regex_compile(melt(determinize(
                reverse(determinize(freeze(fa(Fa0),Flag)))),Flag),Fa).

user:help_info(xpred,fsa_dict_to_perfect_hash_file,
"fsa_dict_to_perfect_hash_file(+FileIn,+FileOut)","\
 \n FileIn is assumed to contain a set of strings: each line is a\
 \n string. A string-to-weight transducer will be written to FileOut\
 \n implementing the perfect hash for the set of strings read from\
 \n FileIn: i.e. the transducer maps each string to its rank (in\
 \n alphabetic order), and does not accept any string not listed in\
 \n FileIn. The transducer is (w_)minimal.").

fsa_dict_to_perfect_hash_file(InFile,OutFile) :-
    fsa_util:time(fsa_dict:get_file(InFile,Lines)),
    fsa_dict_to_perfect_hash(Lines,Fa),
    fsa_io:fsa_write_file(OutFile,Fa).

user:help_info(xpred,fsa_dict_to_fsa,"fsa_dict_to_fsa(+ListOfStrings,-Fa)",
"A minimal recognizer Fa will be constructed recognizing exactly the\
 \n strings in ListOfStrings").

fsa_dict_to_fsa(Lines,Fa) :-
    reverse_dict(Lines,Fa0),
    fsa_regex:fsa_regex_compile(melt(determinize(
                reverse(determinize(freeze(fa(Fa0),Flag)))),Flag),Fa).

user:help_info(xpred,fsa_dict_to_fsa_file,
"fsa_dict_to_fsa_file(+FileIn,+FileOut)",
"FileIn is assumed to contain a set of strings: each line is a\
 \n string. A minimal automaton recognizing exactly those strings is\
 \n written to FileOut").

fsa_dict_to_fsa_file(InFile,OutFile) :-
    fsa_util:time(fsa_dict:get_file(InFile,Lines)),
    fsa_dict_to_fsa(Lines,Fa),
    fsa_io:fsa_write_file(OutFile,Fa).

user:help_info(xpred,fsa_dict_to_trie_file,
"fsa_dict_to_trie_file(+FileIn,+FileOut)",
"FileIn is assumed to contain a set of strings: each line is a\
 \n string. A deterministic automaton recognizing exactly those strings is\
 \n written to FileOut; the automaton has the form of a trie.").

fsa_dict_to_trie_file(InFile,OutFile) :-
    fsa_util:time(fsa_dict:get_file(InFile,Lines)),
    fsa_dict_to_trie(Lines,Fa),
    fsa_io:fsa_write_file(OutFile,Fa).

fsa_dict_to_weighted_trie_file(InFile,OutFile) :-
    fsa_util:time(fsa_dict:get_file(InFile,Lines)),
    fsa_dict_to_weighted_trie(Lines,Fa),
    fsa_io:fsa_write_file(OutFile,Fa).


user:help_info(xpred,fsa_dict_to_trie,
"fsa_dict_to_trie(+ListOfStrings,-Fa)",
"Fa is a deterministic automaton recognizing exactly each of the\
 \n strings in ListOfStrings; the automaton has the form of a trie.").

fsa_dict_to_trie(Lines,Fa):-
    dict(Lines,Fa).

fsa_dict_to_weighted_trie(Lines,Fa):-
    weighted_dict(Lines,Fa).

reverse_dict(Lines,Fa) :-
    fsa_data:fsa_components(r(Module),States,Starts,[0],Trs,[],Fa),
    fsa_util:time(fsa_dict:add_lines(Lines,States,Starts,Trs,Module,r)).

reverse_dict_ph(Lines,Fa) :-
    fsa_data:fsa_components(r(Module,fsa_frozen),States,Starts,[0],Trs,[],Fa),
    fsa_util:time(fsa_dict:add_lines_ph(Lines,States,Starts,Trs,Module,r)).

dict(Lines,Fa) :-
    fsa_data:fsa_components(r(Module),States,[0],Finals,Trs,[],Fa),
    fsa_util:time(fsa_dict:add_lines(Lines,States,Finals,Trs,Module,n)).

weighted_dict(Lines,Fa) :-
    fsa_data:fsa_components(r(Module,fsa_frozen),States,[0],Finals,Trs,[],Fa),
    fsa_util:time(fsa_dict:add_lines_weighted(Lines,States,Finals,Trs,Module)).

add_lines(Lines,NumberOfStates,Starts,Trans,Module,Flag) :-
    fsa_globals:fsa_global_get(pred_module,Module),
    lines_to_key_lists(Lines,KeyList,Module),
    keymerge(KeyList,Graph),
    graph_to_fa(Graph,Starts,[],Trans0,[],0,0,LastState,Flag),
    sort(Trans0,Trans),
    NumberOfStates is LastState+1.

add_lines_weighted(Lines,NumberOfStates,Starts,Trans,Module) :-
    fsa_globals:fsa_global_get(pred_module,Module),
    lines_to_key_lists_with_counts(Lines,KeyList,Module),
    keymerge_with_counts(KeyList,Graph0),
    counts_to_probs(Graph0,Graph),
    graph_to_fa_with_counts(Graph,Starts,[],Trans0,[],0,0,LastState),
    sort(Trans0,Trans),
    NumberOfStates is LastState+1.

add_lines_ph(Lines,NumberOfStates,Starts,Trans,Module,Flag) :-
    fsa_globals:fsa_global_get(pred_module,Module),
    lines_to_key_lists(Lines,KeyList,Module),
    keymerge(KeyList,Graph),
    graph_to_fa_ph(Graph,Starts,[],Trans0,[],0,0,LastState,0,0,_,Flag),
    sort(Trans0,Trans),
    NumberOfStates is LastState+1.

graph_to_fa([],Fs,Fs,Ts,Ts,_,States,States,_Flag).
graph_to_fa([Sym-Succ|Rest],Fs0,Fs,Ts0,Ts,Current,St0,St,Flag) :-
    graph_to_fa0(Flag,Sym,Succ,Fs0,Fs1,Ts0,Ts1,Current,St0,St1),
    graph_to_fa(Rest,Fs1,Fs,Ts1,Ts,Current,St1,St,Flag).

graph_to_fa0(r,Sym,Succ,Fs0,Fs,Ts0,Ts,Current,St0,St) :-
    (   Sym== ''
    ->  Succ=[],  %% sanity check
        Fs0=[Current|Fs],
	Ts0=Ts,
	St0=St
    ;   St1 is St0+1,
	Ts0=[trans(St1,Sym,Current)|Ts1],
	graph_to_fa(Succ,Fs0,Fs,Ts1,Ts,St1,St1,St,r)
    ).

graph_to_fa0(n,Sym,Succ,Fs0,Fs,Ts0,Ts,Current,St0,St) :-
    (   Sym== ''
    ->  Succ=[],  %% sanity check
        Fs0=[Current|Fs],
	Ts0=Ts,
	St0=St
    ;   St1 is St0+1,
	Ts0=[trans(Current,Sym,St1)|Ts1],
	graph_to_fa(Succ,Fs0,Fs,Ts1,Ts,St1,St1,St,n)
    ).



graph_to_fa_with_counts([],Fs,Fs,Ts,Ts,_,States,States).
graph_to_fa_with_counts([f(Sym,SymC,Succ)|Rest],Fs0,Fs,Ts0,Ts,Cur,St0,St) :-
    graph_to_fa0_with_counts(Sym,SymC,Succ,Fs0,Fs1,Ts0,Ts1,Cur,St0,St1),
    graph_to_fa_with_counts(Rest,Fs1,Fs,Ts1,Ts,Cur,St1,St).

graph_to_fa0_with_counts(Sym,SymC,Succ,Fs0,Fs,Ts0,Ts,Cur,St0,St) :-
    (   Sym== ''
    ->  Succ=[],  %% sanity check
	St is St0+1,
        Fs0=[St|Fs],
	Ts0=[trans(Cur,[]/SymC,St)|Ts]
    ;   St1 is St0+1,
	Ts0=[trans(Cur,Sym/SymC,St1)|Ts1],
	graph_to_fa_with_counts(Succ,Fs0,Fs,Ts1,Ts,St1,St1,St)
    ).



graph_to_fa_ph([],Fs,Fs,Ts,Ts,_,States,States,_,Nth,Nth,_Flag).
graph_to_fa_ph([Sym-Succ|Rest],Fs0,Fs,Ts0,Ts,C,St0,St,CN0,Nth0,Nth,Flag) :-
    graph_to_fa0_ph(Flag,Sym,Succ,Fs0,Fs1,Ts0,Ts1,C,St0,St1,CN0,Nth0,Nth1),
    CN is CN0+(Nth1-Nth0),
    graph_to_fa_ph(Rest,Fs1,Fs,Ts1,Ts,C,St1,St,CN,Nth1,Nth,Flag).

graph_to_fa0_ph(r,Sym,Succ,Fs0,Fs,Ts0,Ts,C,St0,St,CN,Nth0,Nth) :-
    (   Sym== ''
    ->  Succ==[], 
	CN=:=0,
        Fs0=[St0|Fs],
	Ts0=Ts,
	St0=St,
	Nth is Nth0+1
    ;	St1 is St0+1,
	Ts0=[trans(St1,Sym/CN,C)|Ts1],
	graph_to_fa_ph(Succ,Fs0,Fs,Ts1,Ts,St1,St1,St,0,Nth0,Nth,r)
    ).

graph_to_fa0_ph(n,Sym,Succ,Fs0,Fs,Ts0,Ts,C,St0,St,CN,Nth0,Nth) :-
    (   Sym== ''
    ->  Succ==[], 
	CN=:=0,
        Fs0=[St0|Fs],
	Ts0=Ts,
	St0=St,
	Nth is Nth0+1
    ;	St1 is St0+1,
	Ts0=[trans(C,Sym/CN,St1)|Ts1],
	graph_to_fa_ph(Succ,Fs0,Fs,Ts1,Ts,St1,St1,St,0,Nth0,Nth,n)
    ).

lines_to_key_lists([],[],_).
lines_to_key_lists([H0|T0],[H|T],M):-
    line_to_key_list(H0,H,M),
    lines_to_key_lists(T0,T,M).

%% '' is special atom to indicate end of word
%% we want an atom that is sorted before any other
line_to_key_list([],[''- []],_).
line_to_key_list([H|T0],[Hpred-T],M):-
    fsa_pred_module:regex_notation_to_predicate(M,H,Hpred),
    line_to_key_list(T0,T,M).

lines_to_key_lists_with_counts([],[],_).
lines_to_key_lists_with_counts([H0|T0],[H|T],M):-
    line_to_key_list_with_counts(H0,H,M),
    lines_to_key_lists_with_counts(T0,T,M).

%% '' is special atom to indicate end of word
%% we want an atom that is sorted before any other
line_to_key_list_with_counts([],[f('',1,[])],_).
line_to_key_list_with_counts([H|T0],[f(Hpred,1,T)],M):-
    fsa_pred_module:regex_notation_to_predicate(M,H,Hpred),
    line_to_key_list_with_counts(T0,T,M).

%% as union/2, but we have keylists, and we merge the lists associated
%% with identical keys as well.  
keymerge([], []).
keymerge([Set|Sets], Union) :-
    length([Set|Sets], NumberOfSets),
    keymerge_all(NumberOfSets, [Set|Sets], Union, []).

keymerge_all(N,Sets0,Union,Sets) :-
    (  N=:=1  -> Sets0=[Union|Sets]
    ;  N=:=2  -> Sets0=[Set1,Set2|Sets], 
                 keymerge(Set1,Set2,Union)
    ;  A is N>>1,
       Z is N-A,
       keymerge_all(A, Sets0, X, Sets1),
       keymerge_all(Z, Sets1, Y, Sets),
       keymerge(X, Y, Union)
    ).

%% much like keymerge, but pairs with the same key merge their values 
%% (which are ordered lists).
keymerge([],S,S).
keymerge([O|Os], Ns, Set) :- 
    keymerge(Ns, O, Os, Set).

keymerge([], O, Os, [O|Os]).
keymerge([N-Nlist|Ns], O-Olist, Os, Set) :-
    compare(C, O, N), 
    keymerge(C, O-Olist, Os, N-Nlist, Ns, Set).

keymerge(<, O1, Os, N, Ns, [O1|Set]) :- 
    keymerge(Os, N, Ns, Set).
keymerge(=, _-Olist, Os, N-Nlist, Ns, [N-Xlist|Set]) :- 
    keymerge(Olist,Nlist,Xlist),
    keymerge(Os,Ns,Set).
keymerge(>, O, Os, N1, Ns, [N1|Set]) :- 
    keymerge(Ns, O, Os, Set).


%% with counts
keymerge_with_counts([], []).
keymerge_with_counts([Set|Sets], Union) :-
    length([Set|Sets], NumberOfSets),
    keymerge_all_with_counts(NumberOfSets, [Set|Sets], Union, []).

keymerge_all_with_counts(N,Sets0,Union,Sets) :-
    (  N=:=1  -> Sets0=[Union|Sets]
    ;  N=:=2  -> Sets0=[Set1,Set2|Sets], 
                 keymerge_with_counts(Set1,Set2,Union)
    ;  A is N>>1,
       Z is N-A,
       keymerge_all_with_counts(A, Sets0, X, Sets1),
       keymerge_all_with_counts(Z, Sets1, Y, Sets),
       keymerge_with_counts(X, Y, Union)
    ).

%% much like keymerge, but pairs with the same key merge their values 
%% (which are ordered lists).
keymerge_with_counts([],S,S).
keymerge_with_counts([O|Os], Ns, Set) :- 
    keymerge_with_counts(Ns, O, Os, Set).

keymerge_with_counts([], O, Os, [O|Os]).
keymerge_with_counts([f(N,Nc,Nlist)|Ns], f(O,Oc,Olist), Os, Set) :-
    compare(C, O, N), 
    keymerge_with_counts(C, f(O,Oc,Olist), Os, f(N,Nc,Nlist), Ns, Set).

keymerge_with_counts(<, O1, Os, N, Ns, [O1|Set]) :- 
    keymerge_with_counts(Os, N, Ns, Set).
keymerge_with_counts(=, f(_,Oc,Olist), Os, f(N,Nc,Nlist),Ns,
		     [f(N,NewC,Xlist)|Set]) :-
    NewC is Oc+Nc,
    keymerge_with_counts(Olist,Nlist,Xlist),
    keymerge_with_counts(Os,Ns,Set).
keymerge_with_counts(>, O, Os, N1, Ns, [N1|Set]) :- 
    keymerge_with_counts(Ns, O, Os, Set).

%% get_file(File,Lines) as in Quintus library getfile,
%% except that we don't use the character codes, but the atoms 'a', 'b', etc.
get_file(File, Lines) :-
    see(File),
    prompt(Old,''),
    user:call_cleanup( fsa_dict:get_file(Lines),		       
			   (   seen, prompt(_,Old) ) ).

get_file(Lines) :-
    get_code(Char),
    (   Char  <   0
    ->	Lines = []
    ;   Char =:= 10
    ->	Lines = [[]|Rest],
	get_file(Rest)
    ;	Lines = [Line|Rest],
	get_file(Char, Line, Rest)
    ).

get_file(Char0, [Char|Line], Lines) :-
    atom_codes(Char,[Char0]),  % NB, this may be unexpected!
    get_code(Next),
    (	Next  >  10
    ->	get_file(Next, Line, Lines)
    ;   Next =:= 10
    ->	Line = [],
	get_file(Lines)
    ;   Next  <   0
    ->	Line = [],
	Lines = []
    ;	get_file(Next, Line, Lines)
    ).

counts_to_probs(G0,G) :-
    sum_counts(G0,0.0,Sum),
    counts_to_probs_(G0,Sum,G).

sum_counts([],S,S).
sum_counts([f(_,S1,_)|T],S0,S) :-
    S2 is S0+S1,
    sum_counts(T,S2,S).

counts_to_probs_([],_,[]).
counts_to_probs_([f(Sym,Count,Rest0)|T0],Sum,[f(Sym,Prob,Rest)|T]) :-
    (	Sum =:= Count
    ->	Prob = 0
    ;	Prob is -log(Count/Sum)/log(2)
    ),
    counts_to_probs(Rest0,Rest),
    counts_to_probs_(T0,Sum,T).

    