:- module(fsa_sync,[]).

#swi# :- style_check([-atom]).

#sic# :- use_module(library(lists), [ reverse/2 ]).
#yap# :- use_module(library(lists), [ reverse/2 ]).

% example of not-easily-synchronizable:
% t_determinize({[a:b,?,b],[a:c,class([x,y,z])x e]})

%% sync(+Fa0,-Fa)
%% Fa0 is a letter transducer possibly containing identity constraints
%% Fa is equivalent letter transducer where identity constraints are all
%% within single transition (synchronized). The first type of automata
%% are the result of t_determinize and t_minimize; the latter type of
%% automata are required as input to most operators.
%%
%% w_sync does the same, for weighted transducers

sync(Fa0,Fa):-
    fsa_data:fsa_symbols_decl(Fa0,t(ModA,ModB)),
    (	ModA \== ModB
    ->	Fa0=Fa			% no identity constraints possible
				% through different modules
    ;   fsa_data:fsa_start_states(Fa0,Starts),
	fsa_data:fsa_final_states(Fa0,Finals),
	fsa_data:fsa_states_number(Fa0,Size),
	fsa_data:fsa_jumps(Fa0,Jumps),
	(   Jumps \== []
	->  user:throw(fsa_error(
		"fsa_sync: input must not have jumps~n",[]))
	;   true
	),
	fsa_array:fsa_array_new(Done,Size),
	process(Starts,Fa0,Trans,Done,Size,transducer),
	fsa_data:fsa_construct_rename_states(t(ModA,ModB),Starts,Finals,Trans,[],Fa)
    ).

sync_w(Fa0,Fa):-
    fsa_data:fsa_symbols_decl(Fa0,t(ModA,ModB,fsa_frozen)),
    (	ModA \== ModB
    ->	Fa0=Fa			% no identity constraints possible
				% through different modules
    ;   fsa_data:fsa_start_states(Fa0,Starts),
	fsa_data:fsa_final_states(Fa0,Finals),
	fsa_data:fsa_states_number(Fa0,Size),
	fsa_data:fsa_jumps(Fa0,Jumps),
	(   Jumps \== []
	->  user:throw(fsa_error(
		"fsa_sync: input must not have jumps~n",[]))
	;   true
	),
	fsa_array:fsa_array_new(Done,Size),
	process(Starts,Fa0,Trans,Done,Size,w_transducer),
	fsa_data:fsa_construct_rename_states(
               t(ModA,ModB,fsa_frozen),Starts,Finals,Trans,[],Fa)
    ).

%% algorithm: maintain an agenda: set of states which are known to be
%%   synchronized (initialized by the start states)
%%   .for each synchronous state find all minimal synchronous paths
%%   .add target states of those paths to agenda (and create new
%%    synchroneous transitions for those paths)
%%
%% process(+Agenda,+Fa0,-Tr,?Done,+N0,transducer/w_transducer)
%% Agenda: synchroneous states yet to be considered
%% Fa0:    input automaton
%% Tr:     list of new transitions
%% Done:   states already added to Agenda
%% N0:     next available state number for new states
process([],_,[],_,_,_).
process([State|States],Fa0,Tr,Done,N0,Flag):-
    %% find all minimal synchroneous paths leaving State
    findall(Trans,fsa_sync:in_sync(Flag,State,Fa0,Trans),Transitions),
    %% add each path to Transitions and Agenda
    add(Transitions,State,States0,States,Tr,Tr1,Done,N0,N),
    process(States0,Fa0,Tr1,Done,N,Flag).

add([],_,S,S,Tr,Tr,_,N,N).
add([H|Tail],Source,St0,St,Tr0,Tr,Done,N0,N):-
    add1(H,Tail,Source,St0,St,Tr0,Tr,Done,N0,N).

add1(t(Target,Lseq,Rseq),Tail,Source,St0,St,Tr0,Tr,Done,N0,N) :-
    add_state(Target,St0,St1,Done),
    %% add synchr transitions for path
    add_seq(Lseq,Rseq,Source,Target,Tr0,Tr1,N0,N1),
    add(Tail,Source,St1,St,Tr1,Tr,Done,N1,N).

add1(t(Target,Lseq,Rseq,W),Tail,Source,St0,St,Tr0,Tr,Done,N0,N) :-
    add_state(Target,St0,St1,Done),
    %% add synchr transitions for path
    add_seq_w(Lseq,Rseq,W,Source,Target,Tr0,Tr1,N0,N1),
    add(Tail,Source,St1,St,Tr1,Tr,Done,N1,N).

add_state(Target,St0,St,Done):-
    fsa_array:fsa_array_access(Target,Val,Done),
    (	var(Val)
    ->	Val=done, St0=[Target|St]
    ;   St0=St
    ).

%% add transitions for a synchr. path
add_seq(Lseq,Rseq,P,Q,Tr0,Tr,N0,N):-
    %% divide synchr path into number of synchr steps
    make_sync(Lseq,Rseq,[H|T]),
    %% create transitions for each synchr step
    seq_to_trans(T,H,P,Q,Tr0,Tr,N0,N).

%% add transitions for a synchr. path
add_seq_w(Lseq,Rseq,W,P,Q,Tr0,Tr,N0,N):-
    %% divide synchr path into number of synchr steps
    make_sync(Lseq,Rseq,List),
    %% create transitions for each synchr step
    seq_to_trans_w(List,W,P,Q,Tr0,Tr,N0,N).

seq_to_trans([],H,P,Q,[trans(P,H,Q)|Tr],Tr,N0,N) :-
    (	var(P)
    ->	P=N0,
	N is N0+1
    ;   N = N0
    ).
    
seq_to_trans([New|T],H,P,Q,[trans(P,H,P1)|Tr0],Tr,N0,N):-
    (	var(P)
    ->	P=N0,
	N1 is N0+1
    ;   N1 = N0
    ),
    seq_to_trans(T,New,P1,Q,Tr0,Tr,N1,N).


seq_to_trans_w([],W,P,Q,[trans(P,[]/([]/W),Q)|Tr],Tr,N,N).
seq_to_trans_w([H|T],W,P,Q,Tr0,Tr,N0,N):-
    seq_to_trans_w(T,H,W,P,Q,Tr0,Tr,N0,N).

seq_to_trans_w([],LH/RH,W,P,Q,[trans(P,LH/(RH/W),Q)|Tr],Tr,N0,N) :-
    (	var(P)
    ->	P=N0,
	N is N0+1
    ;   N = N0
    ).
    
seq_to_trans_w([New|T],LH/RH,W,P,Q,[trans(P,LH/(RH/W),P1)|Tr0],Tr,N0,N):-
    (	var(P)
    ->	P=N0,
	N1 is N0+1
    ;   N1 = N0
    ),
    seq_to_trans_w0(T,New,P1,Q,Tr0,Tr,N1,N).

seq_to_trans_w0([],LH/RH,P,Q,[trans(P,LH/(RH/0),Q)|Tr],Tr,N0,N) :-
    (	var(P)
    ->	P=N0,
	N is N0+1
    ;   N = N0
    ).
    
seq_to_trans_w0([New|T],LH/RH,P,Q,[trans(P,LH/(RH/0),P1)|Tr0],Tr,N0,N):-
    (	var(P)
    ->	P=N0,
	N1 is N0+1
    ;   N1 = N0
    ),
    seq_to_trans_w0(T,New,P1,Q,Tr0,Tr,N1,N).


%% add []/Sym or Sym/[] pairs in order to synchronize.
%% make_sync(+Lseq,+Rseq,-List)
make_sync([],Right,List):-
    complete_right(Right,List).
make_sync([H|T],Right,List):-
    make_sync0(Right,[H|T],List).

make_sync0([],[H|T],List):-
    complete_left([H|T],List).
make_sync0([R|Rs0],[L|Ls0],List0):-
    consume_one(L,R,Ls0,Rs0,Ls,Rs,H),
    (	H==[]/[]
    ->	List0=List
    ;   List0=[H|List]
    ),
    make_sync(Ls,Rs,List).

%% add []/H transitions for list of H's
complete_right([],[]).
complete_right([H0|T0],List):-
    complete_right0(H0,List,T),
    complete_right(T0,T).

complete_right0(normal(H),List,T):-
    (	H==[]
    ->	List=T
    ;	List=[[]/H|T]
    ).

complete_left([],[]).
complete_left([H0|T0],List):-
    complete_left0(H0,List,T),
    complete_left(T0,T).

complete_left0(normal(H),List,T):-
    (	H==[]
    ->	List=T
    ;	List=[H/[]|T]
    ).

consume_one(normal(L),R1,Ls0,Rs0,Ls,Rs,Sym):-
    consume_one_normal(R1,L,Ls0,Rs0,Ls,Rs,Sym).
consume_one('$@'(L),R1,Ls0,Rs0,Ls,Rs,Sym):-
    consume_one_id(R1,L,Ls0,Rs0,Ls,Rs,Sym).

consume_one_normal(normal(R),L,Ls,Rs,Ls,Rs,L/R).
consume_one_normal('$@'(R),L,Ls,Rs0,Ls,['$@'(R)|Rs0],L/[]).

consume_one_id(normal(R),L,Ls0,Rs,['$@'(L)|Ls0],Rs,[]/R).
consume_one_id('$@'(T),T,Ls,Rs,Ls,Rs,'$@'(T)/'$@'(T)).
consume_one_id('$@dequeue',T,Ls,Rs,Ls,Rs,T/[]).

%% in_sync(State,Fa,Path,Flag)
%%     Trans is one minimal synchr. path leaving State
%%     where Path=t(Targetstate,SymbolsLeft,SymbolsRight)
%%
%% nondet algorithm: start with a transition leaving State
%% and then keep on extending the path until synchr.
%% (this works (terminates), because ...)
in_sync(transducer,State,Fa,t(Target,LeftList,RightList)):-
    fsa_data:fsa_transition(Fa,State,Left0/Right0,Next),
    ats(Left0,Left,0,Tleft),
    ats(Right0,Right,0,Tright),
    ensure_sync([Left],[Right],Tleft,Tright,Next,Target,LeftList,RightList,Fa).

in_sync(w_transducer,State,Fa,t(Target,LeftList,RightList,W)):-
    fsa_data:fsa_transition(Fa,State,Left0/(Right0/W0),Next),
    ats(Left0,Left,0,Tleft),
    ats(Right0,Right,0,Tright),
    ensure_sync([Left],[Right],Tleft,Tright,Next,Target,LeftList,RightList,Fa,
                W0,W).

ats(Sym,Pred,L0,L):-
    (	Sym = '$@'(_)
    ->	Pred=Sym,
	L is L0+1
    ;   Sym == '$@dequeue'
    ->  Pred=Sym,
	L is L0+1
    ;   Pred=normal(Sym),
	L0=L
    ).

ensure_sync(LeftL0,RightL0,Tleft,Tright,Target0,Target,LeftL,RightL,Fa):-
    Diff is Tleft-Tright,
    ensure_sync0(LeftL0,RightL0,Tleft,Tright,Target0,Target,LeftL1,RightL1,
		 Fa,[p(Diff,Target0)]),
    reverse(LeftL1,LeftL),
    reverse(RightL1,RightL).

%% we need loop-checking here, otherwise might not terminate.
%% -pm fsa_preds_i
%% -aux eacl99
%% letter_transducer(t_determinize(
%%    replace(e:a,[],[{b,c,d,f,g,h,j,k,l,m,n,p,q,r,s,t,v,w,x,y,z},a])))
%% 
ensure_sync0(LeftL0,RightL0,Tl0,Tr0,Target0,Target,LeftL,RightL,Fa,His):-
    %% synchr: if number of id constraints on left and right is the
    %%         same (because we start from synchr state); in that case
    %%         we shouldn't add another transition anymore, but stop
    (	Tl0=:=Tr0
    ->	LeftL0=LeftL,
	RightL0=RightL,
	Target0=Target
    ;   fsa_data:fsa_transition(Fa,Target0,Left0/Right0,Target1),
	ats(Left0,Left,Tl0,Tl),
	ats(Right0,Right,Tr0,Tr),
	Diff is Tl-Tr,
	\+ lists:memberchk(p(Diff,Target1),His),
	ensure_sync0([Left|LeftL0],[Right|RightL0],Tl,Tr,Target1,Target,
		     LeftL,RightL,Fa,[p(Diff,Target1)|His])
    ).

%% todo: loop check...
ensure_sync(LeftL0,RightL0,Tleft,Tright,Target0,Target,LeftL,RightL,Fa,W0,W):-
    Diff is Tleft-Tright,
    ensure_sync0(LeftL0,RightL0,Tleft,Tright,Target0,Target,LeftL1,RightL1,Fa,
		 W0,W,[p(Diff,Target0)]),
    reverse(LeftL1,LeftL),
    reverse(RightL1,RightL).

ensure_sync0(LeftL0,RightL0,Tl0,Tr0,Target0,Target,LeftL,RightL,Fa,W0,W,His):-
    %% synchr: if number of id constraints on left and right is the
    %%         same (because we start from synchr state); in that case
    %%         we shouldn't add another transition anymore, but stop
    (	Tl0=:=Tr0
    ->	LeftL0=LeftL,
	RightL0=RightL,
	Target0=Target,
	W0 = W
    ;   fsa_data:fsa_transition(Fa,Target0,Left0/(Right0/W1),Target1),
	ats(Left0,Left,Tl0,Tl),
	ats(Right0,Right,Tr0,Tr),
	W2 is W0+W1,
	Diff is Tl-Tr,
	\+ lists:memberchk(p(Diff,Target1),His),
	ensure_sync0([Left|LeftL0],[Right|RightL0],Tl,Tr,Target1,Target,
		     LeftL,RightL,Fa,W2,W,[p(Diff,Target1)|His])
    ).

