:- module(fsa_u_array, [ %% updatable arrays using copying
		       fsa_u_array_new/1,
		       fsa_u_array_new/2,
		       fsa_u_array_get/3,
                       fsa_u_array_get/4,
		       fsa_u_array_put/4,
                       fsa_u_array_put/5
		    ]).

#swi# :- style_check([-atom]).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

:- public help_info/4.
:- discontiguous help_info/4.

user:help_info(module,fsa_u_array,"Updatable Arrays (15+16 trees)",
"This module provides an updatable array datastructure. The arrays are \
 \n implemented using O'Keefe's N+K trees, with N=15 and K=16.\
 \n \
 \n NB. Array indices start at 0: so 0 refers to the first element of\
 \n the array.\
 \n \
 \n Here's an overview of the predicates provided:  \
 \n \
 \n * fsa_u_array_new/[1,2]       create a new updatable array\
 \n * fsa_u_array_get/[3,4]       lookup a value from an updatable array\
 \n * fsa_u_array_put/[4,5]       update a value in an updatable array\
 \n \
 \n The N+K tree data-structure is described in The Craft of Prolog, by \
 \n Richard A. O'Keefe, MIT Press, 1990, chapters 4.5. and 4.6.").

help_info(class,pred,"List of Predicates",
"This section lists the predicates defined by this module.").

help_info(pred,fsa_u_array_new,
"fsa_u_array_new(-UpdatableFsaArray[,?Size])",
"Initializes UpdatableFsaArray as a new mutable array. In this\
 \n implementation the optional Size argument is not used."). 

fsa_u_array_new(_).

fsa_u_array_new(_,_).

help_info(pred,fsa_u_array_get,
"fsa_u_array_get(+Index,?Val[,?Default,]+UpdatableFsaArray)",
"Val is unified with the Index'th entry in UpdatableFsaArray. "). 

help_info(pred,fsa_u_array_put,
"fsa_u_array_put(+Index[,?OldVal],?Val,+UpdatableFsaArray0,?UpdatableFsaArray)",
"The Index'th entry in UpdatableFsaArray0 is updated to Val; UpdatableFsaArray\
 \n is the resulting new array. OldVal is unified with the old value of\
 \n Index."). 

fsa_u_array_put(Index0,Val,Tree,NewTree):-
    Index is Index0+1,
    fsa_u_array_put0(Index,Val,Tree,NewTree,_OldVal).

fsa_u_array_put(Index0,OldVal,Val,Tree,NewTree):-
    Index is Index0+1,
    fsa_u_array_put0(Index,Val,Tree,NewTree,OldVal).

%% 15 + 16
fsa_u_array_put0(Index,Val,Tree,NewTree,OldVal):-
    functor(Tree,' nk ',31),
    (	Index < 16
    ->	replace_arg(Index,Tree,OldVal,NewTree,Val)
    ;   K is (Index /\ 15)+16,
	J is Index >> 4,
	replace_arg(K,Tree,Son,NewTree,NewSon),
	fsa_u_array_put0(J,Val,Son,NewSon,OldVal)
    ).

fsa_u_array_get(Index0,Val,Thing):-
    Index is Index0+1,
    fsa_u_array_get0(Index,Val,Thing).

fsa_u_array_get(Index0,Val,Default,Thing):-
    Index is Index0+1,
    fsa_u_array_get0(Index,Val0,Thing),
    (	var(Val0)
    ->	Val=Default
    ;   Val=Val0
    ).

%% 15 + 16
fsa_u_array_get0(Index,Val,Tree):-
    functor(Tree,' nk ',31),
    (	Index < 16
    ->	arg(Index,Tree,Val)
    ;   K is (Index /\ 15)+16,
	J is Index >> 4,
	arg(K,Tree,Son),
	fsa_u_array_get0(J,Val,Son)
    ).

replace_arg(1,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  A,
  ' nk '(XX,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(2,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  B,
  ' nk '(A,XX,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(3,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  C,
  ' nk '(A,B,XX,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(4,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  D,
  ' nk '(A,B,C,XX,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(5,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  E,
  ' nk '(A,B,C,D,XX,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(6,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  F,
  ' nk '(A,B,C,D,E,XX,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(7,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  G,
  ' nk '(A,B,C,D,E,F,XX,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(8,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  H,
  ' nk '(A,B,C,D,E,F,G,XX,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(9,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  I,
  ' nk '(A,B,C,D,E,F,G,H,XX,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(10,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  J,
  ' nk '(A,B,C,D,E,F,G,H,I,XX,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(11,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  K,
  ' nk '(A,B,C,D,E,F,G,H,I,J,XX,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(12,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  L,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,XX,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(13,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  M,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,XX,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(14,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  N,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,XX,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(15,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  O,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,XX,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(16,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  P,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,XX,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(17,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  Q,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,XX,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(18,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  R,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,XX,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(19,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  S,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,XX,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(20,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  T,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,XX,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(21,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  U,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,XX,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(22,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  V,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,XX,W,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(23,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  W,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,XX,X,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(24,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  X,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,XX,Y,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(25,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  Y,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,XX,Z,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(26,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  Z,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,XX,AA,BB,CC,DD,EE),
  XX).	    
replace_arg(27,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  AA,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,XX,BB,CC,DD,EE),
  XX).	    
replace_arg(28,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  BB,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,XX,CC,DD,EE),
  XX).	    
replace_arg(29,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  CC,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,XX,DD,EE),
  XX).	    
replace_arg(30,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  DD,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,XX,EE),
  XX).	    
replace_arg(31,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,EE),
  EE,
  ' nk '(A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,AA,BB,CC,DD,XX),
  XX).	    

:- multifile user:portray/1.

user:portray(X) :-
    functor(X,' nk ',31),
    !,
    write('<FsaUArray>').



