%% map_bbbtree - implements maps using bounded balanced binary trees.
%% adapted from the Mercury version. Adaption by Gertjan van Noord. The original is
%% available from http://www.cs.mu.oz.au/research/mercury/
%% Implementation based on "Functional Pearls: Efficient maps-a balancing act"
%% by Stephen Adams, J. Functional Programming 3 (4): 553-561, Oct 1993.

:- module(map_bbbtree, [ map_bbbtree__init/1,
			 map_bbbtree__empty/1,
			 map_bbbtree__size/2,
			 map_bbbtree__put/4,
			 map_bbbtree__put_list/3,
			 map_bbbtree__get/3,
			 map_bbbtree__least/3,
			 map_bbbtree__largest/3,
			 map_bbbtree__delete/3,
			 map_bbbtree__delete_list/3,
			 map_bbbtree__remove/3,
			 map_bbbtree__remove_list/3,
			 map_bbbtree__remove_least/4,
			 map_bbbtree__remove_largest/4,
			 map_bbbtree__list_to_map/2,
			 map_bbbtree__sorted_list_to_map/2,
			 map_bbbtree__sorted_list_to_map_len/3,
			 map_bbbtree__to_sorted_list/2
		       ]).

#swi# :- style_check([-atom]).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

:- discontiguous help_info/4.
:- public help_info/4.

user:help_info(module,map_bbbtree,"Balanced Binary Trees: Maps",
"This module implements maps using bounded balanced binary trees.\
 \n It is adapted from set_bbbtree, which itself is adapted from\
 \n the Mercury version. The original of that version is\
 \n available from http://www.cs.mu.oz.au/research/mercury/.\
 \n That implementation is based on `Functional Pearls: Efficient sets\
 \n -a balancing act' by Stephen Adams, J. Functional Programming 3 (4): \
 \n 553-561, Oct 1993.\
 \n \
 \n A map is a set of key/value pairs, such that each key is associated\
 \n with at most one value. Keys are required to be ground. The typical \
 \n operations on maps such as lookup the value of a given key are O(log n)\
 \n where n is the number of pairs in the map. A potentially more efficient\
 \n implementation of maps is provided by the *fsa_hash*, *fsa_m_hash* and\
 \n *fsa_u_hash* modules.").

help_info(class,pred,"List of Predicates",
"This section lists the predicates defined by this module.").

%% map_bbbtree__def_ratio(?Size)
map_bbbtree__def_ratio(5).

help_info(pred,map_bbbtree__init,
"map_bbbtree__init(?Bbbtree)",
"Initializes *Bbbtree* as an empty map.").
map_bbbtree__init(empty).

help_info(pred,map_bbbtree__empty,
"map_bbbtree__empty(?Bbbtree)",
"*Bbbtree* is an empty map.").
%% map_bbbtree__empty(?Bbbtree)
map_bbbtree__empty(empty).

help_info(pred,map_bbbtree__size,
"map_bbbtree__size(+Bbbtree,?Size)",
"*Size* is the number of pairs in map *Bbbtree*.").
map_bbbtree__size(empty,0).
map_bbbtree__size(tree(_K,_V,N,_L,_R),N).

help_info(pred,map_bbbtree__get,
"map_bbbtree__get(+Key,?Val,+Bbbtree)",
"*Val* is the value associated with *Key* in the map *Bbbtree*.\
 \n This predicate fails if *Key* is not a key of *Bbbtree*.").
map_bbbtree__get(Key,Val,tree(K,V,_N,L,R)) :-
    compare(Result,Key,K),
    map_bbbtree__get_(Result,Key,Val,V,L,R).

map_bbbtree__get_((<),Key,Val,_V,tree(K,V,_,L,R),_R)  :-
    compare(Result,Key,K),
    map_bbbtree__get_(Result,Key,Val,V,L,R).
map_bbbtree__get_((>),Key,Val,_V,_L,tree(K,V,_,L,R)) :-
    compare(Result,Key,K),
    map_bbbtree__get_(Result,Key,Val,V,L,R).
map_bbbtree__get_((=),_,Val,Val,_L,_R).


help_info(pred,map_bbbtree__least,
"map_bbbtree__least(+Bbbtree,?Least,?Val)",
"*Key* is the least key in *Bbbtree* (using the standard order\
 \n ordering of terms). Its value is *Val*.").
map_bbbtree__least(tree(_K,V,_N,L,_R),Key,Val) :-
    map_bbbtree__least_(L,V,Key,Val).

map_bbbtree__least_(empty,V,_,V).
map_bbbtree__least_(tree(_K0,V0,_N0,L0,_R0),_,Key,Val):-
    map_bbbtree__least_(L0,V0,Key,Val).

help_info(pred,map_bbbtree__largest,
"map_bbbtree__largest(+Bbbtree,?Largest,?Val)",
"*Key* is the largest key in *Bbbtree* (using the standard order\
 \n ordering of terms). Its value is *Val*.").
map_bbbtree__largest(tree(_K,V,_N,_L,R),Key,Val) :-
    map_bbbtree__largest_(R,V,Key,Val).

map_bbbtree__largest_(empty,V,_,V).
map_bbbtree__largest_(tree(_,V0,_,_,R0),_,Key,Val):-
    map_bbbtree__largest_(R0,V0,Key,Val).
    
help_info(pred,map_bbbtree__put,
"map_bbbtree__put(+Key,?Val,+Bbbtree0,-Bbbtree)",
"*Bbbtree* is the same map as *Bbbtree0*, except that *Key*\
 \n is now associated with *Val*.").
map_bbbtree__put(Key,Val,Map0,Map) :-
    map_bbbtree__def_ratio(Ratio),
    map_bbbtree__insert_r(Map0,Key,Val,Map,Ratio).

%% +Tree,+El,-Tree,+Ratio
map_bbbtree__insert_r(empty,Key,Val,tree(Key,Val,1,empty,empty),_Ratio).

map_bbbtree__insert_r(tree(K,V,N,L,R),Key,Val,Map,Ratio) :-
    compare(Result,Key,K),
    map_bbbtree__insert_r_(Result,K,V,N,L,R,Key,Val,Map,Ratio).

map_bbbtree__insert_r_((<),K,V,_N,L,R,Key,Val,Map,Ratio):-
    map_bbbtree__insert_r(L,Key,Val,NewL,Ratio),
    map_bbbtree__balance(K,V,NewL,R,Map,Ratio).
map_bbbtree__insert_r_((>),K,V,_N,L,R,Key,Val,Map,Ratio) :-
    map_bbbtree__insert_r(R,Key,Val,NewR,Ratio),
    map_bbbtree__balance(K,V,L,NewR,Map,Ratio).
map_bbbtree__insert_r_((=),K,V,N,L,R,K,V,tree(K,V,N,L,R),_).

help_info(pred,map_bbbtree__put_list,
"map_bbbtree__put_list(+Bbbtree0,+KeyValList,-Bbbtree",
"*Bbbtree* is the same map as *Bbbtree0*, except that each of the \
 \n key-value pairs in *KeyValList* are in *Bbbtree*. ").
map_bbbtree__put_list(Map0,List,Map) :-
    map_bbbtree__def_ratio(Ratio),
    map_bbbtree__insert_list_r(List,Map0,Map,Ratio).

%% +List,+Tree,-Tree,+Ratio
map_bbbtree__insert_list_r([],Map,Map,_Ratio).
map_bbbtree__insert_list_r([Key-X|Xs],Map0,Map,Ratio) :-
    map_bbbtree__insert_r(Map0,Key,X,Map1,Ratio),
    map_bbbtree__insert_list_r(Xs,Map1,Map,Ratio).

help_info(pred,map_bbbtree__delete,
"map_bbbtree__delete(+Bbbtree0,+Key,-Bbbtree)",
"*Bbbtree* is the result of removing *Key* and its associated\
 \n value from *Bbbtree0*. Succeeds if *Key* was not a key of \
 \n *Bbbtree0* (cf map_bbbtree__remove).").
map_bbbtree__delete(empty,_,empty).
map_bbbtree__delete(tree(K,V,N,L,R),Key,Map) :-
    compare(Result,Key,K),
    map_bbbtree__delete_(Result,K,V,N,L,R,Key,Map).

map_bbbtree__delete_((<),K,V,N,L,R,Key,tree(K,V,NewN,NewL,R)) :-
    NewN is N-1,
    map_bbbtree__delete(L,Key,NewL).
map_bbbtree__delete_((>),K,V,N,L,R,Key,tree(K,V,NewN,L,NewR)) :-
    NewN is N-1,
    map_bbbtree__delete(R,Key,NewR).
map_bbbtree__delete_((=),_K,_V,_N,L,R,_Key,Map):-
    map_bbbtree__concat3(L,R,Map).

help_info(pred,map_bbbtree__delete_list,
"map_bbbtree__delete_list(+Keys,+Bbbtree0,-Bbbtree)",
"*Bbbtree* is the result of deleting all keys *Keys* with \
 \n associated values from *Bbbtree0*. These keys are not\
 \n required to exist in *Bbbtree0* (cf map_bbbtree__remove_list).").
%% +List +Map -Map
map_bbbtree__delete_list([],Map,Map).
map_bbbtree__delete_list([X|Xs],Map0,Map) :-
    map_bbbtree__delete(Map0,X,Map1),
    map_bbbtree__delete_list(Xs,Map1,Map).

help_info(pred,map_bbbtree__remove,
"map_bbbtree__remove(+Bbbtree0,+Key,-Bbbtree)",
"*Bbbtree* is the result of removing *Key* and its associated\
 \n value from *Bbbtree0*. Fails if *Key* was not a key of \
 \n *Bbbtree0* (cf map_bbbtree__delete).").
map_bbbtree__remove(tree(K,V,N,L,R),X,Map) :-
    compare(Result,X,V),
    map_bbbtree__remove_(Result,K,V,N,L,R,X,Map).

map_bbbtree__remove_((<),K,V,N,L,R,X,tree(K,V,NewN,NewL,R)) :-
    NewN is N-1,
    map_bbbtree__remove(L,X,NewL).
map_bbbtree__remove_((>),K,V,N,L,R,X,tree(K,V,NewN,L,NewR)) :-
    NewN is N-1,
    map_bbbtree__remove(R,X,NewR).
map_bbbtree__remove_((=),_K,_V,_N,L,R,_X,Map):-
    map_bbbtree__concat3(L,R,Map).

help_info(pred,map_bbbtree__remove_list,
"map_bbbtree__remove_list(+Keys,+Bbbtree0,-Bbbtree)",
"*Bbbtree* is the result of removing all keys *Keys* with \
 \n associated values from *Bbbtree0*. These keys are\
 \n required to exist in *Bbbtree0* (cf map_bbbtree__delete_list).").
%% +List +Tree -Tree
map_bbbtree__remove_list([],Map,Map).
map_bbbtree__remove_list([X|Xs],Map0,Map) :-
    map_bbbtree__remove(Map0,X,Map1),
    map_bbbtree__remove_list(Xs,Map1,Map).

%% +Tree ?Least +Tree
help_info(pred,map_bbbtree__remove_least,
"map_bbbtree__remove_least(+Bbbtree0,?Key,?Val,-Bbbtree)",
"*Key* is the least key in *Bbbtree0* (using standard ordering of\
 \n terms). Its value is *Value*. *Bbbtree* is the same map as *Bbbtree0* \
 \n except that *Key* is removed.").
map_bbbtree__remove_least(tree(K,V,N,L,R),Key,Val,Map) :-
    map_bbbtree__remove_least_(L,K,V,N,R,Key,Val,Map).

map_bbbtree__remove_least_(empty,X,V,_,Map,X,V,Map).
map_bbbtree__remove_least_(tree(K1,V1,N1,L1,R1),K,V,N,R,Key,Val,
			   tree(K,V,NewN,NewL,R)):-
    NewN is N - 1,
    map_bbbtree__remove_least_(L1,K1,V1,N1,R1,Key,Val,NewL).

help_info(pred,map_bbbtree__remove_largest,
"map_bbbtree__remove_largest(+Bbbtree0,?Key,?Val,-Bbbtree)",
"*Key* is the largest key in *Bbbtree0* (using standard ordering of\
 \n terms). Its value is *Value*. *Bbbtree* is the same map as *Bbbtree0* \
 \n except that *Key* is removed.").
%% +Tree,?El,-Tree
map_bbbtree__remove_largest(tree(K,V,N,L,R),Key,Val,Map) :-
    map_bbbtree__remove_largest_(R,K,V,N,L,Key,Val,Map).

map_bbbtree__remove_largest_(empty,X,V,_N,Map,X,V,Map).
map_bbbtree__remove_largest_(tree(K1,V1,N1,L1,R1),K,V,N,L,Key,Val,
			     tree(K,V,NewN,L,NewR)):-
    NewN is N - 1,
    map_bbbtree__remove_largest_(R1,K1,V1,N1,L1,Key,Val,NewR).

help_info(pred,map_bbbtree__list_to_map,
"map_bbbtree__list_to_map(+KeyValList,-Bbbtree)",
"*Bbbtree* is the map for the key-value pairs given as a list\
 \n in *KeyValList*.").
%% +List -Tree
map_bbbtree__list_to_map(List,Map) :-
    map_bbbtree__def_ratio(Ratio),
    map_bbbtree__list_to_map_r(List,Map,Ratio).

%% +List -Tree +Ratio
map_bbbtree__list_to_map_r(List,Map,Ratio) :-
    map_bbbtree__init(InitMap),
    map_bbbtree__insert_list_r(List,InitMap,Map,Ratio).

%% The tree is created by first determining it's length. All lists of length
%% N have the same structure. The root of of the tree is the N // 2 element
%% of the list. Elements 1 to N // 2 - 1 make up the left subtree and elements
%% N // 2 + 1 to N make up the right subtree. The structure,which is known
%% due to the length of the list,is just created inorder while passing the
%% list around in an accumulator and taking off elements as needed.
%%
%% The predicate map_bbbtree__sorted_list_to_map_len2 has been unfolded up to
%% trees of size 3 so as to avoid recursive calls all the way to the leaves.
%% This alone approximately halves execution time and stack usage.
%% Unfolding further becomes a case of diminishing returns.
%%
%% Cases N = 3,N = 2 and N = 1 could safely be removed without change to
%% computed results as long as the 3 in N > 3 is adjusted appropriately.

help_info(pred,map_bbbtree__sorted_list_to_map,
"map_bbbtree__sorted_list_to_map(+SortedKeyValueList,-Bbbtree)",
"*SortedKeyValueList* is a sorted list of key value pairs; *Bbbtree*\
 \n is the corresponding map.").
map_bbbtree__sorted_list_to_map(List,Map) :-
    length(List,N),
    map_bbbtree__sorted_list_to_map_len(List,Map,N).

help_info(pred,map_bbbtree__sorted_list_to_map_len,
"map_bbbtree__sorted_list_to_map_len(+SortedKeyValueList,-Bbbtree,+Len)",
"*SortedKeyValueList* is a sorted list of key value pairs; *Bbbtree*\
 \n is the corresponding map. *Len* is the lenth of the list.").
%% +List,-Tree +N
map_bbbtree__sorted_list_to_map_len([],empty,_N).
map_bbbtree__sorted_list_to_map_len([X|Xs],Map,N) :-
    map_bbbtree__sorted_list_to_map_len2([X|Xs],[],N,Map).

map_bbbtree__sorted_list_to_map_len2(List,RestOfList,N,Map) :-
    (	N > 3
    ->	NL is N//2,
	NR is N - NL - 1,
	Map = tree(K,V,N,L,R),
	map_bbbtree__sorted_list_to_map_len2(List,[K-V|RestOfList1],NL,L),
	map_bbbtree__sorted_list_to_map_len2(RestOfList1,RestOfList,NR,R)
    ;	N =:= 3
    ->	List = [X-Xv,Y-Yv,Z-Zv|RestOfList],
	Map = tree(Y,Yv,N,tree(X,Xv,1,empty,empty),tree(Z,Zv,1,empty,empty))
    ;	N =:= 2
    ->	List = [X-Xv,Y-Yv|RestOfList],
	Map = tree(Y,Yv,N,tree(X,Xv,1,empty,empty),empty)
    ;	N =:= 1
    ->	List = [X-Xv|RestOfList],
	Map = tree(X,Xv,N,empty,empty)
    ;	RestOfList = List,
	Map = empty
    ).

help_info(pred,map_bbbtree__to_sorted_list,
"map_bbbtree__to_sorted_list(+Bbbtree,?SortedKeyValList)",
"*SortedKeyValList* is a sorted list of the key-value pairs in\
 \n the map *Bbbtree*.").
map_bbbtree__to_sorted_list(Map,List) :-
    map_bbbtree__to_sorted_list2(Map,[],List).

map_bbbtree__to_sorted_list2(empty,List,List).
map_bbbtree__to_sorted_list2(tree(K,V,_N,L,R),Acc,List) :-
    map_bbbtree__to_sorted_list2(R,Acc,List0),
    map_bbbtree__to_sorted_list2(L,[K-V|List0],List).

map_bbbtree__build_node(K,V,L,R,tree(K,V,N,L,R)) :-
    map_bbbtree__size(L,LSize),
    map_bbbtree__size(R,RSize),
    N is 1 + LSize + RSize.

map_bbbtree__single_rot_l(Ak,Av,X,tree(Bk,Bv,_N,Y,Z),Map) :-
    map_bbbtree__build_node(Ak,Av,X,Y,A_X_and_Y),
    map_bbbtree__build_node(Bk,Bv,A_X_and_Y,Z,Map).

map_bbbtree__single_rot_r(Bk,Bv,tree(Ak,Av,_N,X,Y),Z,Map) :-
    map_bbbtree__build_node(Bk,Bv,Y,Z,B_Y_and_Z),
    map_bbbtree__build_node(Ak,Av,X,B_Y_and_Z,Map).

map_bbbtree__double_rot_l(Ak,Av,X,tree(Ck,Cv,_N0,tree(Bk,Bv,_,Y1,Y2),Z),Map) :-
    map_bbbtree__build_node(Ak,Av,X,Y1,A_X_and_Y1),
    map_bbbtree__build_node(Ck,Cv,Y2,Z,C_Y2_and_Z),
    map_bbbtree__build_node(Bk,Bv,A_X_and_Y1,C_Y2_and_Z,Map).

map_bbbtree__double_rot_r(Ck,Cv,tree(Ak,Av,_N0,X,tree(Bk,Bv,_,Y1,Y2)),Z,Map) :-
    map_bbbtree__build_node(Ak,Av,X,Y1,A_X_and_Y1),
    map_bbbtree__build_node(Ck,Cv,Y2,Z,C_Y2_and_Z),
    map_bbbtree__build_node(Bk,Bv,A_X_and_Y1,C_Y2_and_Z,Map).

map_bbbtree__balance(K,V,L,R,Map,Ratio) :-
    map_bbbtree__size(L,LSize),
    map_bbbtree__size(R,RSize),
    Sum is LSize+RSize,
    (	Sum < 2
    ->	%% The two trees are too small to bother rebalancing
	map_bbbtree__build_node(K,V,L,R,Map)
    ;	RatioLSize is Ratio*LSize,
	(   RSize > RatioLSize
	->  R = tree(_K0,_V0,_N0,RL,RR),
	    map_bbbtree__size(RL,RLSize), % right side too big
	    map_bbbtree__size(RR,RRSize),
	    (	RLSize < RRSize		
	    ->	map_bbbtree__single_rot_l(K,V,L,R,Map)
	    ;	map_bbbtree__double_rot_l(K,V,L,R,Map)
	    )
	;   RatioRSize is Ratio*RSize,
	    (	LSize > RatioRSize
	    ->	L = tree(_K1,_V1,_N1,LL,LR),
		map_bbbtree__size(LL,LLSize), % left side too big
		map_bbbtree__size(LR,LRSize),
		(   LRSize < LLSize
		->  map_bbbtree__single_rot_r(K,V,L,R,Map)
		;   map_bbbtree__double_rot_r(K,V,L,R,Map)
		)
	    ;	map_bbbtree__build_node(K,V,L,R,Map) % Already balanced
	    )
	)
    ).

map_bbbtree__concat3(L,R,Map) :-
    map_bbbtree__size(L,LSize),
    (	LSize =:= 0		% Left tree empty so just
    ->	Map = R			% return the right tree
    ;	map_bbbtree__size(R,RSize),
	(   RSize =:= 0		% Right tree empty so just
	->  Map = L             % just return the left tree
	;   LSize > RSize	
	->  map_bbbtree__remove_largest(L,K,V,NewL),
	    map_bbbtree__build_node(K,V,NewL,R,Map)
	;   map_bbbtree__remove_least(R,K,V,NewR),
	    map_bbbtree__build_node(K,V,L,NewR,Map)
	)
    ).


