%% set_bbbtree - implements sets using bounded balanced binary trees.
%% adapted from the Mercury version. Adaption by Gertjan van Noord. The original is
%% available from http://www.cs.mu.oz.au/research/mercury/
%% Implementation based on "Functional Pearls: Efficient sets-a balancing act"
%% by Stephen Adams, J. Functional Programming 3 (4): 553-561, Oct 1993.

:- module(set_bbbtree, [ set_bbbtree__init/1,
			 set_bbbtree__empty/1,
			 set_bbbtree__non_empty/1,
			 set_bbbtree__size/2,
			 set_bbbtree__member/2,
			 set_bbbtree__is_member/3,
			 set_bbbtree__least/2,
			 set_bbbtree__largest/2,
			 set_bbbtree__singleton_set/2,
			 set_bbbtree__equal/2,
			 set_bbbtree__insert/3,
			 set_bbbtree__insert/4,
			 set_bbbtree__insert_list/3,
			 set_bbbtree__delete/3,
			 set_bbbtree__delete_list/3,
			 set_bbbtree__remove/3,
			 set_bbbtree__remove_list/3,
			 set_bbbtree__remove_least/3,
			 set_bbbtree__remove_largest/3,
			 set_bbbtree__list_to_set/2,
			 set_bbbtree__sorted_list_to_set/2,
			 set_bbbtree__sorted_list_to_set_len/3,
			 set_bbbtree__to_sorted_list/2,
			 set_bbbtree__union/3,
			 set_bbbtree__power_union/2,
			 set_bbbtree__intersect/3,
			 set_bbbtree__power_intersect/2,
			 set_bbbtree__difference/3,
			 set_bbbtree__subset/2,
			 set_bbbtree__superset/2
		       ]).

:- multifile user:help_info/4.
:- discontiguous user:help_info/4.

:- discontiguous help_info/4.
:- public help_info/4.

user:help_info(module,set_bbbtree,"Balanced Binary Trees: Sets",
"This module implements sets using bounded balanced binary trees.\
 \n It is adapted from the Mercury version. The original is\
 \n available from http://www.cs.mu.oz.au/research/mercury/.\
 \n That implementation is based on `Functional Pearls: Efficient sets\
 \n -a balancing act' by Stephen Adams, J. Functional Programming 3 (4): \
 \n 553-561, Oct 1993.").

help_info(class,pred,"List of Predicates",
"This section lists the predicates defined by this module.").

set_bbbtree__def_ratio(5).

help_info(pred,set_bbbtree__init,"set_bbbtree__init(?Bbbtree)",
"*Bbbtree* is initialized as an empty set.").
set_bbbtree__init(empty).

help_info(pred,set_bbbtree__empty,"set_bbbtree__empty(?Bbbtree)",
"Succeeds if *Bbbtree* is the empty set.").
set_bbbtree__empty(empty).

help_info(pred,set_bbbtree__non_empty,"set_bbbtree__non_empty(?Bbbtree)",
"Succeeds if *Bbbtree* is a non-empty set.").
set_bbbtree__non_empty(tree(_,_,_,_)).

help_info(pred,set_bbbtree__size,"set_bbbtree__size(+Bbbtree,?Integer)",
"*Integer* is the cardinality of the set *Bbbtree*.").
set_bbbtree__size(empty,0).
set_bbbtree__size(tree(_V,N,_L,_R),N).

help_info(pred,set_bbbtree__is_member,
"set_bbbtree__is_member(+El,+Bbbtree,?Bool)",
"Bool is the atom *yes* if *El* is an element of *Bbbtree*. Otherwise \
 \n it is the atom *no*.").
set_bbbtree__is_member(X,Tree,Bool):-
    set_bbbtree__is_member_(Tree,X,Bool).

set_bbbtree__is_member_(empty,_,no).
set_bbbtree__is_member_(tree(V,_N,L,R),X,Bool) :-
    compare(Result,X,V),
    set_bbbtree__is_member__(Result,X,V,L,R,Bool).

set_bbbtree__is_member__((<),X,_V,L,_R,Bool)  :-
    set_bbbtree__is_member_(L,X,Bool).
set_bbbtree__is_member__((>),X,_V,_L,R,Bool) :-
    set_bbbtree__is_member_(R,X,Bool).
set_bbbtree__is_member__((=),X,X,_L,_R,yes).

help_info(pred,set_bbbtree__member,
"set_bbbtree__member(?El,+Bbbtree)",
"*El* is an element of *Bbbtree*. Can be used to enumerate all elements\
 \n of *Bbbtree*.").
set_bbbtree__member(El,Tree):-
    (	ground(El)
    ->	set_bbbtree__is_member_(Tree,El,yes)
    ;   set_bbbtree__enum_member(Tree,El)
    ).

set_bbbtree__enum_member(tree(V,_,L,R),El):-
    (	V=El
    ;   set_bbbtree__enum_member(L,El)
    ;   set_bbbtree__enum_member(R,El)
    ).

help_info(pred,set_bbbtree__least,
"set_bbbtree__least(+Bbbtree,?El)",
"*El* is the least element occurring in *Bbbtree*, using the\
 \n standard ordering of terms.").
set_bbbtree__least(tree(V,_N,L,_R),X) :-
    set_bbbtree__least_(L,V,X).

set_bbbtree__least_(empty,V,V).
set_bbbtree__least_(tree(V0,_N0,L0,_R0),_,X):-
    set_bbbtree__least_(L0,V0,X).

help_info(pred,set_bbbtree__largest,
"set_bbbtree__largest(+Bbbtree,?El)",
"*El* is the largest element occurring in *Bbbtree*, using the\
 \n standard ordering of terms.").
set_bbbtree__largest(tree(V,_N,_L,R),X) :-
    set_bbbtree__largest_(R,V,X).

set_bbbtree__largest_(empty,V,V).
set_bbbtree__largest_(tree(V0,_,_,R0),_,X):-
    set_bbbtree__largest_(R0,V0,X).

help_info(pred,set_bbbtree__singleton_set,
"set_bbbtree__singleton_set(?BbbTree,?El)",
"*Bbbtree* is a set with single element *El*.").
set_bbbtree__singleton_set(tree(V,1,empty,empty),V).

help_info(pred,set_bbbtree__equal,
"set_bbbtree__equal(+BbbtreeA,+BbbtreeB)",
"*BbbtreeA* and *BbbtreeB* are the same sets.").
set_bbbtree__equal(SetA,SetB) :-
    set_bbbtree__subset(SetA,SetB),
    set_bbbtree__subset(SetB,SetA).

help_info(pred,set_bbbtree__insert,
"set_bbbtree__insert(+BbbtreeA,+El,-BbbtreeB[,?New])",
"*BbbtreeB* is the result of inserting *El* in *BbbtreeA*. The \
 \n optional fourth argument is the atom *yes* if *El* is not an\
 \n element of *BbbtreeA*; otherwise it is the atom *no*.").
%% +Tree,+El,-Tree
set_bbbtree__insert(Set0,X,Set) :-
    set_bbbtree__insert(Set0,X,Set,_).

set_bbbtree__insert(Set0,X,Set,New):-
    set_bbbtree__def_ratio(Ratio),
    set_bbbtree__insert_r(Set0,X,Set,Ratio,New).

%% +Tree,+El,-Tree,+Ratio
set_bbbtree__insert_r(empty,X,tree(X,1,empty,empty),_Ratio,yes).
set_bbbtree__insert_r(tree(V,N,L,R),X,Set,Ratio,New) :-
    compare(Result,X,V),
    set_bbbtree__insert_r_(Result,V,N,L,R,X,Set,Ratio,New).

set_bbbtree__insert_r_((<),V,_N,L,R,X,Set,Ratio,New):-
    set_bbbtree__insert_r(L,X,NewL,Ratio,New),
    set_bbbtree__balance(V,NewL,R,Set,Ratio).
set_bbbtree__insert_r_((>),V,_N,L,R,X,Set,Ratio,New) :-
    set_bbbtree__insert_r(R,X,NewR,Ratio,New),
    set_bbbtree__balance(V,L,NewR,Set,Ratio).
set_bbbtree__insert_r_((=),V,N,L,R,_,tree(V,N,L,R),_,no).

help_info(pred,set_bbbtree__insert_list,
"set_bbbtree__insert_list(+BbbtreeA,+List,-BbbtreeB)",
"*BbbtreeB* is the result of inserting each of the elements in *List*\
 \n to *BbbtreeA*.").

%% +Tree +List -Tree
set_bbbtree__insert_list(Set0,List,Set) :-
    set_bbbtree__def_ratio(Ratio),
    set_bbbtree__insert_list_r(List,Set0,Set,Ratio).

%% +List,+Tree,-Tree,+Ratio
set_bbbtree__insert_list_r([],Set,Set,_Ratio).
set_bbbtree__insert_list_r([X|Xs],Set0,Set,Ratio) :-
    set_bbbtree__insert_r(Set0,X,Set1,Ratio,_),
    set_bbbtree__insert_list_r(Xs,Set1,Set,Ratio).

help_info(pred,set_bbbtree__delete,
"set_bbbtree__delete(+BbbtreeA,+El,-BbbtreeB",
"*BbbtreeB* is the result of removing the element *El* from \
 \n *BbbtreeA*. The predicate *succeeds* if *El* is not an element of\
 \n *BbbtreeA* (cf. set_bbbtree__remove).").
set_bbbtree__delete(empty,_X,empty).
set_bbbtree__delete(tree(V,N,L,R),X,Set) :-
    compare(Result,X,V),
    set_bbbtree__delete_(Result,V,N,L,R,X,Set).

set_bbbtree__delete_((<),V,N,L,R,X,tree(V,NewN,NewL,R)) :-
    NewN is N-1,
    set_bbbtree__delete(L,X,NewL).
set_bbbtree__delete_((>),V,N,L,R,X,tree(V,NewN,L,NewR)) :-
    NewN is N-1,
    set_bbbtree__delete(R,X,NewR).
set_bbbtree__delete_((=),_V,_N,L,R,_X,Set):-
    set_bbbtree__concat3(L,R,Set).

help_info(pred,set_bbbtree__delete_list,
"set_bbbtree__delete_list(+List,+BbbtreeA,-BbbtreeB)",
"*BbbtreeB* is the result of deleting each of the elements of\
 \n *List* from *BbbtreeA*. The elements are *not* required to be\
 \n contained in *BbbtreeA* (cf. set_bbbtree__remove_list).").
%% +List +Set -Set
set_bbbtree__delete_list([],Set,Set).
set_bbbtree__delete_list([X|Xs],Set0,Set) :-
    set_bbbtree__delete(Set0,X,Set1),
    set_bbbtree__delete_list(Xs,Set1,Set).

help_info(pred,set_bbbtree__remove,
"set_bbbtree__remove(+BbbtreeA,+El,-BbbtreeB",
"*BbbtreeB* is the result of removing the element *El* from \
 \n *BbbtreeA*. The predicate *fails* if *El* is not an element of\
 \n *BbbtreeA* (cf. set_bbbtree__delete)").
set_bbbtree__remove(tree(V,N,L,R),X,Set) :-
    compare(Result,X,V),
    set_bbbtree__remove_(Result,V,N,L,R,X,Set).

set_bbbtree__remove_((<),V,N,L,R,X,tree(V,NewN,NewL,R)) :-
    NewN is N-1,
    set_bbbtree__remove(L,X,NewL).
set_bbbtree__remove_((>),V,N,L,R,X,tree(V,NewN,L,NewR)) :-
    NewN is N-1,
    set_bbbtree__remove(R,X,NewR).
set_bbbtree__remove_((=),_V,_N,L,R,_X,Set):-
    set_bbbtree__concat3(L,R,Set).

help_info(pred,set_bbbtree__remove_list,
"set_bbbtree__remove_list(+List,+BbbtreeA,-BbbtreeB)",
"*BbbtreeB is the result of deleting each of the elements of\
 \n *List* from *BbbtreeA*. The elements *are* required to be\
 \n contained in *BbbtreeA* (cf. set_bbbtree__delete_list).").
set_bbbtree__remove_list([],Set,Set).
set_bbbtree__remove_list([X|Xs],Set0,Set) :-
    set_bbbtree__remove(Set0,X,Set1),
    set_bbbtree__remove_list(Xs,Set1,Set).

help_info(pred,set_bbbtree__remove_least,
"set_bbbtree__remove_least(+BbbtreeA,?Least,-BbbtreeB)",
"*BbbtreeB* is the result of removing the least element *Least* from\
 \n *BbbtreeA*, in the standard ordering of terms.").
set_bbbtree__remove_least(tree(V,N,L,R),X,Set) :-
    set_bbbtree__remove_least_(L,V,N,R,X,Set).

set_bbbtree__remove_least_(empty,X,_,Set,X,Set).
set_bbbtree__remove_least_(tree(V1,N1,L1,R1),V,N,R,X,tree(V,NewN,NewL,R)):-
    NewN is N - 1,
    set_bbbtree__remove_least_(L1,V1,N1,R1,X,NewL).

help_info(pred,set_bbbtree__remove_largest,
"set_bbbtree__remove_largest(+BbbtreeA,?Largest,-BbbtreeB)",
"*BbbtreeB* is the result of removing the largest element *Largest* from\
 \n *BbbtreeA*, in the standard ordering of terms.").
set_bbbtree__remove_largest(tree(V,N,L,R),X,Set) :-
    set_bbbtree__remove_largest_(R,V,N,L,X,Set).

set_bbbtree__remove_largest_(empty,X,_N,Set,X,Set).
set_bbbtree__remove_largest_(tree(V1,N1,L1,R1),V,N,L,X,tree(V,NewN,L,NewR)):-
    NewN is N - 1,
    set_bbbtree__remove_largest_(R1,V1,N1,L1,X,NewR).

help_info(pred,set_bbbtree__list_to_set,
"set_bbbtree__list_to_set(+List,-Bbbtree)",
"*Bbbtree* is a set containing precisely all elements of *List*.").
%% +List -Tree
set_bbbtree__list_to_set(List,Set) :-
    set_bbbtree__def_ratio(Ratio),
    set_bbbtree__list_to_set_r(List,Set,Ratio).

%% +List -Tree +Ratio
set_bbbtree__list_to_set_r(List,Set,Ratio) :-
    set_bbbtree__init(InitSet),
    set_bbbtree__insert_list_r(List,InitSet,Set,Ratio).

%% The tree is created by first determining it's length. All lists of length
%% N have the same structure. The root of of the tree is the N // 2 element
%% of the list. Elements 1 to N // 2 - 1 make up the left subtree and elements
%% N // 2 + 1 to N make up the right subtree. The structure,which is known
%% due to the length of the list,is just created inorder while passing the
%% list around in an accumulator and taking off elements as needed.
%%
%% The predicate set_bbbtree__sorted_list_to_set_len2 has been unfolded up to
%% trees of size 3 so as to avoid recursive calls all the way to the leaves.
%% This alone approximately halves execution time and stack usage.
%% Unfolding further becomes a case of diminishing returns.
%%
%% Cases N = 3,N = 2 and N = 1 could safely be removed without change to
%% computed results as long as the 3 in N > 3 is adjusted appropriately.

help_info(pred,set_bbbtree__sorted_list_to_set,
"set_bbbtree__sorted_list_to_set(+SortedList,?Bbbtree)",
"*Bbbtree* is the set containing precisely the elements of\
 \n *SortedList*.").

%% +List, -Tree
set_bbbtree__sorted_list_to_set(List,Set) :-
    length(List,N),
    set_bbbtree__sorted_list_to_set_len(List,Set,N).

help_info(pred,set_bbbtree__sorted_list_to_set_len,
"set_bbbtree__sorted_list_to_set_len(+SortedList,?Bbbtree,+Len)",
"*Bbbtree* is the set containing precisely the elements of\
 \n *SortedList*. *Len* is the length of *SortedList*.").
%% +List,-Tree +N
set_bbbtree__sorted_list_to_set_len([],empty,_N).
set_bbbtree__sorted_list_to_set_len([ X | Xs ],Set,N) :-
    set_bbbtree__sorted_list_to_set_len2([ X | Xs ],[],N,Set).

set_bbbtree__sorted_list_to_set_len2(List,RestOfList,N,Set) :-
    (	N > 3
    ->	NL is N//2,
	NR is N - NL - 1,
	Set = tree(V,N,L,R),
	set_bbbtree__sorted_list_to_set_len2(List,[V|RestOfList1],NL,L),
	set_bbbtree__sorted_list_to_set_len2(RestOfList1,RestOfList,NR,R)
    ;	N =:= 3
    ->	List = [ X,Y,Z | RestOfList ],
	Set = tree(Y,N,tree(X,1,empty,empty),tree(Z,1,empty,empty))
    ;	N =:= 2
    ->	List = [ X,Y | RestOfList ],
	Set = tree(Y,N,tree(X,1,empty,empty),empty)
    ;	N =:= 1
    ->	List = [ X | RestOfList ],
	Set = tree(X,N,empty,empty)
    ;	RestOfList = List,
	Set = empty
    ).

help_info(pred,set_bbbtree__to_sorted_list,
"set_bbbtree__to_sorted_list(+Bbbtree,?SortedList)",
"*SortedList* is a sorted list of the elements of the set *Bbbtree*.").

set_bbbtree__to_sorted_list(Set,List) :-
    set_bbbtree__to_sorted_list2(Set,[],List).

set_bbbtree__to_sorted_list2(empty,List,List).
set_bbbtree__to_sorted_list2(tree(V,_N,L,R),Acc,List) :-
    set_bbbtree__to_sorted_list2(R,Acc,List0),
    set_bbbtree__to_sorted_list2(L,[V|List0],List).

help_info(pred,set_bbbtree__union,
"set_bbbtree__union(+BbbtreeA,+BbbtreeB,-BbbtreeC)",
"*BbbtreeC* is the union of *BbbtreeA* and *BbbtreeB*.").
set_bbbtree__union(SetA,SetB,Set) :-
    set_bbbtree__def_ratio(Ratio),
    set_bbbtree__union_r(SetA,SetB,Set,Ratio).

set_bbbtree__union_r(empty,Set,Set,_Ratio).
set_bbbtree__union_r(tree(V,_N,LL,LR),R,Set,Ratio) :-
    set_bbbtree__split_lt(R,V,NewRL,Ratio),
    set_bbbtree__split_gt(R,V,NewRR,Ratio),
    set_bbbtree__union_r(LL,NewRL,LSet,Ratio),
    set_bbbtree__union_r(LR,NewRR,RSet,Ratio),
    set_bbbtree__concat4(LSet,RSet,V,Set,Ratio).


help_info(pred,set_bbbtree__power_union,
"set_bbbtree__power_union(+Bbbtrees,-BbbtreeC)",
"*Bbbtrees* is a set of sets. *BbbtreeC* is the union of all\
 \n of of these sets.").
set_bbbtree__power_union(Sets,Set) :-
    set_bbbtree__def_ratio(Ratio),
    set_bbbtree__power_union_r(Sets,Set,Ratio).

set_bbbtree__power_union_r(empty,empty,_Ratio).
set_bbbtree__power_union_r(tree(V,_N,L,R),Set,Ratio) :-
    set_bbbtree__power_union_r(L,LUnion,Ratio),
    set_bbbtree__power_union_r(R,RUnion,Ratio),
    set_bbbtree__union_r(LUnion,RUnion,Union,Ratio),
    set_bbbtree__union_r(V,Union,Set,Ratio).

help_info(pred,set_bbbtree__intersect,
"set_bbbtree__intersect(+BbbtreeA,+BbbtreeB,-BbbtreeC)",
"*BbbtreeC* is the intersection of *BbbtreeA* and *BbbtreeB*.").
set_bbbtree__intersect(SetA,SetB,Set) :-
    set_bbbtree__def_ratio(Ratio),
    set_bbbtree__intersect_r(SetA,SetB,Set,Ratio).

set_bbbtree__intersect_r(empty,_Set,empty,_Ratio).
set_bbbtree__intersect_r(tree(V,_N,LL,LR),R,Set,Ratio) :-
    set_bbbtree__split_lt(R,V,NewRL,Ratio),
    set_bbbtree__split_gt(R,V,NewRR,Ratio),
    set_bbbtree__intersect_r(LL,NewRL,LSet,Ratio),
    set_bbbtree__intersect_r(LR,NewRR,RSet,Ratio),
    set_bbbtree__is_member(V,R,Bool),
    set_bbbtree__intersect_r_(Bool,LSet,RSet,V,Set,Ratio).

set_bbbtree__intersect_r_(yes,LSet,RSet,V,Set,Ratio):-
    set_bbbtree__concat4(LSet,RSet,V,Set,Ratio).
set_bbbtree__intersect_r_(no,LSet,RSet,_,Set,_):-
    set_bbbtree__concat3(LSet,RSet,Set).

help_info(pred,set_bbbtree__power_intersect,
"set_bbbtree__power_intersect(+Bbbtrees,-BbbtreeC)",
"*Bbbtrees* is a set of sets. *BbbtreeC* is the set\
 \n containing the elements which occur in each of *Bbbtrees*").
set_bbbtree__power_intersect(Sets,Set) :-
    set_bbbtree__def_ratio(Ratio),
    set_bbbtree__power_intersect_r(Sets,Set,Ratio).

set_bbbtree__power_intersect_r(empty,empty,_Ratio).
set_bbbtree__power_intersect_r(tree(V,_N,L,R),Set,Ratio) :-
    set_bbbtree__power_intersect_r2(L,V,Intersection0,Ratio),
    set_bbbtree__power_intersect_r2(R,Intersection0,Set,Ratio).


set_bbbtree__power_intersect_r2(empty,empty,empty,_Ratio).
set_bbbtree__power_intersect_r2(empty,tree(_V,_N,_L,_R),empty,_Ratio).
set_bbbtree__power_intersect_r2(tree(_V,_N,_L,_R),empty,empty,_Ratio).
set_bbbtree__power_intersect_r2(tree(V,_N,L,R),tree(AV,AN,AL,AR),Set,Ratio) :-
    set_bbbtree__intersect_r(V,tree(AV,AN,AL,AR),Intersection0,Ratio),
    set_bbbtree__power_intersect_r2(L,Intersection0,Intersection1,Ratio),
    set_bbbtree__power_intersect_r2(R,Intersection1,Set,Ratio).

help_info(pred,set_bbbtree__difference,
"set_bbbtree__difference(+BbbtreeA,+BbbtreeB,-BbbtreeC)",
"*BbbtreeC* is the set *BbbtreeA* minus all elements of *BbbtreeB*.").
set_bbbtree__difference(SetA,SetB,Set) :-
    set_bbbtree__def_ratio(Ratio),
    set_bbbtree__difference_r(SetA,SetB,Set,Ratio).

set_bbbtree__difference_r(empty,_Set,empty,_Ratio).
set_bbbtree__difference_r(tree(V,_N,LL,LR),R,Set,Ratio) :-
    set_bbbtree__split_lt(R,V,NewRL,Ratio),
    set_bbbtree__split_gt(R,V,NewRR,Ratio),
    set_bbbtree__difference_r(LL,NewRL,LSet,Ratio),
    set_bbbtree__difference_r(LR,NewRR,RSet,Ratio),
    set_bbbtree__is_member(V,R,Bool),
    set_bbbtree__difference_r_(Bool,LSet,RSet,V,Set,Ratio).

set_bbbtree__difference_r_(yes,LSet,RSet,_,Set,_):-
    set_bbbtree__concat3(LSet,RSet,Set).
set_bbbtree__difference_r_(no,LSet,RSet,V,Set,Ratio):-
    set_bbbtree__concat4(LSet,RSet,V,Set,Ratio).

help_info(pred,set_bbbtree__subset,
"set_bbbtree__subset(+BbbtreeA,+BbbtreeB)",
"*BbbtreeA* is a subset of *BbbtreeB*.").
set_bbbtree__subset(SetA,SetB) :-
    set_bbbtree__difference(SetA,SetB,Set),
    set_bbbtree__empty(Set).

help_info(pred,set_bbbtree__superset,
"set_bbbtree__superset(+BbbtreeA,+BbbtreeB)",
"*BbbtreeA* is a superset of *BbbtreeB*.").
set_bbbtree__superset(SetA,SetB) :-
    set_bbbtree__subset(SetB,SetA).

set_bbbtree__build_node(X,L,R,tree(X,N,L,R)) :-
    set_bbbtree__size(L,LSize),
    set_bbbtree__size(R,RSize),
    N is 1 + LSize + RSize.

set_bbbtree__single_rot_l(A,X,tree(B,_N,Y,Z),Set) :-
    set_bbbtree__build_node(A,X,Y,A_X_and_Y),
    set_bbbtree__build_node(B,A_X_and_Y,Z,Set).

set_bbbtree__single_rot_r(B,tree(A,_N,X,Y),Z,Set) :-
    set_bbbtree__build_node(B,Y,Z,B_Y_and_Z),
    set_bbbtree__build_node(A,X,B_Y_and_Z,Set).

set_bbbtree__double_rot_l(A,X,tree(C,_N0,tree(B,_,Y1,Y2),Z),Set) :-
    set_bbbtree__build_node(A,X,Y1,A_X_and_Y1),
    set_bbbtree__build_node(C,Y2,Z,C_Y2_and_Z),
    set_bbbtree__build_node(B,A_X_and_Y1,C_Y2_and_Z,Set).

set_bbbtree__double_rot_r(C,tree(A,_N0,X,tree(B,_,Y1,Y2)),Z,Set) :-
    set_bbbtree__build_node(A,X,Y1,A_X_and_Y1),
    set_bbbtree__build_node(C,Y2,Z,C_Y2_and_Z),
    set_bbbtree__build_node(B,A_X_and_Y1,C_Y2_and_Z,Set).

set_bbbtree__balance(V,L,R,Set,Ratio) :-
    set_bbbtree__size(L,LSize),
    set_bbbtree__size(R,RSize),
    Sum is LSize+RSize,
    (	Sum < 2
    ->	%% The two trees are too small to bother rebalancing
	set_bbbtree__build_node(V,L,R,Set)
    ;	RatioLSize is Ratio * LSize,
	(   RSize > RatioLSize
	->  R = tree(_V0,_N0,RL,RR),
	    set_bbbtree__size(RL,RLSize), % right side too big
	    set_bbbtree__size(RR,RRSize),
	    (	RLSize < RRSize		
	    ->	set_bbbtree__single_rot_l(V,L,R,Set)
	    ;	set_bbbtree__double_rot_l(V,L,R,Set)
	    )
	;   RatioRSize is Ratio * RSize,
	    (	LSize > RatioRSize
	    ->	L = tree(_V1,_N1,LL,LR),
		set_bbbtree__size(LL,LLSize), % left side too big
		set_bbbtree__size(LR,LRSize),
		(   LRSize < LLSize
		->  set_bbbtree__single_rot_r(V,L,R,Set)
		;   set_bbbtree__double_rot_r(V,L,R,Set)
		)
	    ;	set_bbbtree__build_node(V,L,R,Set) % Already balanced
	    )
	)
    ).

set_bbbtree__concat3(L,R,Set) :-
    set_bbbtree__size(L,LSize),
    (	LSize =:= 0		% Left tree empty so just
    ->	Set = R			% return the right tree
    ;	set_bbbtree__size(R,RSize),
	(   RSize =:= 0		% Right tree empty so just
	->  Set = L             % just return the left tree
	;   LSize > RSize	
	->  set_bbbtree__remove_largest(L,X,NewL),
	    set_bbbtree__build_node(X,NewL,R,Set)
	;   set_bbbtree__remove_least(R,X,NewR),
	    set_bbbtree__build_node(X,L,NewR,Set)
	)
    ).

set_bbbtree__concat4(empty,R,V,Set,Ratio) :-
    set_bbbtree__insert_r(R,V,Set,Ratio,_).
set_bbbtree__concat4(tree(LV,LN,LL,LR),R,V,Set,Ratio) :-
    set_bbbtree__concat4_(R,LV,LN,LL,LR,V,Set,Ratio).

set_bbbtree__concat4_(empty,LV,LN,LL,LR,V,Set,Ratio):-
    set_bbbtree__insert_r(tree(LV,LN,LL,LR),V,Set,Ratio,_).
set_bbbtree__concat4_(tree(RV,RN,RL,RR),LV,LN,LL,LR,V,Set,Ratio):-
    RatioLN is Ratio*LN,
    (	RatioLN < RN
    ->	set_bbbtree__concat4_(RL,LV,LN,LL,LR,V,NewL,Ratio),
	set_bbbtree__balance(RV,NewL,RR,Set,Ratio)
    ;	RatioRN is Ratio * RN,
	(   RatioRN < LN
	->  set_bbbtree__concat4(LR,tree(RV,RN,RL,RR),V,NewR,Ratio),
	    set_bbbtree__balance(LV,LL,NewR,Set,Ratio)
	;   set_bbbtree__build_node(V,tree(LV,LN,LL,LR),tree(RV,RN,RL,RR),Set)
	)
    ).

set_bbbtree__split_lt(empty,_X,empty,_Ratio).
set_bbbtree__split_lt(tree(V,_N,L,R),X,Set,Ratio) :-
    compare(Result,X,V),
    set_bbbtree__split_lt_(Result,V,L,R,X,Set,Ratio).

set_bbbtree__split_lt_((<),_V,L,_R,X,Set,Ratio):-
    set_bbbtree__split_lt(L,X,Set,Ratio).
set_bbbtree__split_lt_((>),V,L,R,X,Set,Ratio):-
    set_bbbtree__split_lt(R,X,Set0,Ratio),
    set_bbbtree__concat4(L,Set0,V,Set,Ratio).
set_bbbtree__split_lt_((=),_V,L,_R,_X,L,_Ratio).

set_bbbtree__split_gt(empty,_X,empty,_Ratio).
set_bbbtree__split_gt(tree(V,_N,L,R),X,Set,Ratio) :-
    compare(Result,X,V),
    set_bbbtree__split_gt_(Result,V,L,R,X,Set,Ratio).

set_bbbtree__split_gt_((>),_V,_L,R,X,Set,Ratio):-
    set_bbbtree__split_gt(R,X,Set,Ratio).
set_bbbtree__split_gt_((<),V,L,R,X,Set,Ratio):-
    set_bbbtree__split_gt(L,X,Set0,Ratio),
    set_bbbtree__concat4(Set0,R,V,Set,Ratio).
set_bbbtree__split_gt_((=),_V,_L,R,_X,R,_Ratio).

